#include "header.hpp"

#include "ppaa02.amplitude.set.hpp"

void ppaa02_amplitude_initialization(munich * MUC){
  static Logger logger("ppaa02_amplitude_initialization");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  if (MUC->osi->switch_OL){
#ifdef OPENLOOPS
    MUC->asi = new ppaa02_amplitude_OpenLoops_set();
#endif
  }
  else if (MUC->osi->switch_RCL){
#ifdef RECOLA
    MUC->asi = new ppaa02_amplitude_Recola_set();
#endif
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


ppaa02_amplitude_set::ppaa02_amplitude_set(){
  Logger logger("ppaa02_amplitude_set::ppaa02_amplitude_set ()");
  logger << LOG_DEBUG_VERBOSE << "started" << endl;

  logger << LOG_INFO << "constructor called..." << endl;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

ppaa02_amplitude_set::~ppaa02_amplitude_set(){
  Logger logger("ppaa02_amplitude_set::~ppaa02_amplitude_set ()");
  logger << LOG_DEBUG_VERBOSE << "started" << endl;

  logger << LOG_INFO << "destructor called..." << endl;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


void ppaa02_amplitude_set::calculate_H2_QCD_2loop(){
  static Logger logger("ppaa02_amplitude_set::calculate_H2_QCD_2loop");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  //  ppaa22_calculate_H2(osi);
  double s = (osi->esi->p_parton[0][1]+osi->esi->p_parton[0][2]).m2();
  double t = (osi->esi->p_parton[0][1]-osi->esi->p_parton[0][4]).m2();
  double u = (osi->esi->p_parton[0][1]-osi->esi->p_parton[0][3]).m2();
  
//   cout << "s=" << s << ", t=" << t << ", u=" << u << endl;

  double Q_factor;
  double Qq;
  if (osi->csi->subprocess == "uu~_aa") {
    Q_factor=pow(2.0/3,4);
    Qq=2.0/3;
  }
  else if (osi->csi->subprocess == "dd~_aa" || osi->csi->subprocess == "bb~_aa") {
    Q_factor=pow(1.0/3,4);
    Qq=-1.0/3;
  }
  else {
    assert(false && "ERROR: sub-process not known!");
  }
  

  osi->QT_A0 = Q_factor*ppaa22_A0(s,t,u)*pow(4*M_PI*osi->msi->alpha_e,2);
  
  osi->QT_A0 /= (4*9); // averaging factor
  osi->QT_A0 *= 0.5; // symmetry factor
  
  // H1
  osi->QT_A1 = 2*(inv2pi * osi->alpha_S)*ppaa22_H1f(s,t,u)*osi->QT_A0;
  osi->QT_H1_delta_QCD = osi->QT_A1;
  
  // H2
  osi->QT_A2 = pow(2*inv2pi * osi->alpha_S,2)*ppaa22_H2f(s,t,u,osi->N_f,Qq)*osi->QT_A0;
  osi->QT_H2_delta_QCD = osi->QT_A2;
   
  // adapt alpha_S/Pi normalisation
  osi->QT_H1_delta_QCD /= 2;
  osi->QT_H2_delta_QCD /= 4;
  
  osi->QT_H1_delta_QCD /= (inv2pi * osi->alpha_S * osi->QT_A0);
  osi->QT_H2_delta_QCD /= (pow(inv2pi * osi->alpha_S, 2) * osi->QT_A0);
  
//   cout << "H1=" << osi->QT_H1_delta_QCD << endl;
//   cout << "H2=" << osi->QT_H2_delta_QCD << endl;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


void ppaa02_amplitude_set::calculate_H1gg_QCD_2loop(){
  static Logger logger("ppaa02_amplitude_set::calculate_H1gg_QCD_2loop");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  //  ggllll34_calculate_H1(osi);
  logger << LOG_FATAL << "Amplitudes have not been implemented yet." << endl;
  exit(1);
  
  osi->QT_H1_delta_QCD = osi->QT_A1;
  osi->QT_H1_delta_QCD /= 2;
  osi->QT_H1_delta_QCD /= (osi->alpha_S * inv2pi * osi->QT_A0);
  logger << LOG_DEBUG << "osi->QT_H1_delta_QCD = " << osi->QT_H1_delta_QCD << endl;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


#ifdef OPENLOOPS

ppaa02_amplitude_OpenLoops_set::ppaa02_amplitude_OpenLoops_set(){
  Logger logger("ppaa02_amplitude_OpenLoops_set::ppaa02_amplitude_OpenLoops_set ()");
  logger << LOG_DEBUG_VERBOSE << "started" << endl;

  logger << LOG_INFO << "constructor called..." << endl;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

ppaa02_amplitude_OpenLoops_set::~ppaa02_amplitude_OpenLoops_set(){
  Logger logger("ppaa02_amplitude_OpenLoops_set::~ppaa02_amplitude_OpenLoops_set ()");
  logger << LOG_DEBUG_VERBOSE << "started" << endl;

  logger << LOG_INFO << "destructor called..." << endl;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

#endif


#ifdef RECOLA

ppaa02_amplitude_Recola_set::ppaa02_amplitude_Recola_set(){
  Logger logger("ppaa02_amplitude_Recola_set::ppaa02_amplitude_Recola_set ()");
  logger << LOG_DEBUG_VERBOSE << "started" << endl;

  logger << LOG_INFO << "constructor called..." << endl;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

ppaa02_amplitude_Recola_set::~ppaa02_amplitude_Recola_set(){
  Logger logger("ppaa02_amplitude_Recola_set::~ppaa02_amplitude_Recola_set ()");
  logger << LOG_DEBUG_VERBOSE << "started" << endl;

  logger << LOG_INFO << "destructor called..." << endl;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

#endif
