#include "header.hpp"

#include "pplla03.amplitude.set.hpp"

// 1-loop helicity amplitude coefficients for qq_bar -> Wgam/Zgam,
// see T. Gehrmann, L. Tancredi, 1112.1531

double_complex pplla03_amplitude_set::alpha_1_1(){
  double_complex alpha =
    C_F * (
	   + zeta2
	   - GHPLs1[3] * HPLs1[0]
	   + GHPLs2[0*4+3]
	   + GHPLs1[0] * HPLs1[0]
	   - HPLs2[0*2+0]
	   + HPLs2[0*2+1]
	   )
    + C_F * ri * M_PI * (
			 + GHPLs1[3]
			 + HPLs1[0]
			 );
  return alpha;
}


double_complex pplla03_amplitude_set::alpha_1_2(){
  double_complex alpha =
    C_F * pow(u,-2) * (
		       + GHPLs1[2] * HPLs1[0]
		       + GHPLs2[2*4+1]
		       - GHPLs1[1] * HPLs1[0]
		       - GHPLs1[1] * HPLs1[1]
		       )
    + C_F * pow(u,-2) * v * (
			     - 2.0 * GHPLs1[2] * HPLs1[0]
			     - 2.0 * GHPLs2[2*4+1]
			     + 2.0 * GHPLs1[1] * HPLs1[0]
			     + 2.0 * GHPLs1[1] * HPLs1[1]
			     )
    + C_F * pow(u,-2) * pow(v,2) * (
				    + GHPLs1[2] * HPLs1[0]
				    + GHPLs2[2*4+1]
				    - GHPLs1[1] * HPLs1[0]
				    - GHPLs1[1] * HPLs1[1]
				    )
    + C_F * pow(u,-1) * (
			 + GHPLs1[2]
			 - HPLs1[1]
			 )
    + C_F * pow(u,-1) * v * (
			     - 2.0 * GHPLs1[2]
			     + HPLs1[0]
			     + 2.0 * HPLs1[1]
			     )
    + C_F * pow(u,-1) * pow(v,2) * (
				    + GHPLs1[2]
				    - HPLs1[0]
				    - HPLs1[1]
				    )
    + C_F * (
	     - 7.0/2
	     + 1.0/2 * GHPLs1[2]
	     - 1.0/2 * HPLs1[1]
	     )
    + C_F * v * pow(1-u,-1) * (
			       - 1.0/2
			       - GHPLs1[2]
			       + HPLs1[0]
			       + HPLs1[1]
			       )
    + C_F * pow(v,2) * pow(1-u,-2) * (
				      - 1.0/2 * GHPLs1[2]
				      + 1.0/2 * HPLs1[0]
				      + 1.0/2 * HPLs1[1]
				      )
    + C_F * pow(v,2) * pow(1-u,-1) * (
				      + GHPLs1[2]
				      - HPLs1[0]
				      - HPLs1[1]
				      )
    + C_F * ri * M_PI * pow(u,-2) * (
				     + GHPLs1[1]
				     )
    + C_F * ri * M_PI * pow(u,-2) * v * (
					 - 2.0 * GHPLs1[1]
					 )
    + C_F * ri * M_PI * pow(u,-2) * pow(v,2) * (
						+ GHPLs1[1]
						)
    + C_F * ri * M_PI * pow(u,-1) * (
				     + 1.0
				     )
    + C_F * ri * M_PI * pow(u,-1) * v * (
					 - 2.0
					 )
    + C_F * ri * M_PI * pow(u,-1) * pow(v,2) * (
						+ 1.0
						)
    + C_F * ri * M_PI * (
			 + 1.0/2
			 )
    + C_F * ri * M_PI * v * pow(1-u,-1) * (
					   - 1.0
					   )
    + C_F * ri * M_PI * pow(v,2) * pow(1-u,-2) * (
						  - 1.0/2
						  )
    + C_F * ri * M_PI * pow(v,2) * pow(1-u,-1) * (
						  + 1.0
						  );
  return alpha;
}

double_complex pplla03_amplitude_set::alpha_1_3(){
  double_complex alpha =
    C_F * (
	   - 4.0
	   )
    + C_F * u * pow(1-v,-1) * (
			       + 4.0
			       );
  return alpha;
}

double_complex pplla03_amplitude_set::beta_1_1(){
  double_complex beta =
    C_F * pow(1-v,-1) * (
			 - 3.0/2 * HPLs1[0]
			 )
    + C_F * (
	     - 9.0/2
	     + zeta2
	     - GHPLs1[3] * HPLs1[0]
	     + GHPLs2[0*4+3]
	     + GHPLs1[0] * HPLs1[0]
	     + 3.0/2 * HPLs1[0]
	     - HPLs2[0*2+0]
	     + HPLs2[0*2+1]
	     )
    + C_F * u * pow(1-v,-2) * (
			       + 1.0/2 * HPLs1[0]
			       )
    + C_F * u * pow(1-v,-1) * (
			       + 1.0/2
			       - 1.0/2 * HPLs1[0]
			       )
    + C_F * ri * M_PI * (
			 + GHPLs1[3]
			 + HPLs1[0]
			 );
  return beta;
}

double_complex pplla03_amplitude_set::beta_1_2(){
  double_complex beta =
    C_F * pow(u,-2) * (
		       + GHPLs1[2] * HPLs1[0]
		       + GHPLs2[2*4+1]
		       - GHPLs1[1] * HPLs1[0]
		       - GHPLs1[1] * HPLs1[1]
		       )

    + C_F * pow(u,-2) * v * (
			     - GHPLs1[2] * HPLs1[0]
			     - GHPLs2[2*4+1]
			     + GHPLs1[1] * HPLs1[0]
			     + GHPLs1[1] * HPLs1[1]
			     )

    + C_F * pow(u,-1) * (
			 + GHPLs1[2]
			 - HPLs1[1]
			 )

    + C_F * pow(u,-1) * v * (
			     - GHPLs1[2]
			     + GHPLs1[2] * HPLs1[0]
			     + GHPLs2[2*4+1]
			     - GHPLs1[1] * HPLs1[0]
			     - GHPLs1[1] * HPLs1[1]
			     + HPLs1[0]
			     + HPLs1[1]
			     )

    + C_F * pow(1-v,-1) * (
			   + 3.0/2 * HPLs1[0]
			   )

    + C_F * (
	     + 1.0/2
	     + 1.0/2 * GHPLs1[2]
	     - 3.0/2 * HPLs1[0]
	     - 1.0/2 * HPLs1[1]
	     )

    + C_F * v * pow(1-u,-1) * (
			       + 1.0/2 * GHPLs1[2]
			       - 1.0/2 * HPLs1[0]
			       - 1.0/2 * HPLs1[1]
			       )

    + C_F * u * pow(1-v,-2) * (
			       + 1.0/2 * HPLs1[0]
			       )

    + C_F * u * pow(1-v,-1) * (
			       + 1.0/2
			       - 1.0/2 * HPLs1[0]
			       )

    + C_F * ri * M_PI * pow(u,-2) * (
				     + GHPLs1[1]
				     )

    + C_F * ri * M_PI * pow(u,-2) * v * (
					 - GHPLs1[1]
					 )

    + C_F * ri * M_PI * pow(u,-1) * (
				     + 1.0
				     )

    + C_F * ri * M_PI * pow(u,-1) * v * (
					 - 1.0
					 + GHPLs1[1]
					 )

    + C_F * ri * M_PI * (
			 + 1.0/2
			 )

    + C_F * ri * M_PI * v * pow(1-u,-1) * (
					   + 1.0/2
					   );
  return beta;
}

double_complex pplla03_amplitude_set::beta_1_3(){
  double_complex beta =
    C_F * u * pow(1-v,-1) * (
			     + 4.0
			     );
  return beta;
}

double_complex pplla03_amplitude_set::gamma_1_1(){
  double_complex gamma =
    C_F * pow(1-v,-1) * (
			 + 1.0/2 * HPLs1[0]
			 )

    + C_F * (
	     + 1.0/2
	     - 1.0/2 * HPLs1[0]
	     )

    + C_F * u * pow(1-v,-2) * (
			       - 1.0/2 * HPLs1[0]
			       )

    + C_F * u * pow(1-v,-1) * (
			       - 1.0/2
			       + 1.0/2 * HPLs1[0]
			       );
  return gamma;
}

double_complex pplla03_amplitude_set::gamma_1_2(){
  double_complex gamma =
    C_F * pow(u,-2) * v * (
			   + GHPLs1[2] * HPLs1[0]
			   + GHPLs2[2*4+1]
			   - GHPLs1[1] * HPLs1[0]
			   - GHPLs1[1] * HPLs1[1]
			   )

    + C_F * pow(u,-2) * pow(v,2) * (
				    - GHPLs1[2] * HPLs1[0]
				    - GHPLs2[2*4+1]
				    + GHPLs1[1] * HPLs1[0]
				    + GHPLs1[1] * HPLs1[1]
				    )

    + C_F * pow(u,-1) * v * (
			     + GHPLs1[2]
			     - GHPLs1[2] * HPLs1[0]
			     - GHPLs2[2*4+1]
			     + GHPLs1[1] * HPLs1[0]
			     + GHPLs1[1] * HPLs1[1]
			     - HPLs1[1]
			     )

    + C_F * pow(u,-1) * pow(v,2) * (
				    - GHPLs1[2]
				    + HPLs1[0]
				    + HPLs1[1]
				    )

    + C_F * pow(1-v,-1) * (
			   - 1.0/2 * HPLs1[0]
			   )

    + C_F * (
	     + 1.0/2 * HPLs1[0]
	     )

    + C_F * v * pow(1-u,-1) * (
			       + 1.0/2
			       - 1.0/2 * GHPLs1[2]
			       + 1.0/2 * HPLs1[0]
			       + 1.0/2 * HPLs1[1]
			       )

    + C_F * pow(v,2) * pow(1-u,-2) * (
				      + 1.0/2 * GHPLs1[2]
				      - 1.0/2 * HPLs1[0]
				      - 1.0/2 * HPLs1[1]
				      )

    + C_F * pow(v,2) * pow(1-u,-1) * (
				      - GHPLs1[2]
				      + HPLs1[0]
				      + HPLs1[1]
				      )

    + C_F * u * pow(1-v,-2) * (
			       - 1.0/2 * HPLs1[0]
			       )

    + C_F * u * pow(1-v,-1) * (
			       - 1.0/2
			       + 1.0/2 * HPLs1[0]
			       )

    + C_F * ri * M_PI * pow(u,-2) * v * (
					 + GHPLs1[1]
					 )

    + C_F * ri * M_PI * pow(u,-2) * pow(v,2) * (
						- GHPLs1[1]
						)

    + C_F * ri * M_PI * pow(u,-1) * v * (
					 + 1.0
					 - GHPLs1[1]
					 )

    + C_F * ri * M_PI * pow(u,-1) * pow(v,2) * (
						- 1.0
						)

    + C_F * ri * M_PI * v * pow(1-u,-1) * (
					   - 1.0/2
					   )

    + C_F * ri * M_PI * pow(v,2) * pow(1-u,-2) * (
						  + 1.0/2
						  )

    + C_F * ri * M_PI * pow(v,2) * pow(1-u,-1) * (
						  - 1.0
						  );
  return gamma;
}

double_complex pplla03_amplitude_set::gamma_1_3(){
  double_complex gamma = 0;
  //          C_F * u * pow(1-v,-1) * (
  //          + 4.0
  //          );
  return gamma;
}

void pplla03_amplitude_set::Omega_1_ZP(double e_q){
  //, double_complex & alpha, double_complex & beta, double_complex & gamma
  //  double_complex Omega1, Omega2;
  //  const double zeta2 = pi2_6;
  //  const double CF=4.0/3;

  alpha_1 = e_q * (alpha_1_1() + alpha_1_2());
  beta_1 = e_q * (beta_1_1() + beta_1_2());
  gamma_1 = e_q * (gamma_1_1() + gamma_1_2());
  
  /*
  alpha_1_1(Omega1);
  alpha_1_2(Omega2);
  alpha = e_q * (Omega1 + Omega2);

  beta_1_1(Omega1);
  beta_1_2(Omega2);
  beta = e_q * (Omega1 + Omega2);

  gamma_1_1(Omega1);
  gamma_1_2(Omega2);
  gamma = e_q * (Omega1 + Omega2);
  */
}

void pplla03_amplitude_set::Omega_1_W(double e_q, double e_qprime){
  //, double_complex & alpha, double_complex & beta, double_complex & gamma
  //  double_complex Omega1, Omega2, Omega3;
  //  const double zeta2 = pi2_6;
  //  const double CF = 4.0/3;

  alpha_1 = e_qprime * alpha_1_1() + e_q * alpha_1_2() + alpha_1_3();
  beta_1 = e_qprime * beta_1_1() + e_q * beta_1_2() + beta_1_3();
  gamma_1 = e_qprime * gamma_1_1() + e_q * gamma_1_2() + gamma_1_3();

  /*
  alpha_1_1(u, v, C_F, zeta2, GHPLs1, GHPLs2, HPLs1, HPLs2, Omega1);
  alpha_1_2(u, v, C_F, zeta2, GHPLs1, GHPLs2, HPLs1, HPLs2, Omega2);
  alpha_1_3(u, v, C_F, zeta2, GHPLs1, GHPLs2, HPLs1, HPLs2, Omega3);
  alpha = e_qprime * Omega1 + e_q * Omega2 + Omega3;

  beta_1_1(u, v, C_F, zeta2, GHPLs1, GHPLs2, HPLs1, HPLs2, Omega1);
  beta_1_2(u, v, C_F, zeta2, GHPLs1, GHPLs2, HPLs1, HPLs2, Omega2);
  beta_1_3(u, v, C_F, zeta2, GHPLs1, GHPLs2, HPLs1, HPLs2, Omega3);
  beta = e_qprime * Omega1 + e_q * Omega2 + Omega3;

  gamma_1_1(u, v, C_F, zeta2, GHPLs1, GHPLs2, HPLs1, HPLs2, Omega1);
  gamma_1_2(u, v, C_F, zeta2, GHPLs1, GHPLs2, HPLs1, HPLs2, Omega2);
  gamma_1_3(u, v, C_F, zeta2, GHPLs1, GHPLs2, HPLs1, HPLs2, Omega3);
  gamma = e_qprime * Omega1 + e_q * Omega2 + Omega3;
  */
}
