#include "header.hpp"

#include "pplla03.amplitude.set.hpp"
#include "pplla23.amplitude.doublevirtual.hpp"

//double_complex pplla03_amplitude_set::alpha0_Z(double e_q){return e_q;}
//double_complex pplla03_amplitude_set::beta0_Z(double e_q){return e_q;}
//double_complex pplla03_amplitude_set::alpha0_W(double u, double v, double e_q, double e_qprime){return e_q + (e_qprime - e_q) * (1. - u / (1. - v));}
//double_complex pplla03_amplitude_set::beta0_W(double u, double v, double e_q, double e_qprime){return e_qprime - (e_qprime - e_q) * u / (1. - v);}

void pplla03_amplitude_set::calculate_helicity_amplitude_perm(int d, int e, int a, int b, int c, double e_q, vector<double_complex> & Aplus_0, vector<double_complex> & Aplus_1, vector<double_complex> & Aplus_2_VR, vector<double_complex> & Aplus_2_VL, vector<double_complex> & Aplus_2_P, int order){
  static Logger logger("pplla03_amplitude_set::calculate_helicity_amplitude_perm");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  // adapt conventions
  d -= 2;
  e -= 2;
  if (a == 3){a = 5;}
  if (b == 3){b = 5;}
  if (c == 3){c = 5;}

  /////  double q2 = (p_amp[d] + p_amp[e]).m2();
  /////  double shat = (p_amp[a] + p_amp[c]).m2();
  //  double u = -s_inv[a][b] / s_inv[a][c];
  //  double v = s34 / s_inv[a][c];
  //  logger << LOG_DEBUG << "s12         = " << s12 << endl;
  //  logger << LOG_DEBUG << "s_inv[" << a << "][" << c << "] = " << s_inv[a][c] << endl;

  u = -s_inv[a][b] / s12;
  v = s34 / s12;
  double_complex alpha0;
  double_complex beta0;
  double e_qprime = 0.;

  if (no_process_class == 1 || no_process_class == 4){
    alpha0 = e_q;
    beta0 = e_q;
    //    alpha0 = alpha0_Z(e_q);
    //    beta0 = beta0_Z(e_q);
    //    gamma0 = gamma0_Z();
  } 
  else if (no_process_class == 2 || no_process_class == 3){
    if (no_process_class == 2) {
      if (a == 1){
        e_q = -1.0 / 3;
        e_qprime = 2.0 / 3;
      }
      else {
        e_q = -2.0 / 3;
        e_qprime = 1.0 / 3;
      }
    }
    else {
      if (a == 1) {
        e_q = -2.0 / 3;
        e_qprime = 1.0 / 3;
      }
      else {
        e_q = -1.0 / 3;
        e_qprime = 2.0 / 3;
      }
    }
    alpha0 = e_q + (e_qprime - e_q) * (1. - u / (1. - v));
    beta0 = e_qprime - (e_qprime - e_q) * u / (1. - v);
    //    alpha0 = alpha0_W(u, v, e_q, e_qprime);
    //    beta0 = beta0_W(u, v, e_q, e_qprime);
  }

  // Born
  Aplus_0[0] = alpha0 * (za[c][d] * za[a][c] * zb[a][e] / (za[a][b] * za[c][b]));
  Aplus_0[0] -= beta0 * (za[c][d] * zb[b][e] / za[a][b]);
  //  Aplus_0[0] += gamma0 * (za[a][d] * zb[a][b] * zb[b][e] / (za[a][b] * zb[c][b]));
  Aplus_0[0] *= -(2 * sqrt2);

  Aplus_0[1] = alpha0 * conj(za[c][d] * za[a][c] * zb[a][e] / (za[a][b] * za[c][b]));
  Aplus_0[1] -= beta0 * conj(za[c][d] * zb[b][e] / za[a][b]);
  //  Aplus_0[1] += gamma0 * conj(za[a][d] * zb[a][b] * zb[b][e] / (za[a][b] * zb[c][b]));
  Aplus_0[1] *= -(2 * sqrt2);

  //  double GYZ1[4], GYZ2[4 * 4], GYZ3[4 * 4 * 4], GYZ4[4 * 4 * 4 * 4];
  //  double HZ1[2], HZ2[2 * 2], HZ3[2 * 2 * 2], HZ4[2 * 2 * 2 * 2];

  if (no_process_class == 1 || no_process_class == 4) {
    /*
    double_complex alpha1;
    double_complex beta1;
    double_complex gamma1;
    double_complex alpha2_VR;
    double_complex beta2_VR;
    double_complex gamma2_VR;
    double_complex alpha2_VL;
    double_complex beta2_VL;
    double_complex gamma2_VL;
    double_complex alpha2_P;
    double_complex beta2_P;
    double_complex gamma2_P;
    */
    if (order >= 1){
      if (order == 1){
	// old version:        compute_GHPLs(2, u, v, GYZ1, GYZ2, GYZ3, GYZ4, HZ1, HZ2, HZ3, HZ4);
        compute_GHPLs(3, u, v, GHPLs1, GHPLs2, GHPLs3, GHPLs4, HPLs1, HPLs2, HPLs3, HPLs4);
	//        compute_GHPLs(3, u, v, GYZ1, GYZ2, GYZ3, GYZ4, HZ1, HZ2, HZ3, HZ4);
        // compute_GHPLs_Ginac(3, u, v, GYZ1g, GYZ2g, GYZ3g, GYZ4g, HZ1g, HZ2g, HZ3g, HZ4g);
      }
      else if (order == 2){
        compute_GHPLs(4, u, v, GHPLs1, GHPLs2, GHPLs3, GHPLs4, HPLs1, HPLs2, HPLs3, HPLs4);
	//        compute_GHPLs(4, u, v, GYZ1, GYZ2, GYZ3, GYZ4, HZ1, HZ2, HZ3, HZ4);
	// compute_GHPLs_Ginac(4, u, v, GYZ1, GYZ2, GYZ3, GYZ4, HZ1, HZ2, HZ3, HZ4);
      }
      Omega_1_ZP(e_q);
      //      Omega_1_ZP(e_q, alpha_1, beta_1, gamma_1);
      //      Omega_1_ZP(u, v, e_q, GYZ1, GYZ2, HZ1, HZ2, alpha1, beta1, gamma1);

      Aplus_1[0] = alpha_1 * (za[c][d] * za[a][c] * zb[a][e] / (za[a][b] * za[c][b]));
      Aplus_1[0] -= beta_1 * (za[c][d] * zb[b][e] / za[a][b]);
      Aplus_1[0] += gamma_1 * (za[a][d] * zb[a][b] * zb[b][e] / (za[a][b] * zb[c][b]));
      Aplus_1[0] *= -(2 * sqrt2);

      Aplus_1[1] = alpha_1 * conj(za[c][d] * za[a][c] * zb[a][e] / (za[a][b] * za[c][b]));
      Aplus_1[1] -= beta_1 * conj(za[c][d] * zb[b][e] / za[a][b]);
      Aplus_1[1] += gamma_1 * conj(za[a][d] * zb[a][b] * zb[b][e] / (za[a][b] * zb[c][b]));
      Aplus_1[1] *= -(2 * sqrt2);

      if (order == 2){
        Omega_2_ZP(e_q);
	//        Omega_2_ZP(u, v, e_q, msi->csin_w, GHPLs1, GHPLs2, GHPLs3, GHPLs4, HPLs1, HPLs2, HPLs3, HPLs4, alpha2_VR, beta2_VR, gamma2_VR, alpha2_VL, beta2_VL, gamma2_VL, alpha2_P, beta2_P, gamma2_P);
	//        Omega_2_ZP(u, v, e_q, msi->csin_w, GYZ1, GYZ2, GYZ3, GYZ4, HZ1, HZ2, HZ3, HZ4, alpha2_VR, beta2_VR, gamma2_VR, alpha2_VL, beta2_VL, gamma2_VL, alpha2_P, beta2_P, gamma2_P);

        Aplus_2_VR[0] = alpha_2_ZR * (za[c][d] * za[a][c] * zb[a][e] / (za[a][b] * za[c][b]));
        Aplus_2_VR[0] -= beta_2_ZR * (za[c][d] * zb[b][e] / za[a][b]);
        Aplus_2_VR[0] += gamma_2_ZR * (za[a][d] * zb[a][b] * zb[b][e] / (za[a][b] * zb[c][b]));
        Aplus_2_VR[0] *= -(2 * sqrt2);

        Aplus_2_VL[0] = alpha_2_ZL * (za[c][d] * za[a][c] * zb[a][e] / (za[a][b] * za[c][b]));
        Aplus_2_VL[0] -= beta_2_ZL * (za[c][d] * zb[b][e] / za[a][b]);
        Aplus_2_VL[0] += gamma_2_ZL * (za[a][d] * zb[a][b] * zb[b][e] / (za[a][b] * zb[c][b]));
        Aplus_2_VL[0] *= -(2 * sqrt2);

        Aplus_2_P[0] = alpha_2_P * (za[c][d] * za[a][c] * zb[a][e] / (za[a][b] * za[c][b]));
        Aplus_2_P[0] -= beta_2_P * (za[c][d] * zb[b][e] / za[a][b]);
        Aplus_2_P[0] += gamma_2_P * (za[a][d] * zb[a][b] * zb[b][e] / (za[a][b] * zb[c][b]));
        Aplus_2_P[0] *= -(2 * sqrt2);

        Aplus_2_VR[1] = alpha_2_ZR * conj(za[c][d] * za[a][c] * zb[a][e] / (za[a][b] * za[c][b]));
        Aplus_2_VR[1] -= beta_2_ZR * conj(za[c][d] * zb[b][e] / za[a][b]);
        Aplus_2_VR[1] += gamma_2_ZR * conj(za[a][d] * zb[a][b] * zb[b][e] / (za[a][b] * zb[c][b]));
        Aplus_2_VR[1] *= -(2 * sqrt2);

        Aplus_2_VL[1] = alpha_2_ZL * conj(za[c][d] * za[a][c] * zb[a][e] / (za[a][b] * za[c][b]));
        Aplus_2_VL[1] -= beta_2_ZL * conj(za[c][d] * zb[b][e] / za[a][b]);
        Aplus_2_VL[1] += gamma_2_ZL * conj(za[a][d] * zb[a][b] * zb[b][e] / (za[a][b] * zb[c][b]));
        Aplus_2_VL[1] *= -(2 * sqrt2);

        Aplus_2_P[1] = alpha_2_P * conj(za[c][d] * za[a][c] * zb[a][e] / (za[a][b] * za[c][b]));
        Aplus_2_P[1] -= beta_2_P * conj(za[c][d] * zb[b][e] / za[a][b]);
        Aplus_2_P[1] += gamma_2_P * conj(za[a][d] * zb[a][b] * zb[b][e] / (za[a][b] * zb[c][b]));
        Aplus_2_P[1] *= -(2 * sqrt2);
      }
    }
  }
  else if (no_process_class == 2 || no_process_class == 3){
    /*
    double_complex alpha1;
    double_complex beta1;
    double_complex gamma1;
    double_complex alpha2_W;
    double_complex beta2_W;
    double_complex gamma2_W;
    */
    if (order >= 1){
      if (order == 1){
	// Should this one also be changed as for Zgamma production (2->3) ???
	// compute_GHPLs(3, u, v, GYZ1, GYZ2, GYZ3, GYZ4, HZ1, HZ2, HZ3, HZ4);
        compute_GHPLs(2, u, v, GHPLs1, GHPLs2, GHPLs3, GHPLs4, HPLs1, HPLs2, HPLs3, HPLs4);
	//        compute_GHPLs(2, u, v, GYZ1, GYZ2, GYZ3, GYZ4, HZ1, HZ2, HZ3, HZ4);
      }
      else if (order == 2){
        compute_GHPLs(4, u, v, GHPLs1, GHPLs2, GHPLs3, GHPLs4, HPLs1, HPLs2, HPLs3, HPLs4);
	//        compute_GHPLs(4, u, v, GYZ1, GYZ2, GYZ3, GYZ4, HZ1, HZ2, HZ3, HZ4);
      }
      Omega_1_W(e_q, e_qprime);
      //      Omega_1_W(e_q, e_qprime, alpha_1, beta_1, gamma_1);
      //      Omega_1_W(u, v, e_q, e_qprime, GYZ1, GYZ2, HZ1, HZ2, alpha1, beta1, gamma1);

      Aplus_1[0] = alpha_1 * (za[c][d] * za[a][c] * zb[a][e] / (za[a][b] * za[c][b]));
      Aplus_1[0] -= beta_1 * (za[c][d] * zb[b][e] / za[a][b]);
      Aplus_1[0] += gamma_1 * (za[a][d] * zb[a][b] * zb[b][e] / (za[a][b] * zb[c][b]));
      Aplus_1[0] *= -(2 * sqrt2);

      Aplus_1[1] = alpha_1 * conj(za[c][d] * za[a][c] * zb[a][e] / (za[a][b] * za[c][b]));
      Aplus_1[1] -= beta_1 * conj(za[c][d] * zb[b][e] / za[a][b]);
      Aplus_1[1] += gamma_1 * conj(za[a][d] * zb[a][b] * zb[b][e] / (za[a][b] * zb[c][b]));
      Aplus_1[1] *= -(2 * sqrt2);

      if (order == 2) {
        Omega_2_W(e_q, e_qprime);
	//        Omega_2_W(u, v, e_q, e_qprime, GYZ1, GYZ2, GYZ3, GYZ4, HZ1, HZ2, HZ3, HZ4, alpha2_W, beta2_W, gamma2_W);
	//        Omega_2_W(u, v, e_q, e_qprime, GYZ1, GYZ2, GYZ3, GYZ4, HZ1, HZ2, HZ3, HZ4, alpha2_W, beta2_W, gamma2_W);

        Aplus_2_VR[0] = alpha_2_W * (za[c][d] * za[a][c] * zb[a][e] / (za[a][b] * za[c][b]));
        Aplus_2_VR[0] -= beta_2_W * (za[c][d] * zb[b][e] / za[a][b]);
        Aplus_2_VR[0] += gamma_2_W * (za[a][d] * zb[a][b] * zb[b][e] / (za[a][b] * zb[c][b]));
        Aplus_2_VR[0] *= -(2 * sqrt2);

        Aplus_2_VR[1] = alpha_2_W * conj(za[c][d] * za[a][c] * zb[a][e] / (za[a][b] * za[c][b]));
        Aplus_2_VR[1] -= beta_2_W * conj(za[c][d] * zb[b][e] / za[a][b]);
        Aplus_2_VR[1] += gamma_2_W * conj(za[a][d] * zb[a][b] * zb[b][e] / (za[a][b] * zb[c][b]));
        Aplus_2_VR[1] *= -(2 * sqrt2);
      }
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


void pplla03_amplitude_set::calculate_helicity_amplitude_perm_FSR(int d, int e, int a, int b, int c, vector<double_complex> & Aplus_0, vector<double_complex> & Aplus_1, vector<double_complex> & Aplus_2, int order){
  static Logger logger("pplla03_amplitude_set::calculate_helicity_amplitude_perm_FSR");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  // adapt conventions
  d -= 2;
  e -= 2;
  b = 5;
  // d = 5, e = 6; a = 1, b = 3, c = 2

  // Born
  if (no_process_class == 1) {
    Aplus_0[0] = za[a][e] * za[c][d] * zb[a][e] / (za[b][e] * za[a][b]);
    Aplus_0[0] -= za[c][d] * zb[a][b] / za[b][e];
    Aplus_0[0] -= za[a][d] * zb[a][e] * za[c][d] / (za[a][b] * za[b][d]);
  }
  else if (no_process_class == 2 || no_process_class == 3) {
    /////    double q2 = (p_amp[d] + p_amp[e]).m2();
    /////    double shat = (p_amp[a] + p_amp[c]).m2();
    if ( (a == 1 && no_process_class == 2) || (a == 2 && no_process_class == 3)) {
      Aplus_0[0] = -za[a][e] * za[c][d] * zb[a][e] / (za[b][e] * za[a][b]);
      Aplus_0[0] += za[c][d] * zb[a][b] / za[b][e];
    }
    else {
      // note that the crossing can not be done as in 1112.1531 for the FSR part
      Aplus_0[0] = -conj(-za[a][e] * zb[c][d] * zb[a][d] / (zb[b][d] * zb[a][b]));
    }
    Aplus_0[0] += s_inv[c][b] / (s12 - s34) * (za[c][d] * za[a][c] * zb[a][e] / (za[a][b] * za[c][b]));
    Aplus_0[0] += s_inv[a][b] / (s12 - s34) * (za[c][d] * zb[b][e] / za[a][b]);
  }
  else if (no_process_class == 4){
    Aplus_0[0] = 0;
    Aplus_0[1] = 0;
    Aplus_1[0] = 0;
    Aplus_1[1] = 0;
    Aplus_2[0] = 0;
    Aplus_2[1] = 0;
    return;
  }
  
  Aplus_0[0] *= (-2 * sqrt2);
  Aplus_0[1] = conj(Aplus_0[0]);
  if (order >= 1) {
    Aplus_1[0] = F1 * Aplus_0[0];
    Aplus_1[1] = F1 * Aplus_0[1];
    if (order == 2) {
      /////    double LR = log(s_inv[3][4] / s_inv[1][2]);
      F2 = F2_coefficient_1 + (LR - ri * pi) * F2_coefficient_LR;
      Aplus_2[0] = F2 * Aplus_0[0];
      Aplus_2[1] = F2 * Aplus_0[1];
    }
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}
