#include "header.hpp"

#include "pplla03.amplitude.set.hpp"
#include "pplla23.amplitude.doublevirtual.hpp"

void pplla03_amplitude_initialization(munich * MUC){
  static Logger logger("pplla03_amplitude_initialization");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  if (MUC->osi->switch_OL){
#ifdef OPENLOOPS
    MUC->asi = new pplla03_amplitude_OpenLoops_set();
#endif
  }
  else if (MUC->osi->switch_RCL){
#ifdef RECOLA
    MUC->asi = new pplla03_amplitude_Recola_set();
#endif
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


pplla03_amplitude_set::pplla03_amplitude_set(){
  Logger logger("pplla03_amplitude_set::pplla03_amplitude_set ()");
  logger << LOG_DEBUG_VERBOSE << "started" << endl;

  logger << LOG_INFO << "constructor called..." << endl;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

pplla03_amplitude_set::~pplla03_amplitude_set(){
  Logger logger("pplla03_amplitude_set::~pplla03_amplitude_set ()");
  logger << LOG_DEBUG_VERBOSE << "started" << endl;

  logger << LOG_INFO << "destructor called..." << endl;

  if (csi->type_contribution == "VT2" && csi->type_correction == "QCD"){
    delete [] GHPLs1;
    delete [] GHPLs2;
    delete [] GHPLs3;
    delete [] GHPLs4;
    delete [] HPLs1;
    delete [] HPLs2;
    delete [] HPLs3;
    delete [] HPLs4;
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


void pplla03_amplitude_set::calculate_H2_QCD_2loop_initialization_ingredients(){
  static Logger logger("pplla03_amplitude_set::calculate_H2_QCD_2loop_initialization_ingredients");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  order = 2;
  dN_f = double(osi->N_f);
  
  if (csi->process_class == "pp-emepa+X"){ //  Z -> e⁻ e⁺ gamma 
    no_process_class = 1;
    e_f = -1.; // charge of e^-
    I3_f = -0.5; // isospin of e^-
  }
  else if (csi->process_class == "pp-emve~a+X"){ //  W⁻ -> e⁻ ne~ gamma
    no_process_class = 2;
    e_f = -1.;
    I3_f = -0.5;
  }
  else if (csi->process_class == "pp-epvea+X"){ //  W⁺ -> e⁺ ne gamma
    no_process_class = 3;
    e_f = 1.;
    I3_f = 0.5;
  }
  else if (csi->process_class == "pp-veve~a+X"){ //  Z -> ne ne~ gamma
    no_process_class = 4;
    e_f = 0.;
    I3_f = 0.5;
  }
  else {
    logger << LOG_FATAL << "calc_amplitudes: unknown process class: " << csi->process_class << endl;
    exit(1);
  }
  logger << LOG_INFO << "csi->process_class = " <<  csi->process_class << "    no_process_class = " << no_process_class << endl;
  
  if (no_process_class == 1 || no_process_class == 4) { //  Z gamma  processes
    parton_type = (abs(csi->type_parton[0][1]) + 1) % 2 + 1; //  1: down-type; 2: up-type
    if (parton_type == 1){I3_q = 0.5; e_q = 1. / 3.;}
    else if (parton_type == 2){I3_q = -0.5; e_q = -2. / 3.;}
  }

  p_amp.resize(csi->n_particle + 3);
  za.resize(p_amp.size(), vector<double_complex> (p_amp.size(), 0.));
  zb.resize(p_amp.size(), vector<double_complex> (p_amp.size(), 0.));
  s_inv.resize(p_amp.size(), vector<double> (p_amp.size(), 0.));
 
  // amplitude coefficients
  // 56132, resonant piece
  Aplus56132_0.resize(2, 0.);
  Aplus56132_1.resize(2, 0.);
  Aplus56132_2_VR.resize(2, 0.);
  Aplus56132_2_VL.resize(2, 0.);
  Aplus56132_2_P.resize(2, 0.);
  // 56132 DY like piece
  Aplus56132_0_FSR.resize(2, 0.);
  Aplus56132_1_FSR.resize(2, 0.);
  Aplus56132_2_FSR.resize(2, 0.);
  // 65231, resonant piece
  Aplus65231_0.resize(2, 0.);
  Aplus65231_1.resize(2, 0.);
  Aplus65231_2_VR.resize(2, 0.);
  Aplus65231_2_VL.resize(2, 0.);
  Aplus65231_2_P.resize(2, 0.);
  // 65231, DY like piece
  Aplus65231_0_FSR.resize(2, 0.);
  Aplus65231_1_FSR.resize(2, 0.);
  Aplus65231_2_FSR.resize(2, 0.);
  // 56231, resonant piece
  Aplus56231_0.resize(2, 0.);
  Aplus56231_1.resize(2, 0.);
  Aplus56231_2_VR.resize(2, 0.);
  Aplus56231_2_VL.resize(2, 0.);
  Aplus56231_2_P.resize(2, 0.);
  // 56231, DY like piece
  Aplus56231_0_FSR.resize(2, 0.);
  Aplus56231_1_FSR.resize(2, 0.);
  Aplus56231_2_FSR.resize(2, 0.);
  // 65132, resonant piece
  Aplus65132_0.resize(2, 0.);
  Aplus65132_1.resize(2, 0.);
  Aplus65132_2_VR.resize(2, 0.);
  Aplus65132_2_VL.resize(2, 0.);
  Aplus65132_2_P.resize(2, 0.);
  // 65132, DY like piece
  Aplus65132_0_FSR.resize(2, 0.);
  Aplus65132_1_FSR.resize(2, 0.);
  Aplus65132_2_FSR.resize(2, 0.);

  Aplus56132_0_tot.resize(2, 0.);
  Aplus65132_0_tot.resize(2, 0.);
  Aplus56231_0_tot.resize(2, 0.);
  Aplus65231_0_tot.resize(2, 0.);

  Aplus56132_1_tot.resize(2, 0.);
  Aplus65132_1_tot.resize(2, 0.);
  Aplus56231_1_tot.resize(2, 0.);
  Aplus65231_1_tot.resize(2, 0.);

  Aplus56132_2_tot.resize(2, 0.);
  Aplus65132_2_tot.resize(2, 0.);
  Aplus56231_2_tot.resize(2, 0.);
  Aplus65231_2_tot.resize(2, 0.);

  LZ_q1q2 = (I3_q - msi->csin_w * msi->csin_w * e_q) / (msi->csin_w * msi->ccos_w);
  LZ_f5f6 = (I3_f - msi->csin_w * msi->csin_w * e_f) / (msi->csin_w * msi->ccos_w);
  RZ_q1q2 = -msi->csin_w * e_q / msi->ccos_w;
  RZ_f5f6 = -msi->csin_w * e_f / msi->ccos_w;

  LP_q1q2 = -e_q;
  LP_f5f6 = -e_f;
  RP_q1q2 = LP_q1q2;
  RP_f5f6 = LP_f5f6;
 
  LW_q1q2 = 1. / (sqrt2 * msi->csin_w);
  LW_f5f6 = 1. / (sqrt2 * msi->csin_w);
  // There are actually corresponding couplings defined in msi already !!!

  
  delta_I1 = 0.5 * C_F * pi2;
  beta0 = (11 * C_A - 4 * 0.5 * osi->N_f) / 6;

  // the two-loop amplitudes are given at muR=q, so we have to use this scale here
  //  double LR = 0;
  //  double LR = log(q2 / s); // q2 != s apparently - why ???
  F1 = -4 * C_F; // too trivial - apparently not used !!!

  ///  F2 = C_F * C_F * (255.0 / 32 + 29.0 * pi2_24 - 11.0 * pi4_180 - 15.0 * zeta3 / 2) + C_A * C_F * (-51157.0 / 2592 - 107.0 * pi2_144 + 31.0 * pi4_480 + 313.0 * zeta3 / 36 - 11.0 / 48 * PolyGamma_21) + C_F * osi->N_f * (4085.0 / 1296 + 7.0 * pi2_72 + zeta3 / 18 + PolyGamma_21 / 24);
  ///  F2 += (LR - ri * pi) * (C_F * C_F * (-3.0 / 8 + pi2_2 - 6 * zeta3) + C_A * C_F * (-1339.0 / 216 - 23.0 * pi2_48 + 13.0 * zeta3 / 2) + C_F * osi->N_f * (119.0 / 108 + pi2_24));
  F2_coefficient_1 = C_F * C_F * (255.0 / 32 + 29.0 * pi2_24 - 11.0 * pi4_180 - 15.0 * zeta3 / 2)
    + C_A * C_F * (-51157.0 / 2592 - 107.0 * pi2_144 + 31.0 * pi4_480 + 313.0 * zeta3 / 36 - 11.0 / 48 * PolyGamma_21)
    + C_F * osi->N_f * (4085.0 / 1296 + 7.0 * pi2_72 + zeta3 / 18 + PolyGamma_21 / 24);
  F2_coefficient_LR = C_F * C_F * (-3.0 / 8 + pi2_2 - 6 * zeta3)
    + C_A * C_F * (-1339.0 / 216 - 23.0 * pi2_48 + 13.0 * zeta3 / 2)
    + C_F * osi->N_f * (119.0 / 108 + pi2_24);
  //  F2 = F2_coefficient_1 + (LR - ri * pi) * F2_coefficient_LR;

  // I-operator
  K = (67.0 / 18 - pi2_6) * C_A - 10.0 / 9 / 2 * osi->N_f;
  delta_qT = zeta2 * K - 2.0 / 3 * zeta3 * beta0 + C_A * (-1214.0 / 81 + 5 * zeta2 * zeta2) + osi->N_f * 164.0 / 81 + 4 * zeta3 * beta0 - 4 * zeta2 * zeta2 * C_A;
  D2a = -14.0 / 9 * beta0 * C_F - (8.0 / 9 - 7.0 / 2 * zeta3) * C_F * C_A;
  g1a = (-3 + 24 * zeta2 - 48 * zeta3) * C_F * C_F
    + (-17.0 / 3 - 88.0 / 3 * zeta2 + 24 * zeta3) * C_F * C_A
    + (4.0 / 3 + 32.0 / 3 * zeta2) * C_F * 0.5 * osi->N_f;

  //  delta_I2 = -0.5 * C_F * delta_qT + 0.5 * C_F * C_F * pi4 + C_F * (33.0 / 24 * beta0 * pi2 + 7.0 / 8 * K * pi2 + 7.0 / 12 * beta0 * PolyGamma_21);
  // scale dependent part
  // the two-loop amplitudes are given at muR=q, so we have to use this scale here
  //  double LR = log(q2 / s);
  //  delta_I2 = delta_I2 - 1. / 12 * LR * (24 * D2a + 3 * g1a + 36 * C_F * K - 7 * beta0 * C_F * pi2);
  delta_I2_coefficient_1 = -0.5 * C_F * delta_qT
    + 0.5 * C_F * C_F * pi4
    + C_F * (33.0 / 24 * beta0 * pi2 + 7.0 / 8 * K * pi2 + 7.0 / 12 * beta0 * PolyGamma_21);
  delta_I2_coefficient_LR = - 1. / 12 * (24 * D2a + 3 * g1a + 36 * C_F * K - 7 * beta0 * C_F * pi2);
  //  delta_I2 = delta_I2 + LR * delta_I2_coefficient_LR;

  HPLs1 = new double[2];
  HPLs2 = new double[2 * 2];
  HPLs3 = new double[2 * 2 * 2];
  HPLs4 = new double[2 * 2 * 2 * 2];
  GHPLs1 = new double[4];
  GHPLs2 = new double[4 * 4];
  GHPLs3 = new double[4 * 4 * 4];
  GHPLs4 = new double[4 * 4 * 4 * 4];
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


void pplla03_amplitude_set::calculate_H2_QCD_2loop(){
  static Logger logger("pplla03_amplitude_set::calculate_H2_QCD_2loop");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  p_amp[0] = esi->p_parton[0][0];
  p_amp[1] = esi->p_parton[0][2];
  p_amp[2] = esi->p_parton[0][1];
  p_amp[3] = esi->p_parton[0][4];
  p_amp[4] = esi->p_parton[0][3];
  p_amp[5] = -esi->p_parton[0][5];

  if (no_process_class == 3){swap(p_amp[3], p_amp[4]);}

  calculate_spinorproducts(p_amp, za, zb, s_inv);

  // misleading notation ??? q2 is normally the m2 of the full colourless system !!! -> renamed q2 -> s34
  s34 = (p_amp[3] + p_amp[4]).m2();
  s12 = (p_amp[1] + p_amp[2]).m2();
  LR = log(s34 / s12);
 
  // 56132, resonant piece
  calculate_helicity_amplitude_perm(5, 6, 1, 3, 2, e_q, Aplus56132_0, Aplus56132_1, Aplus56132_2_VR, Aplus56132_2_VL, Aplus56132_2_P, order);
  // 56132 DY like piece
  calculate_helicity_amplitude_perm_FSR(5, 6, 1, 3, 2, Aplus56132_0_FSR, Aplus56132_1_FSR, Aplus56132_2_FSR, order);
  // 65231, resonant piece
  calculate_helicity_amplitude_perm(6, 5, 2, 3, 1, e_q, Aplus65231_0, Aplus65231_1, Aplus65231_2_VR, Aplus65231_2_VL, Aplus65231_2_P, order);
  // 65231, DY like piece
  calculate_helicity_amplitude_perm_FSR(6, 5, 2, 3, 1, Aplus65231_0_FSR, Aplus65231_1_FSR, Aplus65231_2_FSR, order);
  
  if (no_process_class == 1 || no_process_class == 4){
    // 56231, resonant piece
    calculate_helicity_amplitude_perm(5, 6, 2, 3, 1, e_q, Aplus56231_0, Aplus56231_1, Aplus56231_2_VR, Aplus56231_2_VL, Aplus56231_2_P, order);
    // 56231, DY like piece
    calculate_helicity_amplitude_perm_FSR(5, 6, 2, 3, 1, Aplus56231_0_FSR, Aplus56231_1_FSR, Aplus56231_2_FSR, order);
    // 65132, resonant piece
    calculate_helicity_amplitude_perm(6, 5, 1, 3, 2, e_q, Aplus65132_0, Aplus65132_1, Aplus65132_2_VR, Aplus65132_2_VL, Aplus65132_2_P, order);
    // 65132, DY like piece
    calculate_helicity_amplitude_perm_FSR(6, 5, 1, 3, 2, Aplus65132_0_FSR, Aplus65132_1_FSR, Aplus65132_2_FSR, order);
  }

  
  osi->QT_A0 = 0.;
  osi->QT_A1 = 0.;
  osi->QT_A2 = 0.;

  if (no_process_class == 1 || no_process_class == 4){ //  Z gamma  processes
    // propagator denominators
    DZ = s34 - msi->M2_Z + ri * msi->Gamma_Z * msi->M_Z;
    DP = s34;
    // propagator denominators (FSR)
    DZ_FSR = s12 - msi->M2_Z + ri * msi->Gamma_Z * msi->M_Z;
    DP_FSR = s12;

    Aplus56132_0_tot[0] = (RZ_q1q2 * RZ_f5f6 / DZ + RP_q1q2 * RP_f5f6 / DP) * Aplus56132_0[0] + (RZ_q1q2 * RZ_f5f6 / DZ_FSR + RP_q1q2 * RP_f5f6 / DP_FSR) * Aplus56132_0_FSR[0];
    Aplus65132_0_tot[1] = -(LZ_q1q2 * RZ_f5f6 / DZ + LP_q1q2 * RP_f5f6 / DP) * Aplus65132_0[1] + (LZ_q1q2 * RZ_f5f6 / DZ_FSR + LP_q1q2 * RP_f5f6 / DP_FSR) * Aplus65132_0_FSR[1];
    Aplus56231_0_tot[0] = -(LZ_q1q2 * RZ_f5f6 / DZ + LP_q1q2 * RP_f5f6 / DP) * Aplus56231_0[0] + (LZ_q1q2 * RZ_f5f6 / DZ_FSR + LP_q1q2 * RP_f5f6 / DP_FSR) * Aplus56231_0_FSR[0];
    Aplus65231_0_tot[1] = (RZ_q1q2 * RZ_f5f6 / DZ + RP_q1q2 * RP_f5f6 / DP) * Aplus65231_0[1] + (RZ_q1q2 * RZ_f5f6 / DZ_FSR + RP_q1q2 * RP_f5f6 / DP_FSR) * Aplus65231_0_FSR[1];

    Aplus65132_0_tot[0] = (RZ_q1q2 * LZ_f5f6 / DZ + RP_q1q2 * LP_f5f6 / DP) * Aplus65132_0[0] - (RZ_q1q2 * LZ_f5f6 / DZ_FSR + RP_q1q2 * LP_f5f6 / DP_FSR) * Aplus65132_0_FSR[0];
    Aplus56132_0_tot[1] = -(LZ_q1q2 * LZ_f5f6 / DZ + LP_q1q2 * LP_f5f6 / DP) * Aplus56132_0[1] - (LZ_q1q2 * LZ_f5f6 / DZ_FSR + LP_q1q2 * LP_f5f6 / DP_FSR) * Aplus56132_0_FSR[1];
    Aplus65231_0_tot[0] = -(LZ_q1q2 * LZ_f5f6 / DZ + LP_q1q2 * LP_f5f6 / DP) * Aplus65231_0[0] - (LZ_q1q2 * LZ_f5f6 / DZ_FSR + LP_q1q2 * LP_f5f6 / DP_FSR) * Aplus65231_0_FSR[0];
    Aplus56231_0_tot[1] = (RZ_q1q2 * LZ_f5f6 / DZ + RP_q1q2 * LP_f5f6 / DP) * Aplus56231_0[1] - (RZ_q1q2 * LZ_f5f6 / DZ_FSR + RP_q1q2 * LP_f5f6 / DP_FSR) * Aplus56231_0_FSR[1];

    osi->QT_A0 += norm(Aplus56132_0_tot[0]);
    osi->QT_A0 += norm(Aplus65132_0_tot[1]);
    osi->QT_A0 += norm(Aplus56231_0_tot[0]);
    osi->QT_A0 += norm(Aplus65231_0_tot[1]);

    osi->QT_A0 += norm(Aplus65132_0_tot[0]);
    osi->QT_A0 += norm(Aplus56132_0_tot[1]);
    osi->QT_A0 += norm(Aplus65231_0_tot[0]);
    osi->QT_A0 += norm(Aplus56231_0_tot[1]);

    if (order >= 1) {
      Aplus56132_1_tot[0] = (RZ_q1q2 * RZ_f5f6 / DZ + RP_q1q2 * RP_f5f6 / DP) * Aplus56132_1[0] + (RZ_q1q2 * RZ_f5f6 / DZ_FSR + RP_q1q2 * RP_f5f6 / DP_FSR) * Aplus56132_1_FSR[0];
      Aplus65132_1_tot[1] = -(LZ_q1q2 * RZ_f5f6 / DZ + LP_q1q2 * RP_f5f6 / DP) * Aplus65132_1[1] + (LZ_q1q2 * RZ_f5f6 / DZ_FSR + LP_q1q2 * RP_f5f6 / DP_FSR) * Aplus65132_1_FSR[1];
      Aplus56231_1_tot[0] = -(LZ_q1q2 * RZ_f5f6 / DZ + LP_q1q2 * RP_f5f6 / DP) * Aplus56231_1[0] + (LZ_q1q2 * RZ_f5f6 / DZ_FSR + LP_q1q2 * RP_f5f6 / DP_FSR) * Aplus56231_1_FSR[0];
      Aplus65231_1_tot[1] = (RZ_q1q2 * RZ_f5f6 / DZ + RP_q1q2 * RP_f5f6 / DP) * Aplus65231_1[1] + (RZ_q1q2 * RZ_f5f6 / DZ_FSR + RP_q1q2 * RP_f5f6 / DP_FSR) * Aplus65231_1_FSR[1];
      
      Aplus65132_1_tot[0] = (RZ_q1q2 * LZ_f5f6 / DZ + RP_q1q2 * LP_f5f6 / DP) * Aplus65132_1[0] - (RZ_q1q2 * LZ_f5f6 / DZ_FSR + RP_q1q2 * LP_f5f6 / DP_FSR) * Aplus65132_1_FSR[0];
      Aplus56132_1_tot[1] = -(LZ_q1q2 * LZ_f5f6 / DZ + LP_q1q2 * LP_f5f6 / DP) * Aplus56132_1[1] - (LZ_q1q2 * LZ_f5f6 / DZ_FSR + LP_q1q2 * LP_f5f6 / DP_FSR) * Aplus56132_1_FSR[1];
      Aplus65231_1_tot[0] = -(LZ_q1q2 * LZ_f5f6 / DZ + LP_q1q2 * LP_f5f6 / DP) * Aplus65231_1[0] - (LZ_q1q2 * LZ_f5f6 / DZ_FSR + LP_q1q2 * LP_f5f6 / DP_FSR) * Aplus65231_1_FSR[0];
      Aplus56231_1_tot[1] = (RZ_q1q2 * LZ_f5f6 / DZ + RP_q1q2 * LP_f5f6 / DP) * Aplus56231_1[1] - (RZ_q1q2 * LZ_f5f6 / DZ_FSR + RP_q1q2 * LP_f5f6 / DP_FSR) * Aplus56231_1_FSR[1];
      
      osi->QT_A1 += 2 * real(Aplus56132_0_tot[0] * conj(Aplus56132_1_tot[0]));
      osi->QT_A1 += 2 * real(Aplus65132_0_tot[1] * conj(Aplus65132_1_tot[1]));
      osi->QT_A1 += 2 * real(Aplus56231_0_tot[0] * conj(Aplus56231_1_tot[0]));
      osi->QT_A1 += 2 * real(Aplus65231_0_tot[1] * conj(Aplus65231_1_tot[1]));
      
      osi->QT_A1 += 2 * real(Aplus65132_0_tot[0] * conj(Aplus65132_1_tot[0]));
      osi->QT_A1 += 2 * real(Aplus56132_0_tot[1] * conj(Aplus56132_1_tot[1]));
      osi->QT_A1 += 2 * real(Aplus65231_0_tot[0] * conj(Aplus65231_1_tot[0]));
      osi->QT_A1 += 2 * real(Aplus56231_0_tot[1] * conj(Aplus56231_1_tot[1]));

      if (order == 2) {
  	Aplus56132_2_tot[0] = RZ_q1q2 * RZ_f5f6 / DZ * Aplus56132_2_VR[0] + RP_q1q2 * RP_f5f6 / DP * Aplus56132_2_P[0] + (RZ_q1q2 * RZ_f5f6 / DZ_FSR + RP_q1q2 * RP_f5f6 / DP_FSR) * Aplus56132_2_FSR[0];
	Aplus65132_2_tot[1] = -(LZ_q1q2 * RZ_f5f6 / DZ * Aplus65132_2_VL[1] + LP_q1q2 * RP_f5f6 / DP * Aplus65132_2_P[1]) + (LZ_q1q2 * RZ_f5f6 / DZ_FSR + LP_q1q2 * RP_f5f6 / DP_FSR) * Aplus65132_2_FSR[1];
	Aplus56231_2_tot[0] = -(LZ_q1q2 * RZ_f5f6 / DZ * Aplus56231_2_VL[0] + LP_q1q2 * RP_f5f6 / DP * Aplus56231_2_P[0]) + (LZ_q1q2 * RZ_f5f6 / DZ_FSR + LP_q1q2 * RP_f5f6 / DP_FSR) * Aplus56231_2_FSR[0];
	Aplus65231_2_tot[1] = RZ_q1q2 * RZ_f5f6 / DZ * Aplus65231_2_VR[1] + RP_q1q2 * RP_f5f6 / DP * Aplus65231_2_P[1] + (RZ_q1q2 * RZ_f5f6 / DZ_FSR + RP_q1q2 * RP_f5f6 / DP_FSR) * Aplus65231_2_FSR[1];
	
	Aplus65132_2_tot[0] = RZ_q1q2 * LZ_f5f6 / DZ * Aplus65132_2_VR[0] + RP_q1q2 * LP_f5f6 / DP * Aplus65132_2_P[0] - (RZ_q1q2 * LZ_f5f6 / DZ_FSR + RP_q1q2 * LP_f5f6 / DP_FSR) * Aplus65132_2_FSR[0];
	Aplus56132_2_tot[1] = -(LZ_q1q2 * LZ_f5f6 / DZ * Aplus56132_2_VL[1] + LP_q1q2 * LP_f5f6 / DP * Aplus56132_2_P[1]) - (LZ_q1q2 * LZ_f5f6 / DZ_FSR + LP_q1q2 * LP_f5f6 / DP_FSR) * Aplus56132_2_FSR[1];
	Aplus65231_2_tot[0] = -(LZ_q1q2 * LZ_f5f6 / DZ * Aplus65231_2_VL[0] + LP_q1q2 * LP_f5f6 / DP * Aplus65231_2_P[0]) - (LZ_q1q2 * LZ_f5f6 / DZ_FSR + LP_q1q2 * LP_f5f6 / DP_FSR) * Aplus65231_2_FSR[0];
	Aplus56231_2_tot[1] = RZ_q1q2 * LZ_f5f6 / DZ * Aplus56231_2_VR[1] + RP_q1q2 * LP_f5f6 / DP * Aplus56231_2_P[1] - (RZ_q1q2 * LZ_f5f6 / DZ_FSR + RP_q1q2 * LP_f5f6 / DP_FSR) * Aplus56231_2_FSR[1];

        // one loop times one loop
        osi->QT_A2 += norm(Aplus56132_1_tot[0]);
        osi->QT_A2 += norm(Aplus65132_1_tot[1]);
        osi->QT_A2 += norm(Aplus56231_1_tot[0]);
        osi->QT_A2 += norm(Aplus65231_1_tot[1]);

        osi->QT_A2 += norm(Aplus65132_1_tot[0]);
        osi->QT_A2 += norm(Aplus56132_1_tot[1]);
        osi->QT_A2 += norm(Aplus65231_1_tot[0]);
        osi->QT_A2 += norm(Aplus56231_1_tot[1]);

        // two loop times Born
        osi->QT_A2 += 2 * real(Aplus56132_0_tot[0] * conj(Aplus56132_2_tot[0]));
        osi->QT_A2 += 2 * real(Aplus65132_0_tot[1] * conj(Aplus65132_2_tot[1]));
        osi->QT_A2 += 2 * real(Aplus56231_0_tot[0] * conj(Aplus56231_2_tot[0]));
        osi->QT_A2 += 2 * real(Aplus65231_0_tot[1] * conj(Aplus65231_2_tot[1]));

        osi->QT_A2 += 2 * real(Aplus65132_0_tot[0] * conj(Aplus65132_2_tot[0]));
        osi->QT_A2 += 2 * real(Aplus56132_0_tot[1] * conj(Aplus56132_2_tot[1]));
        osi->QT_A2 += 2 * real(Aplus65231_0_tot[0] * conj(Aplus65231_2_tot[0]));
        osi->QT_A2 += 2 * real(Aplus56231_0_tot[1] * conj(Aplus56231_2_tot[1]));
      }
    }
  } 
  else if (no_process_class == 2 || no_process_class == 3){ //  W gamma  processes
    // propagator denominators
    DW = s34 - msi->M2_W + ri * msi->Gamma_W * msi->M_W;
    // propagator denominators (FSR)
    DW_FSR = s12 - msi->M2_W + ri * msi->Gamma_W * msi->M_W;

    Aplus56132_0_tot[1] = -LW_q1q2 * LW_f5f6 * (1.0 / DW * Aplus56132_0[1] + 1.0 / DW_FSR * Aplus56132_0_FSR[1]);
    Aplus65231_0_tot[0] = -LW_q1q2 * LW_f5f6 * (1.0 / DW * Aplus65231_0[0] + 1.0 / DW_FSR * Aplus65231_0_FSR[0]);

    osi->QT_A0 += norm(Aplus56132_0_tot[1]);
    osi->QT_A0 += norm(Aplus65231_0_tot[0]);

    if (order >= 1){
      Aplus56132_1_tot[1] = -LW_q1q2 * LW_f5f6 * (1.0 / DW * Aplus56132_1[1] + 1.0 / DW_FSR * Aplus56132_1_FSR[1]);
      Aplus65231_1_tot[0] = -LW_q1q2 * LW_f5f6 * (1.0 / DW * Aplus65231_1[0] + 1.0 / DW_FSR * Aplus65231_1_FSR[0]);
      
      osi->QT_A1 += 2 * real(Aplus56132_0_tot[1] * conj(Aplus56132_1_tot[1]));
      osi->QT_A1 += 2 * real(Aplus65231_0_tot[0] * conj(Aplus65231_1_tot[0]));

      if (order == 2){
	Aplus56132_2_tot[1] = -LW_q1q2 * LW_f5f6 * (1.0 / DW * Aplus56132_2_VR[1] + 1.0 / DW_FSR * Aplus56132_2_FSR[1]);
        Aplus65231_2_tot[0] = -LW_q1q2 * LW_f5f6 * (1.0 / DW * Aplus65231_2_VR[0] + 1.0 / DW_FSR * Aplus65231_2_FSR[0]);

        // one loop times one loop
        osi->QT_A2 += norm(Aplus56132_1_tot[1]);
        osi->QT_A2 += norm(Aplus65231_1_tot[0]);

        // two loop times Born
        osi->QT_A2 += 2 * real( Aplus56132_0_tot[1] * conj(Aplus56132_2_tot[1]));
        osi->QT_A2 += 2 * real( Aplus65231_0_tot[0] * conj(Aplus65231_2_tot[0]));
      }
    }
  }
    
  static double factor_squared_amp_coupling_LO = pow(f4pi * msi->alpha_e, 3) / (4 * 9) * 3;
  // EW coupling // averaging factor // color factor
  osi->QT_A0 *= factor_squared_amp_coupling_LO;
  osi->QT_A1 *= factor_squared_amp_coupling_LO * (osi->alpha_S * inv2pi);
  osi->QT_A2 *= factor_squared_amp_coupling_LO * pow(osi->alpha_S * inv2pi, 2);
  

  if (order == 2){
    osi->QT_H2_delta_QCD = osi->QT_A2;
    /**/
    // shift here from  pplla23_calculate_H_coefficients !!!
    /////  double s_hat = 2 * esi->p_parton[0][1] * esi->p_parton[0][2];
    //  double s_hat = s12;
    /////    double LR = log(s34 / s12);
    //  F2 = F2_coefficient_1 + (LR - ri * pi) * F2_coefficient_LR;
    delta_I2 = delta_I2_coefficient_1 + LR * delta_I2_coefficient_LR;

    osi->QT_H2_delta_QCD += 2 * osi->QT_A1 * delta_I1 * osi->alpha_S * inv2pi;
    osi->QT_H2_delta_QCD += osi->QT_A0 * delta_I2 * pow(osi->alpha_S * inv2pi,2);
  }
  osi->QT_H1_delta_QCD = osi->QT_A1;

  // adapt I1 conventions (Lorenzo -> Massimiliano)
  osi->QT_A1 += 2 * osi->QT_A0 * delta_I1 * osi->alpha_S * inv2pi;
  osi->QT_H1_delta_QCD += 2 * osi->QT_A0 * delta_I1 * osi->alpha_S * inv2pi;

  
  // go to right scale
  // Should not always be q2 == s here ???
  // Apparently, s = s12 = (p1 + p2)^2, but q2 = (p3 + p4)^4 = m²_ll !!!
  //  osi->QT_H2_delta_QCD += beta0 * log(s / q2) * osi->QT_H1_delta_QCD * osi->alpha_S * inv2pi;
  osi->QT_H2_delta_QCD += beta0 * log(s12 / s34) * osi->QT_H1_delta_QCD * osi->alpha_S * inv2pi;

  // adapt alpha_S/Pi normalisation
  osi->QT_H1_delta_QCD /= 2;
  osi->QT_H2_delta_QCD /= 4;
  
  osi->QT_H1_delta_QCD /= (inv2pi * osi->alpha_S * osi->QT_A0);
  osi->QT_H2_delta_QCD /= (pow(inv2pi * osi->alpha_S, 2) * osi->QT_A0);

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


void pplla03_amplitude_set::calculate_H1gg_QCD_2loop(){
  static Logger logger("pplla03_amplitude_set::calculate_H1gg_QCD_2loop");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  //  gglla33_calculate_H1(osi);
  logger << LOG_FATAL << "Amplitudes have not been implemented yet." << endl;
  exit(1);
  
  osi->QT_H1_delta_QCD = osi->QT_A1;
  osi->QT_H1_delta_QCD /= 2;
  osi->QT_H1_delta_QCD /= (osi->alpha_S * inv2pi * osi->QT_A0);
  logger << LOG_DEBUG << "osi->QT_H1_delta_QCD = " << osi->QT_H1_delta_QCD << endl;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


#ifdef OPENLOOPS

pplla03_amplitude_OpenLoops_set::pplla03_amplitude_OpenLoops_set(){
  Logger logger("pplla03_amplitude_OpenLoops_set::pplla03_amplitude_OpenLoops_set ()");
  logger << LOG_DEBUG_VERBOSE << "started" << endl;

  logger << LOG_INFO << "constructor called..." << endl;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

pplla03_amplitude_OpenLoops_set::~pplla03_amplitude_OpenLoops_set(){
  Logger logger("pplla03_amplitude_OpenLoops_set::~pplla03_amplitude_OpenLoops_set ()");
  logger << LOG_DEBUG_VERBOSE << "started" << endl;

  logger << LOG_INFO << "destructor called..." << endl;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


void pplla03_amplitude_OpenLoops_set::initialization_process(){
  Logger logger("pplla03_amplitude_OpenLoops_set::initialization_process");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  if (csi->type_contribution == "VT2" && csi->type_correction == "QCD"){
    calculate_H2_QCD_2loop_initialization_ingredients();
  }
  amplitude_OpenLoops_set::initialization_process();
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

#endif



#ifdef RECOLA

pplla03_amplitude_Recola_set::pplla03_amplitude_Recola_set(){
  Logger logger("pplla03_amplitude_Recola_set::pplla03_amplitude_Recola_set ()");
  logger << LOG_DEBUG_VERBOSE << "started" << endl;

  logger << LOG_INFO << "constructor called..." << endl;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

pplla03_amplitude_Recola_set::~pplla03_amplitude_Recola_set(){
  Logger logger("pplla03_amplitude_Recola_set::~pplla03_amplitude_Recola_set ()");
  logger << LOG_DEBUG_VERBOSE << "started" << endl;

  logger << LOG_INFO << "destructor called..." << endl;

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


void pplla03_amplitude_Recola_set::initialization_process(){
  Logger logger("pplla03_amplitude_Recola_set::initialization_process");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  if (csi->type_contribution == "VT2" && csi->type_correction == "QCD"){
    calculate_H2_QCD_2loop_initialization_ingredients();
  }
  amplitude_Recola_set::initialization_process();
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

#endif
