#include "header.hpp"

#include "ppllll04.amplitude.set.hpp"

void ppllll04_amplitude_set::initialization_H2_QCD_2loop(){
  static Logger logger("ppllll04_amplitude_set::initialization_H2_QCD_2loop");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;

  q1 = 0;
  if (osi->csi->subprocess == "uu~_emmumepmup" ||
      osi->csi->subprocess == "uu~_ememepep" ||
      osi->csi->subprocess == "uu~_emmupvmve~" ||
      osi->csi->subprocess == "uu~_emepvmvm~" ||
      osi->csi->subprocess == "uu~_emepveve~" ||
      osi->csi->subprocess == "ud~_emepmupvm" ||
      osi->csi->subprocess == "ud~_emepepve") {
    q1 = 2;
  }
  else if (osi->csi->subprocess == "dd~_emmumepmup" ||
	   osi->csi->subprocess == "dd~_ememepep" ||
	   osi->csi->subprocess == "dd~_emmupvmve~" ||
	   osi->csi->subprocess == "dd~_emepvmvm~" ||
	   osi->csi->subprocess == "dd~_emepveve~" ||
	   osi->csi->subprocess == "bb~_emmumepmup" ||
	   osi->csi->subprocess == "bb~_ememepep" ||
	   osi->csi->subprocess == "bb~_emepvmvm~" ||
	   osi->csi->subprocess == "bb~_emmupvmve~" ||
	   osi->csi->subprocess == "bb~_emepveve~" ||
	   osi->csi->subprocess == "du~_emmumepvm~" ||
	   osi->csi->subprocess == "du~_ememepve~") {
    q1 = 1;
  }
  else {
    logger << LOG_WARN << osi->csi->subprocess << endl;
    assert(false);
  }

  M_dressed.resize(2, vector<vector<vector<double_complex> > > (2, vector<vector<double_complex> > (2,vector<double_complex> (3, 0.))));
  M_dressed2.resize(2, vector<vector<vector<double_complex> > > (2, vector<vector<double_complex> > (2,vector<double_complex> (3, 0.))));
  M_dressed_tmp.resize(2, vector<vector<vector<double_complex> > > (2, vector<vector<double_complex> > (2,vector<double_complex> (3, 0.))));

  p_VVamp.resize(7);
  za.resize(7, vector<double_complex> (7, 0.));
  zb.resize(7, vector<double_complex> (7, 0.));
  s_inv.resize(7, vector<double> (7, 0.));

  /*
  // sym_factor is the symmetry factor of the final state leptons - should be identical with the one in csi -> 1. / ... !!!
  sym_factor = 1;
  if (osi->csi->subprocess == "uu~_emmumepmup" ||
      osi->csi->subprocess == "dd~_emmumepmup" ||
      osi->csi->subprocess == "bb~_emmumepmup") {
    sym_factor = 1;
  }
  else if (osi->csi->subprocess == "uu~_ememepep" ||
	   osi->csi->subprocess == "dd~_ememepep" ||
	   osi->csi->subprocess == "bb~_ememepep") {
    sym_factor=1.0/4;
  }
  else if (osi->csi->subprocess == "uu~_emmupvmve~" ||
	   osi->csi->subprocess == "dd~_emmupvmve~" ||
	   osi->csi->subprocess == "bb~_emmupvmve~") {
    sym_factor = 1;
  }
  else if (osi->csi->subprocess == "uu~_emepvmvm~" ||
	   osi->csi->subprocess == "dd~_emepvmvm~" ||
	   osi->csi->subprocess == "bb~_emepvmvm~" ||
	   osi->csi->subprocess == "uu~_emepveve~" ||
	   osi->csi->subprocess == "dd~_emepveve~" ||
	   osi->csi->subprocess == "bb~_emepveve~") {
  }
  else if (osi->csi->subprocess == "ud~_emepmupvm") {
    sym_factor = 1;
  }
  else if (osi->csi->subprocess == "ud~_emepepve") {
    sym_factor = 1.0/2;
  }
  else if (osi->csi->subprocess == "du~_emmumepvm~") {
    sym_factor = 1;
  }
  else if (osi->csi->subprocess == "du~_ememepve~") {
    sym_factor = 1.0/2;
  }

  if (sym_factor != 1. / csi->symmetry_factor){
    logger << LOG_WARN << "sym_factor = " << sym_factor << endl;
    logger << LOG_WARN << "csi->symmetry_factor = " << csi->symmetry_factor << endl;
  }
  */
  sym_factor = 1. / csi->symmetry_factor;
  
  E.resize(3, vector<vector<double_complex> >(4, vector<double_complex> (9)));
  F.resize(3);

  if (osi->csi->subprocess == "uu~_emmumepmup" ||
      osi->csi->subprocess == "dd~_emmumepmup" ||
      osi->csi->subprocess == "bb~_emmumepmup"){
    permutation.resize(1);
    DR.resize(1);
    SR.resize(1);
    permutation[0] = {0,1,2,3,5,4,6};
    // ZZ contribution
    DR[0].push_back(ppllll04_decay_DR(V_ZLL, V_ZLL, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(1, V_ZLL, V_ZLL, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(2, V_ZLL, V_ZLL, this, permutation[0]));
    // gamgam contribution
    DR[0].push_back(ppllll04_decay_DR(V_PHOT, V_PHOT, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(1, V_PHOT, V_PHOT, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(2, V_PHOT, V_PHOT, this, permutation[0]));
    // gamZ contribution
    DR[0].push_back(ppllll04_decay_DR(V_PHOT, V_ZLL, this, permutation[0]));
    // Zgam contribution
    DR[0].push_back(ppllll04_decay_DR(V_ZLL, V_PHOT, this, permutation[0]));
  }
  else if (osi->csi->subprocess == "uu~_ememepep" ||
	   osi->csi->subprocess == "dd~_ememepep" ||
	   osi->csi->subprocess == "bb~_ememepep"){
    permutation.resize(2);
    DR.resize(2);
    SR.resize(2);
    permutation[0] = {0,1,2,3,5,4,6};
    // ZZ contribution
    DR[0].push_back(ppllll04_decay_DR(V_ZLL, V_ZLL, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(1, V_ZLL, V_ZLL, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(2, V_ZLL, V_ZLL, this, permutation[0]));
    // gamgam contribution
    DR[0].push_back(ppllll04_decay_DR(V_PHOT, V_PHOT, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(1, V_PHOT, V_PHOT, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(2, V_PHOT, V_PHOT, this, permutation[0]));
    // gamZ contribution
    DR[0].push_back(ppllll04_decay_DR(V_PHOT, V_ZLL, this, permutation[0]));
    // Zgam contribution
    DR[0].push_back(ppllll04_decay_DR(V_ZLL, V_PHOT, this, permutation[0]));
    // same flavour case: take into account additional crossings
    permutation[1] = {0,1,2,3,6,4,5};
    // ZZ contribution
    DR[1].push_back(ppllll04_decay_DR(V_ZLL, V_ZLL, this, permutation[1]));
    SR[1].push_back(ppllll04_decay_SR(1, V_ZLL, V_ZLL, this, permutation[1]));
    SR[1].push_back(ppllll04_decay_SR(2, V_ZLL, V_ZLL, this, permutation[1]));
    // gamgam contribution
    DR[1].push_back(ppllll04_decay_DR(V_PHOT, V_PHOT, this, permutation[1]));
    SR[1].push_back(ppllll04_decay_SR(1, V_PHOT, V_PHOT, this, permutation[1]));
    SR[1].push_back(ppllll04_decay_SR(2, V_PHOT, V_PHOT, this, permutation[1]));
    // gamZ contribution
    DR[1].push_back(ppllll04_decay_DR(V_PHOT, V_ZLL, this, permutation[1]));
    // Zgam contribution
    DR[1].push_back(ppllll04_decay_DR(V_ZLL, V_PHOT, this, permutation[1]));
  }
  else if (osi->csi->subprocess == "uu~_emmupvmve~" ||
	   osi->csi->subprocess == "dd~_emmupvmve~" ||
	   osi->csi->subprocess == "bb~_emmupvmve~"){
    permutation.resize(1);
    DR.resize(1);
    SR.resize(1);
    permutation[0] = {0,1,2,3,6,5,4};
    // WW contribution
    DR[0].push_back(ppllll04_decay_DR(V_WMIN, V_WPLU, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(1, V_WMIN, V_WPLU, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(2, V_WMIN, V_WPLU, this, permutation[0]));
  }
  else if (osi->csi->subprocess == "uu~_emepvmvm~" ||
	   osi->csi->subprocess == "dd~_emepvmvm~" ||
	   osi->csi->subprocess == "bb~_emepvmvm~"){
    permutation.resize(1);
    DR.resize(1);
    SR.resize(1);
    permutation[0] = {0,1,2,3,4,5,6};
    // ZZ contribution
    DR[0].push_back(ppllll04_decay_DR(V_ZLL, V_ZNN, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(1, V_ZLL, V_ZNN, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(2, V_ZLL, V_ZNN, this, permutation[0]));
    // gamZ contribution
    DR[0].push_back(ppllll04_decay_DR(V_PHOT, V_ZNN, this, permutation[0]));
  }
  else if (osi->csi->subprocess == "uu~_emepveve~" ||
	   osi->csi->subprocess == "dd~_emepveve~" ||
	   osi->csi->subprocess == "bb~_emepveve~"){
    permutation.resize(2);
    DR.resize(2);
    SR.resize(2);
    permutation[0] = {0,1,2,3,4,5,6};
    // ZZ contribution
    DR[0].push_back(ppllll04_decay_DR(V_ZLL, V_ZNN, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(1, V_ZLL, V_ZNN, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(2, V_ZLL, V_ZNN, this, permutation[0]));
    // gamZ contribution
    DR[0].push_back(ppllll04_decay_DR(V_PHOT, V_ZNN, this, permutation[0]));
    // same flavour case: take into account additional crossings
    permutation[1] = {0,1,2,3,6,5,4};
    // WW contribution
    DR[1].push_back(ppllll04_decay_DR(V_WMIN, V_WPLU, this, permutation[1]));
    SR[1].push_back(ppllll04_decay_SR(1, V_WMIN, V_WPLU, this, permutation[1]));
    SR[1].push_back(ppllll04_decay_SR(2, V_WMIN, V_WPLU, this, permutation[1]));
  }
  else if (osi->csi->subprocess == "ud~_emepmupvm"){
    permutation.resize(1);
    DR.resize(1);
    SR.resize(1);
    permutation[0] = {0,1,2,3,4,6,5};
    // ZW^+ contribution
    DR[0].push_back(ppllll04_decay_DR(V_ZLL, V_WPLU, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(1, V_ZLL, V_WPLU, this, permutation[0]));
    // this contribution only occurs once, as it only contains intermediate W's
    SR[0].push_back(ppllll04_decay_SR(2, V_ZLL, V_WPLU, this, permutation[0]));
    // gamW^+ contribution
    DR[0].push_back(ppllll04_decay_DR(V_PHOT, V_WPLU, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(1, V_PHOT, V_WPLU, this, permutation[0]));
  }
  else if (osi->csi->subprocess == "ud~_emepepve"){
    permutation.resize(2);
    DR.resize(2);
    SR.resize(2);
    permutation[0] = {0,1,2,3,4,6,5};
    // ZW^+ contribution
    DR[0].push_back(ppllll04_decay_DR(V_ZLL, V_WPLU, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(1, V_ZLL, V_WPLU, this, permutation[0]));
    // this contribution only occurs once, as it only contains intermediate W's
    SR[0].push_back(ppllll04_decay_SR(2, V_ZLL, V_WPLU, this, permutation[0]));
    // gamW^+ contribution
    DR[0].push_back(ppllll04_decay_DR(V_PHOT, V_WPLU, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(1, V_PHOT, V_WPLU, this, permutation[0]));
    // same flavour case: take into account additional crossings
    permutation[1] = {0,1,2,3,5,6,4};
    // ZW^+ contribution
    DR[1].push_back(ppllll04_decay_DR(V_ZLL, V_WPLU, this, permutation[1]));
    SR[1].push_back(ppllll04_decay_SR(1, V_ZLL, V_WPLU, this, permutation[1]));
    // this contribution only occurs once, as it only contains intermediate W's
    SR[1].push_back(ppllll04_decay_SR(2, V_ZLL, V_WPLU, this, permutation[1]));
    // gamW^+ contribution
    DR[1].push_back(ppllll04_decay_DR(V_PHOT, V_WPLU, this, permutation[1]));
    SR[1].push_back(ppllll04_decay_SR(1, V_PHOT, V_WPLU, this, permutation[1]));
  }
  else if (osi->csi->subprocess == "du~_emmumepvm~"){
    permutation.resize(1);
    DR.resize(1);
    SR.resize(1);
    permutation[0] = {0,1,2,3,5,4,6};
    // ZW^- contribution
    DR[0].push_back(ppllll04_decay_DR(V_ZLL, V_WMIN, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(1, V_ZLL, V_WMIN, this, permutation[0]));
    // this contribution only occurs once, as it only contains intermediate W's
    SR[0].push_back(ppllll04_decay_SR(2, V_ZLL, V_WMIN, this, permutation[0]));
    // gamW^- contribution
    DR[0].push_back(ppllll04_decay_DR(V_PHOT, V_WMIN, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(1, V_PHOT, V_WMIN, this, permutation[0]));
  }
  else if (osi->csi->subprocess == "du~_ememepve~"){
    permutation.resize(2);
    DR.resize(2);
    SR.resize(2);
    permutation[0] = {0,1,2,3,5,4,6};
    // ZW^- contribution
    DR[0].push_back(ppllll04_decay_DR(V_ZLL, V_WMIN, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(1, V_ZLL, V_WMIN, this, permutation[0]));
    // this contribution only occurs once, as it only contains intermediate W's
    SR[0].push_back(ppllll04_decay_SR(2, V_ZLL, V_WMIN, this, permutation[0]));
    // gamW^- contribution
    DR[0].push_back(ppllll04_decay_DR(V_PHOT, V_WMIN, this, permutation[0]));
    SR[0].push_back(ppllll04_decay_SR(1, V_PHOT, V_WMIN, this, permutation[0]));
    // same flavour case: take into account additional crossings
    permutation[1] = {0,1,2,4,5,3,6};
    // ZW^- contribution
    DR[1].push_back(ppllll04_decay_DR(V_ZLL, V_WMIN, this, permutation[1]));
    SR[1].push_back(ppllll04_decay_SR(1, V_ZLL, V_WMIN, this, permutation[1]));
    // this contribution only occurs once, as it only contains intermediate W's
    SR[1].push_back(ppllll04_decay_SR(2, V_ZLL, V_WMIN, this, permutation[1]));
    // gamW^- contribution
    DR[1].push_back(ppllll04_decay_DR(V_PHOT, V_WMIN, this, permutation[1]));
    SR[1].push_back(ppllll04_decay_SR(1, V_PHOT, V_WMIN, this, permutation[1]));
  }
  else {
    assert(false);
  }

  VVprime::getInstance().initAmplitude(osi->N_f);

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


void ppllll04_amplitude_set::add_hel_amplitudes_H2_QCD_2loop(int i_x){
  static Logger logger("ppllll04_amplitude_set::add_hel_amplitudes_H2_QCD_2loop");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;
  
  for (int h1 = 0; h1 < 2; h1++){
    for (int h2 = 0; h2 < 2; h2++){
      for (int h3 = 0; h3 < 2; h3++){
        for (int loop = 0; loop < 3; loop++){
          if (i_x == 0){M_dressed[h1][h2][h3][loop] += M_dressed_tmp[h1][h2][h3][loop];}
	  else {M_dressed2[h1][h2][h3][loop] += M_dressed_tmp[h1][h2][h3][loop];}
        }
      }
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}

void ppllll04_amplitude_set::reset_hel_amplitudes_H2_QCD_2loop(int i_x){
  static Logger logger("ppllll04_amplitude_set::reset_hel_amplitudes_H2_QCD_2loop");
  logger << LOG_DEBUG_VERBOSE << "called" << endl;
  
  for (int h1 = 0; h1 < 2; h1++){
    for (int h2 = 0; h2 < 2; h2++){
      for (int h3 = 0; h3 < 2; h3++){
	for (int loop = 0; loop < 3; loop++){
	  if (i_x == 0){M_dressed[h1][h2][h3][loop] = 0.;}
	  else {M_dressed2[h1][h2][h3][loop] = 0.;}
	}
      }
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endl;
}


// Needs to be checked !!!
void ppllll24_computeE(double s, double t, double ma2, double mb2, int type, vector<double_complex> &E) {
  // tree level form factors (for debugging purposes)
  double u = ma2+mb2-s-t;

  if (type==0) {
    E[0] = 0;
    E[1] = -2.0/s/t;
    E[2] = 2.0/s/t;
    E[3] = 0;
    E[4] = 2.0/t;
    E[5] = -2.0*(s-t+mb2)/s/t;
    E[6] = 2.0*(t-ma2)/s/t;
    E[7] = -2.0*(t-mb2)/s/t;
    E[8] = 2.0*(s-t+ma2)/s/t;
  } else if (type==1) {
    E[0] = 0;
    E[1] = -2.0/s/u;
    E[2] = 2.0/s/u;
    E[3] = 0;
    E[4] = -2.0/u;
    E[5] = -2.0*(u-ma2)/s/u;
    E[6] = 2.0*(s-u+mb2)/s/u;
    E[7] = -2.0*(s-u+ma2)/s/u;
    E[8] = 2.0*(u-mb2)/s/u;
  } else if (type==2) {
    E[0] = 0;
    E[1] = 0;
    E[2] = 0;
    E[3] = 0;
    E[4] = 0;
    E[5] = 0;
    E[6] = 0;
    E[7] = 0;
    E[8] = 0;
  } else if (type==3) {
    E[0] = 0;
    E[1] = 0;
    E[2] = 0;
    E[3] = 0;
    E[4] = -4;
    E[5] = 4;
    E[6] = 4;
    E[7] = -4;
    E[8] = -4;
  }
}
