#include "header.hpp"

void observable_set::determination_collinear_splitting(){
  Logger logger("observable_set::selection_collinear_splitting");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  CA_dipole_splitting.resize(4, vector<int> (3, 0));
  for (int i_c = 0; i_c < (*CA_collinear).size(); i_c++){
    for (int i_em = 0; i_em < 3; i_em++){
      if ((*CA_collinear)[i_c][0].in_collinear()[i_em] == 1){
	CA_dipole_splitting[(*CA_collinear)[i_c][0].type()][i_em] = 1;
      }
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::selection_collinear(){
  Logger logger("observable_set::selection_collinear");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (int i_c = (*CA_collinear).size() - 1; i_c >=0; i_c--){
    for (int j_c = (*CA_collinear)[i_c].size() - 1; j_c >= 0; j_c--){
      logger << LOG_DEBUG << "(*CA_collinear)[" << i_c << "][" << j_c << "].no_emitter() = " << (*CA_collinear)[i_c][j_c].no_emitter() << "   no_spectator() = " << (*CA_collinear)[i_c][j_c].no_spectator() << endline_flush;
      // Possibly: switch_off_CS_ii && switch_off_CS_if && (*CA_collinear)[i_c][j_c].no_spectator() == 0
      if ((*CA_collinear)[i_c][j_c].no_spectator() == 0){continue;}
      // remove final-final contribution:
      if (switch_off_CS_ff && ((*CA_collinear)[i_c][j_c].no_emitter() > 2 && (*CA_collinear)[i_c][j_c].no_spectator() > 2)){
	(*CA_collinear)[i_c].erase((*CA_collinear)[i_c].begin() + j_c);
	logger << LOG_WARN << "SHOULD NOT HAPPEN !!!   kill ff:   (*CA_collinear)[" << i_c << "][" << j_c << "].name() = " << (*CA_collinear)[i_c][j_c].name() << endline_flush;
	continue;
      }
      // remove final-initial contribution:
      if (switch_off_CS_fi && ((*CA_collinear)[i_c][j_c].no_emitter() > 2 && (*CA_collinear)[i_c][j_c].no_spectator() < 3)){
	(*CA_collinear)[i_c].erase((*CA_collinear)[i_c].begin() + j_c);
	logger << LOG_WARN << "SHOULD NOT HAPPEN !!!   kill fi:   (*CA_collinear)[" << i_c << "][" << j_c << "].name() = " << (*CA_collinear)[i_c][j_c].name() << endline_flush;
	continue;
      }
      // remove initial-final & final-initial contribution:
      if (switch_off_CS_if && switch_off_CS_fi && ((*CA_collinear)[i_c][j_c].no_emitter() < 3 && (*CA_collinear)[i_c][j_c].no_spectator() > 2)){
	(*CA_collinear)[i_c].erase((*CA_collinear)[i_c].begin() + j_c);
	logger << LOG_DEBUG << "kill if & fi:   (*CA_collinear)[" << i_c << "][" << j_c << "].name() = " << (*CA_collinear)[i_c][j_c].name() << endline_flush;
	continue;
      }
      // remove initial-initial contribution:
      if (switch_off_CS_ii && ((*CA_collinear)[i_c][j_c].no_emitter() < 3 && (*CA_collinear)[i_c][j_c].no_spectator() < 3)){
	(*CA_collinear)[i_c].erase((*CA_collinear)[i_c].begin() + j_c);
	logger << LOG_DEBUG << "kill ii :   (*CA_collinear)[" << i_c << "][" << j_c << "].name() = " << (*CA_collinear)[i_c][j_c].name() << endline_flush;
	continue;
      }
    }
    if ((*CA_collinear)[i_c].size() == 0){
      (*CA_collinear).erase((*CA_collinear).begin() + i_c);
      CA_Q2f.erase(CA_Q2f.begin() + i_c);
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}



void observable_set::fill_CA_combination_pdf(){
  Logger logger("observable_set::fill_CA_combination_pdf");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  CA_combination_pdf.resize((*CA_collinear).size(), vector<vector<vector<int> > > (csi->combination_pdf.size()));
  for (int i_c = 0; i_c < (*CA_collinear).size(); i_c++){
    for (int i_i = 0; i_i < CA_combination_pdf[i_c].size(); i_i++){
      // CA_combination_pdf[i_c] contains a vector, which contains the usual csi->combination_pdf(3) with 0 -> direction (1, -1), 1 -> parton with x1 (in hadron 1/2 for +1/-1), 2 -> parton with x2 (in hadron 2/1 for +1/-1)
      if (((*CA_collinear)[i_c][0].all_pdf()[i_i][1] == 10) && ((*CA_collinear)[i_c][0].all_pdf()[i_i][2] == 10)){
	for (int i_q = -N_f_active; i_q < N_f_active + 1; i_q++){
	  if (i_q == 0){continue;}
	  for (int j_q = -N_f_active; j_q < N_f_active + 1; j_q++){
	    if (j_q == 0){continue;}
	    vector<int> new_temp_combination_pdf = (*CA_collinear)[i_c][0].all_pdf()[i_i];
	    new_temp_combination_pdf[1] = i_q;
	    new_temp_combination_pdf[2] = j_q;
	    CA_combination_pdf[i_c][i_i].push_back(new_temp_combination_pdf);
	  }
	}
      }
      else if ((*CA_collinear)[i_c][0].all_pdf()[i_i][1] == 10){
	for (int i_q = -N_f_active; i_q < N_f_active + 1; i_q++){
	  if (i_q == 0){continue;}
	  vector<int> new_temp_combination_pdf = (*CA_collinear)[i_c][0].all_pdf()[i_i];
	  new_temp_combination_pdf[1] = i_q;
	  CA_combination_pdf[i_c][i_i].push_back(new_temp_combination_pdf);
	}
      }
      else if ((*CA_collinear)[i_c][0].all_pdf()[i_i][2] == 10){
	for (int j_q = -N_f_active; j_q < N_f_active + 1; j_q++){
	  if (j_q == 0){continue;}
	  vector<int> new_temp_combination_pdf = (*CA_collinear)[i_c][0].all_pdf()[i_i];
	  new_temp_combination_pdf[2] = j_q;
	  CA_combination_pdf[i_c][i_i].push_back(new_temp_combination_pdf);
	}
      }
      else {
	CA_combination_pdf[i_c][i_i].push_back((*CA_collinear)[i_c][0].all_pdf()[i_i]);
      }
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::determine_CA_list_combination_pdf(){
  Logger logger("observable_set::determine_CA_list_combination_pdf");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  // direct determination of list_combination_pdf from ncollinear (including list_combination_pdf_emission):
  // - Put no-emission/endpoint contribution as 0 component !!!
  // - Introduce 'no_pdf' variable in collinear_set to refer to PDF list.
  
  list_combination_pdf.resize((*CA_collinear).size() + 1, vector<vector<vector<int> > > (csi->combination_pdf.size()));
  list_combination_pdf_emission.resize((*CA_collinear).size() + 1, vector<int> (3, 0));

  // Extra entry for endpoint contribution:
  list_combination_pdf[0].resize(csi->combination_pdf.size(), vector<vector<int> > (1));
  for (int i_i = 0; i_i < csi->combination_pdf.size(); i_i++){list_combination_pdf[0][i_i].push_back(csi->combination_pdf[i_i]);}
  //  list_combination_pdf_emission[0] = vector<int> (3, 0);
  
  // Alternative: Introduce a (basically empty 0-component for CA_collinear) - to avoid the +1 shift !!!
  for (int i_c = 0; i_c < (*CA_collinear).size(); i_c++){
    list_combination_pdf_emission[i_c + 1][(*CA_collinear)[i_c][0].no_emitter()] = 1;
    //    list_combination_pdf_emission[i_c + 1] = (*CA_collinear)[i_c][0].in_collinear();
    //    ncollinear[i_c].no_pdf = i_c;
    for (int i_i = 0; i_i < list_combination_pdf[i_c + 1].size(); i_i++){
      // list_combination_pdf[i_c + 1] contains a vector, which contains the usual osi_csi->combination_pdf(3) with 0 -> direction (1, -1), 1 -> parton with x1 (in hadron 1/2 for +1/-1), 2 -> parton with x2 (in hadron 2/1 for +1/-1)
      if (((*CA_collinear)[i_c][0].all_pdf()[i_i][1] == 10) && ((*CA_collinear)[i_c][0].all_pdf()[i_i][2] == 10)){
	for (int i_q = -N_f_active; i_q < N_f_active + 1; i_q++){
	  if (i_q == 0){continue;}
	  for (int j_q = -N_f_active; j_q < N_f_active + 1; j_q++){
	    if (j_q == 0){continue;}
	    vector<int> new_temp_combination_pdf = (*CA_collinear)[i_c][0].all_pdf()[i_i];
	    new_temp_combination_pdf[1] = i_q;
	    new_temp_combination_pdf[2] = j_q;
	    list_combination_pdf[i_c + 1][i_i].push_back(new_temp_combination_pdf);
	  }
	}
      }
      else if ((*CA_collinear)[i_c][0].all_pdf()[i_i][1] == 10){
	for (int i_q = -N_f_active; i_q < N_f_active + 1; i_q++){
	  if (i_q == 0){continue;}
	  vector<int> new_temp_combination_pdf = (*CA_collinear)[i_c][0].all_pdf()[i_i];
	  new_temp_combination_pdf[1] = i_q;
	  list_combination_pdf[i_c + 1][i_i].push_back(new_temp_combination_pdf);
	}
      }
      else if ((*CA_collinear)[i_c][0].all_pdf()[i_i][2] == 10){
	for (int j_q = -N_f_active; j_q < N_f_active + 1; j_q++){
	  if (j_q == 0){continue;}
	  vector<int> new_temp_combination_pdf = (*CA_collinear)[i_c][0].all_pdf()[i_i];
	  new_temp_combination_pdf[2] = j_q;
	  list_combination_pdf[i_c + 1][i_i].push_back(new_temp_combination_pdf);
	}
      }
      else {
	list_combination_pdf[i_c + 1][i_i].push_back((*CA_collinear)[i_c][0].all_pdf()[i_i]);
      }
    }

    for (int j_c = 0; j_c < (*CA_collinear)[i_c].size(); j_c++){(*CA_collinear)[i_c][j_c].no_pdf = i_c + 1;}
  }

  for (int i_c = 0; i_c < (*CA_collinear).size(); i_c++){
    for (int j_c = 0; j_c < (*CA_collinear)[i_c].size(); j_c++){
      logger << LOG_DEBUG << "(*CA_collinear)[" << i_c << "][" << j_c << "].no_pdf = " <<(*CA_collinear)[i_c][j_c].no_pdf << endline_flush;
    }
  }
	

  
  /*
  // Extra entry for endpoint contribution:
  int flag = false;
  for (int i_c = 0; i_c < (*CA_collinear).size(); i_c++){if ((*CA_collinear)[i_c][0].in_collinear()[0] == 1){flag = true; break;}}
  if (flag){
    //      list_combination_pdf.resize((*CA_collinear).size(), vector<vector<vector<int> > > (csi->combination_pdf.size()));
    //      list_combination_pdf_emission.resize((*CA_collinear).size(), vector<int> (3, 0));
    list_combination_pdf.push_back(vector<vector<vector<int> > > (csi->combination_pdf.size(), vector<vector<int> > (1)));
    for (int i_i = 0; i_i < csi->combination_pdf.size(); i_i++){list_combination_pdf[list_combination_pdf.size() - 1][i_i][0] = csi->combination_pdf[i_i];}
    list_combination_pdf_emission.push_back(vector<int> (3, 0));
  }
  */
  
  logger << LOG_DEBUG << "list_combination_pdf determined " << endl << endline_flush;


logger << LOG_INFO << "(output_list_combination_pdf)" << endline << output_list_combination_pdf() << endline_flush;
/*
  logger.newLine(LOG_INFO);
  logger << LOG_INFO << "pdf contributions:   list_combination_pdf, list_combination_pdf_emission" << endline_flush;
  logger.newLine(LOG_INFO);
  for (int i_l = 0; i_l < list_combination_pdf.size(); i_l++){
    stringstream temp_em;
    for (int i_e = 1; i_e < 3; i_e++){
      temp_em << "   ";
      if (list_combination_pdf_emission[i_l][i_e] == 1){temp_em << "x_" << i_e << " / z_" << i_e;}
      else if (list_combination_pdf_emission[i_l][i_e] == 0){temp_em << "x_" << i_e;}
      else {logger << LOG_FATAL << temp_em.str() << " is not a valid value." << endline_flush; exit(1);}
    }
    logger << LOG_INFO << "pdf contributions [" << setw(2) << i_l << "]   @" << temp_em.str() << endline_flush;// << setw(15) << left << ncollinear[i_c].all_name[i_i] << ":   " << endline_flush;

    for (int i_i = 0; i_i < list_combination_pdf[i_l].size(); i_i++){
       for (int i_z = 0; i_z < list_combination_pdf[i_l][i_i].size(); i_z++){
	stringstream temp;
	temp.str("");
	temp << setw(32) << "";
	temp << "list_combination_pdf[" << setw(2) << i_l << "][" << setw(2) << i_i << "][" << setw(2) << i_z << "] = ";
	for (int i_y = 0; i_y < list_combination_pdf[i_l][i_i][i_z].size(); i_y++){temp << setw(2) << right << list_combination_pdf[i_l][i_i][i_z][i_y] << "  ";}
	logger << LOG_INFO << temp.str() << endline_flush;
      }
    }
  }
  logger.newLine(LOG_INFO);
*/
 
  //  output_CX_ncollinear_QCD();

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

