#include "header.hpp"

void observable_set::header_integration(ofstream & out_integration){
  Logger logger("observable_set::header_integration");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  out_integration << setprecision(15);
  out_integration << left << setw(30) << "process" << " = " << csi->process_class << endline;
  if (csi->process_class != csi->basic_process_class){
    out_integration << left << setw(30) << "basic process" << " = " << csi->basic_process_class << endline;
  }
  if (csi->coll_choice == 0){out_integration << setw(30) << "collision type" << " = " << "partonic" << endline;}
  else if (csi->coll_choice == 1){out_integration << setw(30) << "collision type" << " = " << "proton-proton" << endline;}
  else if (csi->coll_choice == 2){out_integration << setw(30) << "collision type" << " = " << "proton-antiproton" << endline;}
  else if (csi->coll_choice == 3){out_integration << setw(30) << "collision type" << " = " << "antiproton-antiproton" << endline;}
  out_integration << left << setw(30) << "beam energy" << " = " << psi->E << " GeV" << endline;
  out_integration << left << setw(30) << "partonic subprocess" << " = " << csi->subprocess << endline;

  out_integration << endline;
  out_integration << "Particle masses: " << endline;
  out_integration << left << setw(30) << "M_W" << " = " << msi->M_W << " GeV" << endline;
  out_integration << left << setw(30) << "M_Z" << " = " << msi->M_Z << " GeV" << endline;
  out_integration << left << setw(30) << "M_H" << " = " << msi->M_H << " GeV" << endline;

  out_integration << left << setw(30) << "M_d" << " = " << msi->M_d << " GeV" << endline;
  out_integration << left << setw(30) << "M_u" << " = " << msi->M_u << " GeV" << endline;
  out_integration << left << setw(30) << "M_s" << " = " << msi->M_s << " GeV" << endline;
  out_integration << left << setw(30) << "M_c" << " = " << msi->M_c << " GeV" << endline;
  out_integration << left << setw(30) << "M_b" << " = " << msi->M_b << " GeV" << endline;
  out_integration << left << setw(30) << "M_t" << " = " << msi->M_t << " GeV" << endline;
  out_integration << endline;
  out_integration << "Particle widths: " << endline;
  out_integration << left << setw(30) << "Gamma_W" << " = " << msi->Gamma_W << " GeV" << endline;
  out_integration << left << setw(30) << "Gamma_Z" << " = " << msi->Gamma_Z << " GeV" << endline;
  out_integration << left << setw(30) << "Gamma_H" << " = " << msi->Gamma_H << " GeV" << endline;
  out_integration << left << setw(30) << "Gamma_t" << " = " << msi->Gamma_t << " GeV" << endline;
  out_integration << endline;
  out_integration << "Branching ratios: " << endline;
  out_integration << left << setw(30) << "BR_Wlv" << " = " << msi->BR_Wlv << endline;
  out_integration << left << setw(30) << "BR_Zll" << " = " << msi->BR_Zll << endline;
  out_integration << left << setw(30) << "BR_tWb" << " = " << msi->BR_tWb << endline;
  out_integration << endline;
  out_integration << "Electroweak Standard Model parameters: " << endline;
  if (msi->ew_scheme == 0 || msi->use_adapted_ew_coupling == 0){
    out_integration << left << setw(30) << "1 / alpha_e_0" << " = " << setw(30) << 1 / msi->alpha_e_0 << setw(5) << setw(30) << "alpha_e_0" << " = " << msi->alpha_e_0 << endline;
  }
  if (msi->ew_scheme == 1 || msi->ew_scheme == -1 || msi->use_adapted_ew_coupling == 1){
    out_integration << left << setw(30) << "G_F" << " = " << msi->G_F << endline;
    out_integration << left << setw(30) << "1 / alpha_e_Gmu" << " = " << setw(30) << 1 / msi->alpha_e_Gmu << setw(5) << setw(30) << "alpha_e_Gmu" << " = " << msi->alpha_e_Gmu << endline;
  }
  if (msi->ew_scheme == 2 || msi->use_adapted_ew_coupling == 2){
    out_integration << left << setw(30) << "1 / alpha_e_MZ" << " = " << setw(30) << 1 / msi->alpha_e_MZ << setw(5) << setw(30) << "alpha_e_MZ" << " = " << msi->alpha_e_MZ << endline;
  }
  out_integration << endline;

  out_integration << left << setw(30) << "ew_scheme" << " = " << msi->ew_scheme;
  if (msi->ew_scheme == 0){out_integration << setw(24) << " (alpha(0) scheme)" << "->   alpha_e = alpha_e_0 = " << msi->alpha_e << endline;}
  else if (msi->ew_scheme == 1 || msi->ew_scheme == -1){out_integration << setw(24) << " (G_mu scheme)" << "->   alpha_e = alpha_e_Gmu = " << msi->alpha_e << endline;}
  else if (msi->ew_scheme == 2){out_integration << setw(24) << " (alpha(M_Z) scheme)" << "->   alpha_e = alpha_e_MZ = " << msi->alpha_e << endline;}
  else {out_integration << " (No allowed scheme specified!)" << endline; exit(1);}
  //  out_integration << left << setw(30) << "1 / alpha_e" << " = " << setw(30) << 1 / msi->alpha_e << setw(5) << setw(30) << "alpha_e" << " = " << msi->alpha_e << endline;
  if (msi->use_adapted_ew_coupling != -1){
    out_integration << left << setw(30) << "use_adapted_ew_coupling" << " = " << msi->use_adapted_ew_coupling << "   ->   ";
    if (msi->use_adapted_ew_coupling == 0){
      if (msi->ew_scheme == 1 || msi->ew_scheme == -1){
	out_integration << "Amplitudes are rescaled by a factor of (alpha_e_0 / alpha_e_Gmu) ^ " << csi->n_photon_born << " ." << endline;
      }
      else if (msi->ew_scheme == 2){
	out_integration << "Amplitudes are rescaled by a factor of (alpha_e_0 / alpha_e_MZ) ^ " << csi->n_photon_born << " ." << endline;
      }
      else {exit(1);}
    }
    else {
      int temp_exp = 0;
      if (csi->type_correction == "QEW" || csi->type_correction == "MIX"){temp_exp = csi->contribution_order_alpha_e - 1 - csi->n_photon_born;}
      else {temp_exp = csi->contribution_order_alpha_e - csi->n_photon_born;}
      if (msi->use_adapted_ew_coupling == 1){
	out_integration << "Amplitudes are rescaled by a factor of (alpha_e_Gmu / alpha_e_0) ^ " << temp_exp << " ." << endline;
      }
      else if (msi->use_adapted_ew_coupling == 2){
	out_integration << "Amplitudes are rescaled by a factor of (alpha_e_MZ / alpha_e_0) ^ " << temp_exp << " ." << endline;

      }
    }
  }
  //  out_integration << left << setw(30) << "use_adapted_ew_coupling" << " = " << msi->use_adapted_ew_coupling << endline;

  out_integration << left << setw(30) << "e" << " = " << msi->e_pow[1] << endline;
  out_integration << left << setw(30) << "cos_w" << " = " << msi->cos_w << endline;
  out_integration << left << setw(30) << "sin_w" << " = " << msi->sin_w << endline;
  out_integration << left << setw(30) << "cos2_w" << " = " << msi->cos2_w << endline;
  out_integration << left << setw(30) << "sin2_w" << " = " << msi->sin2_w << endline;
  //  out_integration << endline;

  /*
  if (ckm_choice == 0){
    out_integration << left << setw(30) << "V_ckm" << " = " << "trivial" << endline;
  }
  else {
    for (int i1 = 1; i1 < 4; i1++){
      if (i1 == 2){out_integration << left << setw(30) << "V_ckm" << " = ";}
      else {out_integration << left << setw(30) << "";}
      for (int i2 = 1; i2 < 4; i2++){
	out_integration << setprecision(8) << setw(20) << msi->V_ckm[i1][i2];
      }
      out_integration << endline;
    }
  }
  */
  out_integration << endline;

  out_integration << "QCD parameters: " << endline;
  if (csi->coll_choice != 0){
   out_integration << left << setw(30) << "LHAPDF set" << " = "  << LHAPDFname << endline;
   out_integration << left << setw(30) << "LHAPDF subset" << " = "  << LHAPDFsubset << endline;
   out_integration << left << setw(30) << "factorization scale" << " = " << scale_fact << " GeV" << endline;
   out_integration << endline;
  }
  out_integration << left << setw(30) << "renormalization scale" << " = " << scale_ren << " GeV" << endline;
  out_integration << endline;
  out_integration << left << setw(30) << "N_f" << " = " << N_f << endline;
  out_integration << left << setw(30) << "N_f_active" << " = " << N_f_active << endline;
  stringstream temp_alpha_S;
  temp_alpha_S << "alpha_S(mu" << " = " << setprecision(6) << mu_ren[1] << ")";
  out_integration << left << setw(30) << temp_alpha_S.str() << " = " << setprecision(15) << alpha_S << endline;
  stringstream temp_g_S;
  temp_g_S << "g_S(mu" << " = " << setprecision(6) << mu_ren[1] << ")";
  out_integration << left << setw(30) << temp_g_S.str() << " = " << setprecision(15) << g_from_alpha(alpha_S) << endline;
  out_integration << endline;
  out_integration << "Definitions for event selection: " << endline;
  out_integration << setw(5) << "no" << setw(20) << "object" << setw(20) << "n_observed_min" << setw(20) << "n_observed_max" << setw(20) << "define_pT_min" << setw(20) << "define_ET_min" << setw(20) << "define_|eta|_max"  << setw(20) << "define_|y|_max"<< endline;
  for (int i = 1; i < esi->object_list.size(); i++){
    out_integration << setw(5) << i << setw(20) << esi->object_list[i] << setw(20) << esi->pda[i].n_observed_min << setw(20) << esi->pda[i].n_observed_max << setw(20) << esi->pda[i].define_pT << setw(20) << esi->pda[i].define_ET << setw(20) << esi->pda[i].define_eta << setw(20) << esi->pda[i].define_y << endline;
  }
  out_integration << endline;
  out_integration << "Relevant object definitions for event selection: -> oset" << endline;
  out_integration << setw(5) << "no" << setw(20) << "object" << setw(20) << "n_observed_min" << setw(20) << "n_observed_max" << setw(20) << "define_pT_min" << setw(20) << "define_ET_min" << setw(20) << "define_|eta|_max"  << setw(20) << "define_|y|_max"<< endline;
  for (int i = 1; i < (esi->object_list_selection).size(); i++){
    out_integration << setw(5) << i << setw(20) << esi->object_list_selection[i] << setw(20) << esi->pds[i].n_observed_min << setw(20) << esi->pds[i].n_observed_max << setw(20) << esi->pds[i].define_pT << setw(20) << esi->pds[i].define_ET << setw(20) << esi->pds[i].define_eta << setw(20) << esi->pds[i].define_y << endline;
  }
  out_integration << endline;

  out_integration << csi->output_parton_content() <<endline;
  out_integration << endline;

  out_integration << "Additional cuts: " << endline;
  out_integration << endline;
  out_integration << "Integration parameters: " << endline;
  out_integration << setw(30) << "n_events_max" << " = " << psi->n_events_max << endline;
  out_integration << setw(30) << "n_events_min" << " = " << psi->n_events_min << endline;
  out_integration << setw(30) << "n_step" << " = " << psi->n_step << endline;
  out_integration << endline;
  out_integration << setw(30) << "mass0" << " = " << psi->mass0 << endline;
  out_integration << setw(30) << "nuxs" << " = " << psi->nuxs << endline;
  out_integration << setw(30) << "nuxt" << " = " << psi->nuxt << endline;
  out_integration << setw(30) << "cut_technical" << " = " << psi->cut_technical << endline;
  out_integration << setw(30) << "exp_pdf" << " = " << psi->exp_pdf << endline;
  if (csi->class_contribution_CS_real){
    out_integration << setw(30) << "exp_ij_k_y" << " = " << psi->exp_ij_k_y << endline;
    out_integration << setw(30) << "exp_ij_k_z" << " = " << psi->exp_ij_k_z << endline;
    out_integration << setw(30) << "exp_ij_a_x" << " = " << psi->exp_ij_a_x << endline;
    out_integration << setw(30) << "exp_ij_a_z" << " = " << psi->exp_ij_a_z << endline;
    out_integration << setw(30) << "exp_ai_k_x" << " = " << psi->exp_ai_k_x << endline;
    out_integration << setw(30) << "exp_ai_k_u" << " = " << psi->exp_ai_k_u << endline;
    out_integration << setw(30) << "exp_ai_b_x" << " = " << psi->exp_ai_b_x << endline;
    out_integration << setw(30) << "exp_ai_b_v" << " = " << psi->exp_ai_b_v << endline;
  }
  out_integration << endline;
  out_integration << endline;

  for (int i_m = 0; i_m < psi->input_MC.size(); i_m++){
    out_integration << setw(16) << psi->input_MC[i_m]->name << "   "
		    << right << setw(2) << psi->input_MC[i_m]->switch_optimization << "   "
		    << setw(3) << psi->input_MC[i_m]->n_optimization_step << " steps" << "   "
		    << setw(7) << psi->input_MC[i_m]->n_event_per_step << " events" << "   "
		    << setw(5) << psi->input_MC[i_m]->n_channel << " channels" << "   "
		    << "min: " << setw(1) << psi->input_MC[i_m]->switch_minimum_weight << "   "
		    << "(" << left << setw(5) << psi->input_MC[i_m]->reserved_minimum_weight << " reserved)" << "   "
		    << "max: " << setw(1) << psi->input_MC[i_m]->alpha_maximum_weight << "   "
		    << "mode " << setw(1) << psi->input_MC[i_m]->switch_mode_optimization << "   "
		    << "validation " << setw(1) << psi->input_MC[i_m]->switch_validation_optimization << "   "
		    << "adaptation exponent " << setw(5) << psi->input_MC[i_m]->weight_adaptation_exponent << "   "
		    << endline;
  }
  out_integration << endline;
  for (int i_s = 0; i_s < psi->input_IS.size(); i_s++){
    out_integration << setw(16) << psi->input_IS[i_s]->name << "   "
		    << right << setw(2) << psi->input_IS[i_s]->switch_optimization << "   "
		    << setw(3) << psi->input_IS[i_s]->n_optimization_step << " steps" << "   "
		    << setw(7) << psi->input_IS[i_s]->n_event_per_step << " events" << "   "
		    << setw(5) << psi->input_IS[i_s]->n_gridsize << " gridbins" << "   "
		    << "min: " << setw(1) << psi->input_IS[i_s]->switch_minimum_weight << "   "
		    << "(" << left << setw(5) << psi->input_IS[i_s]->reserved_minimum_weight << " reserved)" << "   "
		    << "max: " << setw(1) << psi->input_IS[i_s]->alpha_maximum_weight << "   "
		    << "mode " << setw(1) << psi->input_IS[i_s]->switch_mode_optimization << "   "
		    << "validation " << setw(1) << psi->input_IS[i_s]->switch_validation_optimization << "   "
		    << "adaptation exponent " << setw(5) << psi->input_IS[i_s]->weight_adaptation_exponent << "   "
		    << endline;
  }
  out_integration << endline;

  if (csi->type_contribution == "RA" || csi->type_contribution == "RRA"){
    for (int i_a = 0; i_a < psi->MC_n_channel_phasespace.size(); i_a++){
      out_integration << setw(21) << "MC_n_channel_phasespace[" << setw(3) << right << i_a << "]   = " << left << psi->MC_n_channel_phasespace[i_a] << endline;
    }
    out_integration << endline;
  }

  out_integration << setw(30) << "switch_MC_tau" << " = " << psi->input_MC_tau.switch_optimization << endline;
  for (int i_c = 0; i_c < psi->tau_MC_map.size(); i_c++){
    stringstream temp;
    temp <<    "tau_MC_map[" << i_c << "]";
    out_integration << setw(30) << temp.str() << " = " << psi->tau_MC_map[i_c] << endline;
  }
  out_integration << endline;

  if (csi->type_contribution == "RA" || csi->type_contribution == "RRA"){
    out_integration << setw(30) << "switch_MC_x_dipole" << " = " << psi->input_MC_x_dipole.switch_optimization << endline;
    for (int i_a = 1; i_a < psi->MC_x_dipole_mapping.size(); i_a++){
      for (int i_m = 0; i_m < psi->MC_x_dipole_mapping[i_a].size(); i_m++){
	out_integration << setw(20) << "MC_x_dipole_mapping[" << setw(3) << right << i_a << "][" << setw(1) << right << i_m << "]    = " << left << psi->MC_x_dipole_mapping[i_a][i_m] << endline;
      }
    }
    out_integration << endline;
  }
  out_integration << endline;

  out_integration << "Relevant objects:" << endline;

  out_integration << setw(10) << "no" << setw(20) << "object" << setw(20) << "n_observed_min" << setw(20) << "n_observed_max" << setw(20) << "define_pT" << setw(20) << "define_ET" << setw(20) << "define_eta"  << setw(20) << "define_y"<< endline;
  for (int i = 0; i < (esi->object_list_selection).size(); i++){
    out_integration << setw(5) << i << setw(20) << esi->object_list_selection[i] << setw(20) << esi->pds[i].n_observed_min << setw(20) << esi->pds[i].n_observed_max << setw(20) << esi->pds[i].define_pT << setw(20) << esi->pds[i].define_ET << setw(20) << esi->pds[i].define_eta << setw(20) << esi->pds[i].define_y << endline;
  }

  out_integration << endline;

  out_integration << "Relevant objects:   runtime_jet_recombination   collects partons that enter jet recombination." << endline;
  for (int i = 0; i < esi->runtime_jet_recombination.size(); i++){
    out_integration << "runtime_jet_recombination[" << setw(3) << i << "] = " << setw(3) << esi->runtime_jet_recombination[i] << "[" << setw(10) <<  left << esi->object_list_selection[esi->runtime_jet_recombination[i]] << "]" << endline;
  }
  out_integration << endline;

  out_integration << "Relevant objects:   runtime_photon_isolation   collects partons that could become isolated photons a la Frixione." << endline;
  for (int i = 0; i < esi->runtime_photon_isolation.size(); i++){
    out_integration << "runtime_photon_isolation[" << setw(3) << i << "] = " << setw(3) << esi->runtime_photon_isolation[i] << "[" << setw(10) <<   left << esi->object_list_selection[esi->runtime_photon_isolation[i]] << "]" << endline;
  }
  out_integration << endline;

  out_integration << "Relevant objects:   runtime_original   collects partons that enter event selection with their parton-level momenta." << endline;
  for (int i = 0; i < esi->runtime_original.size(); i++){
    out_integration << "runtime_original[" << setw(3) << i << "] = " << setw(3) << esi->runtime_original[i] << "[" << setw(10) << left << esi->object_list_selection[esi->runtime_original[i]] << "]" << endline;
  }
  out_integration << endline;

  out_integration << "Relevant objects:   runtime_order   collects partons that belong to the respective object class." << endline;
  for (int i = 0; i < esi->object_list_selection.size(); i++){
    stringstream sstemp;
    for (int j = 0; j < esi->runtime_order[i].size(); j++){sstemp << setw(3) << esi->runtime_order[i][j] << "[" << setw(6) << csi->type_parton[0][esi->runtime_order[i][j]] << "]";}
    out_integration << setw(10) << esi->object_list_selection[i] << " -> " << sstemp.str() << endline;
  }
  out_integration << endline;

  out_integration << "Relevant objects:   runtime_order_inverse   collects object classes a parton belongs to." << endline;
  // !!! csi->type_parton has n_ps entries (vector<vector<int> >) !!!
  for (int i = 3; i < csi->type_parton[0].size(); i++){
    stringstream sstemp;
    for (int j = 0; j < esi->runtime_order_inverse[i].size(); j++){
      sstemp << setw(10) << esi->object_list_selection[esi->runtime_order_inverse[i][j]] << " [" << setw(3) << esi->runtime_order_inverse[i][j] << "]";
    }
    out_integration << "pa[" << setw(3) << i << "] = " << setw(5) << csi->type_parton[0][i] << " -> " << sstemp.str() << endline;
  }
  out_integration << endline;
  out_integration << endline;
  out_integration << "   generated [nan]    accepted (techcut)    rejected    Xsection contribution +- absolute error        dev/XScont   dev/XS@norm" << endline;
  out_integration << "*******************************************************************************************************************************" << endline;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


// function not used at present !!!
void observable_set::output_check_value_scale(){
  Logger logger("observable_set::output_check_value_scale");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;
  for (int i_v = 0; i_v < max_dyn_ren + 1; i_v++){
    logger << LOG_DEBUG_VERBOSE << left << "DS: " << i_v << right << setw(25) << "rel_scale_ren" << setw(24) << "rel_scale_ren" << "²" << setw(25) << "rel_factor_alpha_S" << setw(25) << "scale_ren" << setw(24) << "scale_ren" << "²" << endline_flush;
    for (int i_m = 0; i_m < value_relative_scale_ren[i_v].size(); i_m++){
      logger << LOG_DEBUG_VERBOSE << setw(5) << "" << setw(25) << setprecision(15) << value_relative_scale_ren[i_v][i_m] << setw(25) << setprecision(15) << value_relative_scale2_ren[i_v][i_m] << setw(25) << setprecision(15) << value_relative_factor_alpha_S[0][i_v][i_m] << setw(25) << setprecision(15) << value_scale_ren[0][i_v][i_m] << setw(25) << setprecision(15) << value_scale2_ren[0][i_v][i_m] << endline_flush;
    }
    logger.newLine(LOG_DEBUG_VERBOSE);
  }
  logger.newLine(LOG_DEBUG_VERBOSE);
  for (int i_v = 0; i_v < max_dyn_fact + 1; i_v++){
    logger << LOG_DEBUG_VERBOSE << "DS: " << i_v << setw(25) << "rel_scale_fact" << setw(24) << "rel_scale_fact" << "²" << setw(23) << "log(rel_scale_fact" << "²)" << setw(25) << "scale_fact" << setw(24) << "scale_fact" << "²" << endline_flush;
    for (int i_m = 0; i_m < value_relative_scale_fact[i_v].size(); i_m++){
      logger << LOG_DEBUG_VERBOSE << setw(5) << "" << setw(25) << setprecision(15) << value_relative_scale_fact[i_v][i_m] << setw(25) << setprecision(15) << value_relative_scale2_fact[i_v][i_m] << setw(25) << setprecision(15) << value_relative_logscale2_fact[i_v][i_m] << setw(25) << setprecision(15) << value_scale_fact[0][i_v][i_m] << setw(25) << setprecision(15) << value_scale2_fact[0][i_v][i_m] << endline_flush;
    }
    logger.newLine(LOG_DEBUG_VERBOSE);
  }
  logger.newLine(LOG_DEBUG_VERBOSE);
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}



void observable_set::output_check_running_alpha_S(){
  Logger logger("observable_set::output_check_running_alpha_S");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ofstream out_check_alpha_S;
  string filename_check_alpha_S = "alphaS." + LHAPDFname + ".dat";
  out_check_alpha_S.open(filename_check_alpha_S.c_str(), ofstream::out | ofstream::trunc);
  out_check_alpha_S << "# " << left << setw(40) << LHAPDFname << "getNf = " << LHAPDF::getNf() << "     N_nondecoupled = " << asi->N_nondecoupled << endline;
  int decade = 3;
  int step = 1000;
  for (int i_d = 0; i_d < decade; i_d++){
    int temp_step = step;
    if (i_d == decade - 1){temp_step++;}
    for (int i_s = 0; i_s < temp_step; i_s++){
      double temp_mu_ren = pow(10., double(i_d) + double(i_s) / step);
      double temp_alpha_S = LHAPDF::alphasPDF(temp_mu_ren);
      out_check_alpha_S << setprecision(15) << setw(23) << temp_mu_ren << setprecision(15) << setw(23) << temp_alpha_S << endline;
    }
  }
  out_check_alpha_S.close();

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_finalization_integration(){
  Logger logger("observable_set::output_finalization_integration");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger.newLine(LOG_INFO);
  logger << LOG_INFO << "CPU time:   integration evaluation time = " << h << "h " << min << "min " << sec << "sec     n_gen = " << psi->i_gen << "   n_acc = " << psi->i_acc << endline;
  logger << LOG_INFO << "CPU time:   integration evaluation time in seconds = " << h * 3600 + min * 60 + sec << "     n_gen = " << psi->i_gen << "   n_acc = " << psi->i_acc << endline;
  logger << LOG_INFO << "evaluation time per generated event = "  << showpoint << setprecision(4) << double(3600 * h + 60 * min + sec) / psi->i_gen << " / sec" << endline;
  logger << LOG_INFO << "evaluation time per accepted event  = "  << showpoint << setprecision(4) << double(3600 * h + 60 * min + sec) / psi->i_acc << " / sec" << endline_flush;

  ofstream out_integration;
  out_integration.open(filename_integration.c_str(), ofstream::out | ofstream::app);
  static string stars = "********************************************************";
  out_integration << stars << " final result " << stars << endline;
  out_integration.close();

  ofstream out_execution;
  out_execution.open(filename_execution.c_str(), ofstream::out | ofstream::trunc);
  out_execution << "final result" << endline;
  out_execution.close();

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_integrand_maximum_psp(){
  Logger logger("observable_set::output_integrand_maximum_psp");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ofstream out_maxevent;
  out_maxevent.open(filename_maxevent.c_str(), ofstream::out | ofstream::app);
  for (int i_p = 0; i_p < esi->p_parton[0].size(); i_p++){
    out_maxevent << "esi->p_parton[0][" << i_p << "] = " << esi->p_parton[0][i_p] << "   pT = " << esi->p_parton[0][i_p].pT() << endline;
  }
  out_maxevent << endline;

  for (int i_a = 0; i_a < n_ps; i_a++){
    for (int i_u = 2; i_u < esi->p_parton[i_a].size() - 2; i_u ++){
      //    for (int i_u = 2; i_u < csi->n_particle + 1; i_u ++){
      out_maxevent << "Possible candidates for " << i_u << "-particle resonances for phase-space " << i_a << ":" << endline;
      for (int i_x = 4; i_x < psi->xbp_all[i_a].size(); i_x += 4){
	vector<int> xbno_contained_particle = vectorint_from_binary(i_x);
	//      if (xbno_contained_particle.size() > 1){
	if (xbno_contained_particle.size() == i_u){
	  stringstream tempss;
	  tempss << setw(3) << i_x << " =^= ";
	  for (int i_xi = 0; i_xi < xbno_contained_particle.size(); i_xi++){
	    tempss << setw(3) << xbno_contained_particle[i_xi];
	    if (i_xi < xbno_contained_particle.size() - 1){tempss << " + ";}
	  }
	  if (psi->xbp_all[i_a][i_x] != nullvector){out_maxevent << "sqrt(s[" << tempss.str() << "]) = " << psi->xbp_all[i_a][i_x].m() << endline;}
	}
      }
      out_maxevent << endline;
    }
  }

  out_maxevent << "sqrt(s_part)    = " << (esi->p_parton[0][1] + esi->p_parton[0][2]).m() << endline;
  out_maxevent << "tau             = " << psi->x_pdf[0] << endline;
  out_maxevent << "g_tot           = " << setw(23) << setprecision(15) << psi->g_tot << endline;
  out_maxevent << endline;
  out_maxevent << "psi->MC_tau.channel  = " << setw(5) << psi->MC_tau.channel << "   (absolute probability = " << setw(23) << setprecision(15) << psi->MC_tau.alpha[psi->MC_tau.channel] << endline;
  out_maxevent << "psi->MC_phasespace.channel      = " << setw(5) << psi->MC_phasespace.channel << "   (absolute probability = " << setw(23) << setprecision(15) << psi->MC_phasespace.alpha[psi->MC_phasespace.channel] << endline;

  logger << LOG_DEBUG_VERBOSE << "psi->IS_tau.active_optimization = " << psi->IS_tau.active_optimization << endline_flush;
  if (psi->IS_tau.active_optimization != -1){
    logger << LOG_DEBUG_VERBOSE << "psi->IS_tau.channel = " << psi->IS_tau.channel << endline_flush;
    out_maxevent << "tau_channel     = " << setw(5) << psi->IS_tau.channel << "   (absolute probability = " << setw(23) << setprecision(15) << psi->IS_tau.alpha[psi->IS_tau.channel] << endline;
  }

  logger << LOG_DEBUG_VERBOSE << "psi->IS_x1x2.active_optimization = " << psi->IS_x1x2.active_optimization << endline_flush;
  if (psi->IS_x1x2.active_optimization != -1){
    logger << LOG_DEBUG_VERBOSE << "psi->IS_x1x2.channel = " << psi->IS_x1x2.channel << endline_flush;
    out_maxevent << "x1x2_channel    = " << setw(5) << psi->IS_x1x2.channel << "   (absolute probability = " << setw(23) << setprecision(15) << psi->IS_x1x2.alpha[psi->IS_x1x2.channel] << endline;
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_integrand_maximum(){
  Logger logger("observable_set::output_integrand_maximum");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  if (abs(this_psp_weight) > max_integrand){
    ofstream out_maxevent;
    out_maxevent.open(filename_maxevent.c_str(), ofstream::out | ofstream::app);
    max_integrand = abs(this_psp_weight);
    out_maxevent << right << setw(10) << psi->i_gen << right << setw(10) << psi->i_acc << "   max_integrand = " << this_psp_weight << "   " << "max_integrand / sigma_normalization = " << max_integrand / sigma_normalization << endline;
    out_maxevent.close();

    output_integrand_maximum_psp();

    out_maxevent.open(filename_maxevent.c_str(), ofstream::out | ofstream::app);
    out_maxevent << endline;
    out_maxevent << "integrand = " << setw(23) << setprecision(15) << integrand << endline;
    out_maxevent << "ME2       = " << setw(23) << setprecision(15) << ME2 << endline;

    out_maxevent << endline;
    out_maxevent.close();
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_integrand_maximum_VT2(){
  Logger logger("observable_set::output_integrand_maximum_VT2");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;
  /*
  // TODO: put into perform.maxintegrand.checks
  if (fabs(this_psp_weight) > max_integrand) {
    max_integrand = fabs(this_psp_weight);
    logger << LOG_DEBUG_POINT << psi->i_acc << ", maxint: " << max_integrand << endline_flush;
    logger << LOG_DEBUG_POINT << psi->zz_pdf[1] << ", " << psi->zz_pdf[2] << endline_flush;
  }
  //  logger << LOG_DEBUG_POINT << QT_H1_delta << ", " << QT_H2_delta << endline_flush;
  */
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::perform_integration_step_complete(){
  Logger logger("observable_set::perform_integration_step_complete");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  if (temp_n_step == 0){temp_n_step = psi->n_step;}

  if (switch_output_integration){output_step_integration();}
  if (switch_output_result){output_step_result();}
  if (switch_output_gnuplot){output_step_errorplot();}
  if (switch_output_execution){output_step_execution();}
  if (switch_output_moment){output_step_moment();}

  if (psi->i_acc == 10 * temp_n_step){
    temp_n_step = 10 * temp_n_step;
    logger << LOG_DEBUG_VERBOSE << "temp_n_step = " << temp_n_step << endline_flush;
  }

  if (switch_output_time){output_step_time();}

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_zero_contribution_complete(){
  Logger logger("observable_set::output_zero_contribution_complete");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_INFO << "Contribution is zero. Output is printed." << endline_flush;
  logger << LOG_DEBUG << "before psi->output_optimization_complete" << endline_flush;
  /*
  logger << LOG_DEBUG << "psi->MC_opt_end = " << psi->MC_opt_end << endline_flush;
  logger << LOG_DEBUG << "psi->active_optimization = " << psi->active_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->end_optimization    = " << psi->end_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->MC_phasespace.active_optimization = " << psi->MC_phasespace.active_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->MC_phasespace.end_optimization    = " << psi->MC_phasespace.end_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->MC_tau.active_optimization = " << psi->MC_tau.active_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->MC_tau.end_optimization    = " << psi->MC_tau.end_optimization << endline_flush;
  if (csi->class_contribution_CS_real){
    for (int i_a = 1; i_a < psi->MC_x_dipole.size(); i_a++){
      logger << LOG_DEBUG << "psi->MC_x_dipole[" << i_a << "].active_optimization = " << psi->MC_x_dipole[i_a].active_optimization << endline_flush;
      logger << LOG_DEBUG << "psi->MC_x_dipole[" << i_a << "].end_optimization    = " << psi->MC_x_dipole[i_a].end_optimization << endline_flush;
    }
  }
  logger << LOG_DEBUG << "psi->IS_tau.active_optimization = " << psi->IS_tau.active_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->IS_tau.end_optimization    = " << psi->IS_tau.end_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->IS_x1x2.active_optimization = " << psi->IS_x1x2.active_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->IS_x1x2.end_optimization    = " << psi->IS_x1x2.end_optimization << endline_flush;
  if (csi->class_contribution_collinear){
    for (int i_z = 1; i_z < 3; i_z++){
      logger << LOG_DEBUG << "psi->IS_z1z2[" << i_z << "].active_optimization = " << psi->IS_z1z2[i_z].active_optimization << endline_flush;
      logger << LOG_DEBUG << "psi->IS_z1z2[" << i_z << "].end_optimization    = " << psi->IS_z1z2[i_z].end_optimization << endline_flush;
    }
  }

  psi->MC_opt_end = 1;
  psi->end_optimization = 1;
  if (psi->MC_phasespace.active_optimization){psi->MC_phasespace.end_optimization = 1;}
  if (psi->MC_tau.active_optimization){psi->MC_tau.end_optimization = 1;}
  if (csi->class_contribution_CS_real){
    for (int i_a = 1; i_a < psi->MC_x_dipole.size(); i_a++){
      if (psi->MC_x_dipole[i_a].active_optimization){psi->MC_x_dipole[i_a].end_optimization = 1;}
    }
  }
  if (psi->IS_tau.active_optimization){psi->IS_tau.end_optimization = 1;}
  if (psi->IS_x1x2.active_optimization){psi->IS_x1x2.end_optimization = 1;}
  if (csi->class_contribution_collinear){for (int i_z = 1; i_z < 3; i_z++){if (psi->IS_z1z2[i_z].active_optimization){psi->IS_z1z2[i_z].end_optimization = 1;}}}

  psi->random_manager.end_optimization = 1;
  for (int i_r = 0; i_r < psi->random_manager.random_psp.size(); i_r++){psi->random_manager.random_psp[i_r].IS_PS.end_optimization = 1;}
  logger << LOG_DEBUG << "before psi->output_optimization_complete" << endline_flush;
  logger << LOG_DEBUG << "psi->MC_opt_end = " << psi->MC_opt_end << endline_flush;
  logger << LOG_DEBUG << "psi->active_optimization = " << psi->active_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->end_optimization    = " << psi->end_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->MC_phasespace.active_optimization = " << psi->MC_phasespace.active_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->MC_phasespace.end_optimization    = " << psi->MC_phasespace.end_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->MC_tau.active_optimization = " << psi->MC_tau.active_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->MC_tau.end_optimization    = " << psi->MC_tau.end_optimization << endline_flush;
  if (csi->class_contribution_CS_real){
    for (int i_a = 1; i_a < psi->MC_x_dipole.size(); i_a++){
      logger << LOG_DEBUG << "psi->MC_x_dipole[" << i_a << "].active_optimization = " << psi->MC_x_dipole[i_a].active_optimization << endline_flush;
      logger << LOG_DEBUG << "psi->MC_x_dipole[" << i_a << "].end_optimization    = " << psi->MC_x_dipole[i_a].end_optimization << endline_flush;
    }
  }
  logger << LOG_DEBUG << "psi->IS_tau.active_optimization = " << psi->IS_tau.active_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->IS_tau.end_optimization    = " << psi->IS_tau.end_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->IS_x1x2.active_optimization = " << psi->IS_x1x2.active_optimization << endline_flush;
  logger << LOG_DEBUG << "psi->IS_x1x2.end_optimization    = " << psi->IS_x1x2.end_optimization << endline_flush;
  if (csi->class_contribution_collinear){
    for (int i_z = 1; i_z < 3; i_z++){
      logger << LOG_DEBUG << "psi->IS_z1z2[" << i_z << "].active_optimization = " << psi->IS_z1z2[i_z].active_optimization << endline_flush;
      logger << LOG_DEBUG << "psi->IS_z1z2[" << i_z << "].end_optimization    = " << psi->IS_z1z2[i_z].end_optimization << endline_flush;
    }
  }

  psi->output_optimization_complete();
  */
  logger << LOG_DEBUG << "after psi->output_optimization_complete" << endline_flush;

  //      output_step_integration();
  logger << LOG_DEBUG << "before temp_n_step = " << temp_n_step << endline_flush;
  if (temp_n_step == 0){temp_n_step = psi->n_step;}
  logger << LOG_DEBUG << "after temp_n_step = " << temp_n_step << endline_flush;

  if (switch_output_integration){output_step_integration();}
  if (switch_output_result){output_step_result();}
  if (switch_output_moment){output_step_moment();}
  if (switch_output_execution){output_step_execution();}
  if (switch_output_distribution){output_distribution_complete();}

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_step_integration_TSV(){
  Logger logger("observable_set::output_step_integration_TSV");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "n_extended_set_TSV = " << n_extended_set_TSV << endline_flush;
  for (int i_s = 0; i_s < n_extended_set_TSV; i_s++){
    ofstream out_integration;
    out_integration.open((filename_integration_TSV[i_s]).c_str(), ofstream::out | ofstream::app);

    int next_no_qTcut = 0;
    for (int i_q = 0; i_q < output_n_qTcut; i_q++){
      if (i_q != no_qTcut_integration[next_no_qTcut]){continue;}

      logger << LOG_DEBUG << "i_q = " << i_q << "   no_qTcut_integration[next_no_qTcut = " << next_no_qTcut << "] = " << no_qTcut_integration[next_no_qTcut] << endline_flush;

      out_integration << endline;
      if (active_qTcut == 0){
	out_integration << left << setw(21) << "qTcut - independent";
      }
      else {
	out_integration << left << setw(8) << "qTcut = " << setprecision(3) << setw(6) << showpoint << value_qTcut[i_q] << setw(7) << " GeV   ";
      }
      if (i_q == 0){
	out_integration << "n(gen) = " << setw(12) << psi->i_gen << "   n(acc) = " << setw(12) << psi->i_acc << "   n(rej) = " << setw(12) << psi->i_rej << "   n(nan) = " << setw(12) << psi->i_nan << "   n(tec) = " << setw(12) << psi->i_tec;
      }
      out_integration << endline;
      out_integration << endline;

      double abs_central_Xsection = abs(Xsection_TSV[i_s][i_q][no_central_scale_ren_TSV[i_s]][no_central_scale_fact_TSV[i_s]]);

      for (int i_r = 0; i_r < n_scale_ren_TSV[i_s]; i_r++){
	for (int i_f = 0; i_f < n_scale_fact_TSV[i_s]; i_f++){
	  double delta_NLO_LO = abs(Xsection_delta_TSV[i_s][i_q][i_r][i_f] / sigma_normalization);

	  int shift;
	  string unit;
	  double output_Xsection_delta = 0.;
	  double output_Xsection = 0.;
	  if (abs_central_Xsection >= 1.e3){output_Xsection = Xsection_TSV[i_s][i_q][i_r][i_f] / 1.e3; output_Xsection_delta = Xsection_delta_TSV[i_s][i_q][i_r][i_f] / 1.e3; unit = "pb";}
	  else if (abs_central_Xsection >= 1.e0){output_Xsection = Xsection_TSV[i_s][i_q][i_r][i_f]; output_Xsection_delta = Xsection_delta_TSV[i_s][i_q][i_r][i_f]; unit = "fb"; shift = 0;}
	  else if (abs_central_Xsection >= 1.e-3){output_Xsection = Xsection_TSV[i_s][i_q][i_r][i_f] * 1.e3; output_Xsection_delta = Xsection_delta_TSV[i_s][i_q][i_r][i_f] * 1.e3; unit = "ab";}
	  else if (abs_central_Xsection >= 1.e-6){output_Xsection = Xsection_TSV[i_s][i_q][i_r][i_f] * 1.e6; output_Xsection_delta = Xsection_delta_TSV[i_s][i_q][i_r][i_f] * 1.e6; unit = "zb";}
	  else if (abs_central_Xsection >= 1.e-9){output_Xsection = Xsection_TSV[i_s][i_q][i_r][i_f] * 1.e9; output_Xsection_delta = Xsection_delta_TSV[i_s][i_q][i_r][i_f] * 1.e9; unit = "yb";}

	  double dlog_delta_NLO_LO = log10(abs(delta_NLO_LO));
	  int log_delta_NLO_LO;
	  log_delta_NLO_LO = int(dlog_delta_NLO_LO);
	  if (dlog_delta_NLO_LO < 0.){log_delta_NLO_LO--;}

	  double dlog_Xsection = log10(abs(output_Xsection));
	  int log_Xsection;
	  log_Xsection = int(dlog_Xsection);
	  if (dlog_Xsection < 0.){log_Xsection--;}

	  double dlog_Xsection_delta = log10(abs(output_Xsection_delta));
	  int log_Xsection_delta;
	  log_Xsection_delta = int(dlog_Xsection_delta);
	  if (dlog_Xsection_delta < 0.){log_Xsection_delta--;}

	  double dlog_rel = log10(abs(output_Xsection_delta / output_Xsection));
	  int log_rel;
	  log_rel = int(dlog_rel);
	  if (dlog_rel < 0.){log_rel--;}

	  double dlog_rel_mu_R = log10(relative_scale_ren_TSV[i_s][i_r]);
	  int log_rel_mu_R;
	  log_rel_mu_R = int(dlog_rel_mu_R);
	  if (dlog_rel_mu_R < 0.){log_rel_mu_R--;}
	  //	  out_integration << log_rel_mu_R << endline;

	  double dlog_rel_mu_F = log10(relative_scale_fact_TSV[i_s][i_f]);
	  int log_rel_mu_F;
	  log_rel_mu_F = int(dlog_rel_mu_F);
	  if (dlog_rel_mu_F < 0.){log_rel_mu_F--;}
	  //	  out_integration << log_rel_mu_F << endline;

	  out_integration << setw(10) << "";
	  out_integration << setw(11) << "rel_mu_R = " << setprecision(6 + log_rel_mu_R) << setw(8) << showpoint << relative_scale_ren_TSV[i_s][i_r] << "   ";
	  out_integration << setw(11) << "rel_mu_F = " << setprecision(6 + log_rel_mu_F) << setw(8) << showpoint << relative_scale_fact_TSV[i_s][i_f] << "   ";

	  out_integration << "cXS = " << setw(10) << setprecision(7 + log_Xsection + shift) << showpoint << right << output_Xsection << " +- " << setw(10) << setprecision(7 + log_Xsection_delta + shift) << output_Xsection_delta << " " << unit << "   d(cXS) / cXS = " << left << setw(11) << setprecision(6) << abs(output_Xsection_delta / output_Xsection) << "   d(cXS) / XS = " << setw(11) << setprecision(6) << delta_NLO_LO << endline;
	}
      }
      if (next_no_qTcut < no_qTcut_integration.size() - 1){next_no_qTcut++;}

    }
    out_integration.close();
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_step_result_TSV(){
  Logger logger("observable_set::output_step_result_TSV");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (int i_s = 0; i_s < n_extended_set_TSV; i_s++){
    ofstream out_result;
    out_result.open((filename_result_TSV[i_s]).c_str(), ofstream::out | ofstream::trunc);
    out_result << setw(12) << left << psi->i_gen << endline;
    for (int i_q = 0; i_q < output_n_qTcut; i_q++){

      // Add feature to print only selected values later !!!

      if (active_qTcut == 0){
	out_result << "#   qTcut - independent" << endline;
      }
      else {
	out_result << "#   qTcut[" << setw(3) << i_q << "] = " << setw(15) << setprecision(8) << value_qTcut[i_q] << endline;
      }
      for (int i_r = 0; i_r < n_scale_ren_TSV[i_s]; i_r++){
	for (int i_f = 0; i_f < n_scale_fact_TSV[i_s]; i_f++){
	  out_result << "  " << setw(3) << i_r << " " << setw(15) << setprecision(8) << relative_scale_ren_TSV[i_s][i_r];
	  out_result << "  " << setw(3) << i_f << " " << setw(15) << setprecision(8) << relative_scale_fact_TSV[i_s][i_f];
	  out_result << "  " << setw(24) << setprecision(15) << unit_factor_calculation * Xsection_TSV[i_s][i_q][i_r][i_f] << setw(24) << setprecision(15) << unit_factor_calculation * Xsection_delta_TSV[i_s][i_q][i_r][i_f];
	  out_result << endline;
	}
      }
    }
    out_result.close();
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_step_moment_TSV(){
  Logger logger("observable_set::output_step_moment_TSV");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (int i_s = 0; i_s < n_extended_set_TSV; i_s++){
    if (!switch_moment_TSV[i_s]){continue;}

    logger << LOG_DEBUG << "Moment output into " << filename_moment_TSV[i_s] << endline_flush;

    ofstream out_moment;
    out_moment.open((filename_moment_TSV[i_s]).c_str(), ofstream::out | ofstream::trunc);
    out_moment << setw(12) << left << psi->i_gen << endline;
    for (int i_m = 0; i_m < n_moments; i_m++){
      out_moment << "#   moment " << i_m << endline;

      for (int i_q = 0; i_q < output_n_qTcut; i_q++){

	// Add feature to print only selected values later !!!

	if (active_qTcut == 0){
	  out_moment << "#   qTcut - independent" << endline;
	}
	else {
	  out_moment << "#   qTcut[" << setw(3) << i_q << "] = " << setw(15) << setprecision(8) << value_qTcut[i_q] << endline;
	}
	for (int i_r = 0; i_r < n_scale_ren_TSV[i_s]; i_r++){
	  for (int i_f = 0; i_f < n_scale_fact_TSV[i_s]; i_f++){
	    out_moment << "  " << setw(3) << i_r << " " << setw(15) << setprecision(8) << relative_scale_ren_TSV[i_s][i_r];
	    out_moment << "  " << setw(3) << i_f << " " << setw(15) << setprecision(8) << relative_scale_fact_TSV[i_s][i_f];
	    out_moment << "  " << setw(24) << setprecision(15) << unit_factor_calculation * Xsection_moment_TSV[i_s][i_m][i_q][i_r][i_f] << setw(24) << setprecision(15) << unit_factor_calculation * Xsection_moment_delta_TSV[i_s][i_m][i_q][i_r][i_f];
	    out_moment << endline;
	  }
	}
      }
    }
    out_moment.close();
  }
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


// not used !!!
void observable_set::output_step_distribution_TSV(){
  Logger logger("observable_set::output_step_distribution_TSV");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ofstream out_distribution;
  for (int i_s = 0; i_s < n_extended_set_TSV; i_s++){
    if (switch_distribution_TSV[i_s] > 0){
      out_distribution.open((filename_distribution_TSV[i_s]).c_str(), ofstream::out | ofstream::trunc);
      // Replaced 'psi->i_gen' by 'n_ps * psi->i_gen' !!!
      logger << LOG_DEBUG_VERBOSE << "n_ps = " << n_ps << endline_flush;
      logger << LOG_DEBUG_VERBOSE << "psi->i_gen = " << psi->i_gen << endline_flush;
      logger << LOG_DEBUG_VERBOSE << "n_ps * psi->i_gen = " << n_ps * psi->i_gen << endline_flush;
      out_distribution << setw(12) << left << n_ps * psi->i_gen << endline;
      //      out_distribution << setw(12) << left << psi->i_gen << endline;
      //      out_distribution << setw(12) << left << psi->i_gen << endline;

      for (int i_q = 0; i_q < output_n_qTcut; i_q++){
	//	if (active_qTcut > 0 || i_q == 0){
	if (active_qTcut == 0){out_distribution << "#   no qTcut - independent" << endline;}
	else {out_distribution << "#   qTcut[" << setw(3) << i_q << "] = " << setw(15) << setprecision(8) << value_qTcut[i_q] << endline;}
	for (int i_d = 0; i_d < dat.size(); i_d++){
	  out_distribution << "#   dat.name[" << setw(3) << i_d << "] = " << dat[i_d].xdistribution_name << endline;
	  for (int i_r = 0; i_r < n_scale_ren_TSV[i_s]; i_r++){
	    for (int i_f = 0; i_f < n_scale_fact_TSV[i_s]; i_f++){
	      out_distribution << "#   ";
	      out_distribution << "relative_scale_ren[" << setw(3) << i_r << "] = " << setw(15) << setprecision(8) << relative_scale_ren_TSV[i_s][i_r];
	      out_distribution << "   relative_scale_fact[" << setw(3) << i_f << "] = " << setw(15) << setprecision(8) << relative_scale_fact_TSV[i_s][i_f];
	      out_distribution << endline;
	      for (int i_b = 0; i_b < dat[i_d].n_bins; i_b++){
		out_distribution << setw(12) << bin_count_TSV[i_q][i_d][i_b] << "  ";
		out_distribution << setw(24) << setprecision(15) << unit_factor_calculation * bin_weight_TSV[i_s][i_q][i_r][i_f][i_d][i_b] << "  ";
		out_distribution << setw(24) << setprecision(15) << unit_factor2_calculation * bin_weight2_TSV[i_s][i_q][i_r][i_f][i_d][i_b];
		out_distribution << endline;
	      }
	    }
	  }
	}
      }
      out_distribution.close();
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}
// until here ???


void observable_set::output_step_integration(){
  Logger logger("observable_set::output_step_integration");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ofstream out_integration;
  out_integration.open(filename_integration.c_str(), ofstream::out | ofstream::app);
  double delta_NLO_LO = abs(Xsection_delta / sigma_normalization);
  int log_rel = int(log10(abs(Xsection_delta / Xsection)));
  out_integration << setw(12) << psi->i_gen << " [" << setw(3) << psi->i_nan << "] " << " ";
  out_integration << setw(10) << psi->i_acc << " (" << setw(7) << psi->i_tec << ") " << " ";
  out_integration << setw(10) << psi->i_rej << " ";
  out_integration << "  sigma = " << setw(14) << setprecision(8) << showpoint << unit_factor_calculation * Xsection << " +- " << setw(14) << setprecision(8 + log_rel) << unit_factor_calculation * Xsection_delta << " " << unit_calculation << "    " << setw(11) << setprecision(7 + log_rel) << abs(Xsection_delta / Xsection);
  out_integration << "   " << setw(11) << setprecision(6) << delta_NLO_LO << endline;
  out_integration.close();
 
  if (switch_CV && temp_n_step != 0){
    logger << LOG_DEBUG_VERBOSE << "filename_integration_CV = " << filename_integration_CV << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "output_n_qTcut = " << output_n_qTcut << endline_flush;
    logger << LOG_DEBUG_VERBOSE << "n_scales_CV = " << n_scales_CV << endline_flush;
    out_integration.open((filename_integration_CV).c_str(), ofstream::out | ofstream::app);
    for (int i_c = 0; i_c < output_n_qTcut; i_c++){
      for (int i_s = 0; i_s < n_scales_CV; i_s++){
	logger << LOG_DEBUG_VERBOSE << "i_c = " << i_c << endline_flush;
	logger << LOG_DEBUG_VERBOSE << "i_s = " << i_s << endline_flush;
	logger << LOG_DEBUG_VERBOSE << "Xsection_CV.size() = " << Xsection_CV.size() << endline_flush;
	logger << LOG_DEBUG_VERBOSE << "Xsection_delta_CV.size() = " << Xsection_delta_CV.size() << endline_flush;
	logger << LOG_DEBUG_VERBOSE << "Xsection_CV[" << i_c << "].size() = " << Xsection_CV[i_c].size() << endline_flush;
	logger << LOG_DEBUG_VERBOSE << "Xsection_delta_CV[" << i_c << "].size() = " << Xsection_delta_CV[i_c].size() << endline_flush;
	logger << LOG_DEBUG_VERBOSE << "temp_n_step = " << temp_n_step << endline_flush;
	logger << LOG_DEBUG_VERBOSE << "(psi->i_acc % temp_n_step) = " << (psi->i_acc % temp_n_step) << endline_flush;
	logger << LOG_DEBUG_VERBOSE << "psi->i_acc == psi->n_events_max = " << (psi->i_acc == psi->n_events_max) << endline_flush;
	logger << LOG_DEBUG_VERBOSE << "int_end == 1 = " << (int_end == 1) << endline_flush;
	if ((psi->i_acc % temp_n_step) == 0 || psi->i_acc == psi->n_events_max || int_end == 1){
	  double delta_NLO_LO = abs(Xsection_delta_CV[i_c][i_s] / sigma_normalization);
	  int log_rel = int(log10(abs(Xsection_delta_CV[i_c][i_s] / Xsection_CV[i_c][i_s])));
	  if (i_c == 0 && i_s == 0){
	    out_integration << "n(gen) = " << setw(12) << psi->i_gen << "   n(acc) = " << setw(12) << psi->i_acc << "   n(rej) = " << setw(12) << psi->i_rej << "   n(nan) = " << setw(12) << psi->i_nan << "   n(tec) = " << setw(12) << psi->i_tec << endline;
	  }
	  if (i_s == 0){
	  logger << LOG_DEBUG_VERBOSE << "value_qTcut.size() = " << value_qTcut.size() << endline_flush;
	    if (active_qTcut == 1){out_integration << setw(6) << "qTcut = " << setprecision(3) << setw(6) << showpoint << value_qTcut[i_c] << " GeV   ";}
	    else {out_integration << setw(6) << "qTcut - independent  ";}
	  }
	  else {out_integration << setw(21) << "";}

	  if (dynamic_scale_CV == 0){
	    out_integration << setw(9) << "mu_ren = " << setprecision(5) << setw(8) << showpoint << mu_ren_CV[1][i_s] << " GeV   " << setw(10) << "mu_fact = " << setprecision(5) << setw(8) << showpoint << mu_fact_CV[1][i_s] << " GeV ";
	  }
	  else {
	    out_integration << setw(9) << "mu_ren = " << setprecision(5) << setw(8) << showpoint << rel_scale_factor_ren_CV[i_s] << " DS" << dynamic_scale_CV << setw(10) << "    mu_fact = " << setprecision(5) << setw(8) << showpoint << rel_scale_factor_fact_CV[i_s] << " DS" << dynamic_scale_CV;
	  }

	  out_integration << "  sigma = " << setw(15) << setprecision(10) << showpoint << unit_factor_calculation * Xsection_CV[i_c][i_s] << " +- " << setw(15) << setprecision(10 + log_rel) << unit_factor_calculation * Xsection_delta_CV[i_c][i_s] << " " << unit_calculation << "    " << setw(11) << setprecision(7 + log_rel) << abs(Xsection_delta_CV[i_c][i_s] / Xsection_CV[i_c][i_s]) << "   " << setw(11) << setprecision(6) << delta_NLO_LO << endline;

	  if ((i_c == output_n_qTcut - 1) && (i_s == n_scales_CV - 1)){out_integration << endline;}
	}
      }
    }
    out_integration.close();
  }

  if (switch_TSV){output_step_integration_TSV();}

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_step_integration_weight_optimization(){
  Logger logger("observable_set::output_step_integration_weight_optimization");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;
  /*
void output_weight_optimization(phasespace_set & psi, observable_set & oset, int & size_proc_generic, vector<int> & size_proceeding){
  Logger logger("output_weight_optimization");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;
  */

  if (psi->MC_phasespace.active_optimization == 1){
    /////    if (psi->MC_phasespace.end_optimization < 2 && *psi->i_step_mode % psi->MC_phasespace.n_event_per_step == 0){
    logger << LOG_INFO << "psi->i_step = " << psi->i_step << "   psi->MC_phasespace.n_optimization_repetition = " << psi->MC_phasespace.n_optimization_repetition << "   (psi->i_step % psi->MC_phasespace.n_optimization_repetition = " << (psi->i_step % psi->MC_phasespace.n_optimization_repetition) << endline_flush;
    if (psi->MC_phasespace.end_optimization < 2 && (psi->i_step % psi->MC_phasespace.n_optimization_repetition == 0) && (*psi->i_step_mode == psi->MC_phasespace.n_event_per_step / psi->MC_phasespace.n_optimization_repetition || psi->enforce_step)){
      ///    if (psi->MC_phasespace.end_optimization < 2 && (*psi->i_step_mode == psi->MC_phasespace.n_event_per_step / psi->MC_phasespace.n_optimization_repetition || psi->enforce_step)){
      //  if (psi->MC_opt_end < 2 && *psi->i_step_mode % psi->n_event_per_step == 0){
      ofstream out_integration;
      out_integration.open(filename_integration.c_str(), ofstream::out | ofstream::app);
      out_integration << "************************************************* weight optimization step " << setw(2) << psi->MC_phasespace.i_alpha_it << " *************************************************" << endline;
      out_integration << "   D_min[alpha(i),alpha(j)](step " << psi->MC_phasespace.i_alpha_it - 1 << ") = " << setw(20) << setprecision(15) << psi->MC_phasespace.diff_w[psi->MC_phasespace.i_alpha_it - 1] << endline;
      if (!(*psi->i_step_mode == psi->MC_phasespace.n_event_per_step  * psi->MC_phasespace.n_optimization_step)){
	if (psi->MC_phasespace.counter_minimum_weight != 0){
	  out_integration << "   switch_minimum_weight:   " << psi->MC_phasespace.counter_minimum_weight << " alpha's are set to " << psi->MC_phasespace.alpha_reserved_min << " in the next step." << endline;
	}
      out_integration << "*******************************************************************************************************************************" << endline;
      }
      out_integration.close();
    }

      //    if (psi->MC_phasespace.end_optimization < 2 && *psi->i_step_mode == psi->MC_phasespace.n_event_per_step  * psi->MC_phasespace.n_optimization_step){
    //    if (psi->MC_phasespace.end_optimization < 2 && (*psi->i_step_mode == psi->MC_phasespace.n_event_per_step / psi->MC_phasespace.n_optimization_repetition  || psi->enforce_step) && psi->i_step == psi->MC_phasespace.n_optimization_step){
    if (psi->MC_phasespace.end_optimization < 2 && (*psi->i_step_mode == psi->MC_phasespace.n_event_per_step / psi->MC_phasespace.n_optimization_repetition  || psi->enforce_step) && psi->i_step == psi->MC_phasespace.n_optimization_step * psi->MC_phasespace.n_optimization_repetition){
      ofstream out_integration;
      out_integration.open(filename_integration.c_str(), ofstream::out | ofstream::app);
      out_integration << "************************************************* weight optimization result **************************************************" << endline;
      for (int j = 0; j < psi->MC_phasespace.diff_w.size(); j++){
	out_integration << "diff_w[" << j << "] = " << showpoint << setprecision(15) << setw(25) << psi->MC_phasespace.diff_w[j] << "   " << psi->MC_phasespace.alpha_it[j][0] << "   " << psi->MC_phasespace.alpha_it[j][psi->MC_phasespace.alpha_it[j].size() - 1] << endline;
      }
      out_integration << "************************************************* weight optimization result **************************************************" << endline;
      out_integration << "   min(D_min[alpha(i),alpha(j)]) = " << showpoint << setprecision(15) << setw(25) << psi->MC_phasespace.diff_w[psi->MC_phasespace.x_minimum_diff_w] << "   at optimization step "<< setw(2) << psi->MC_phasespace.x_minimum_diff_w << endline;
      out_integration << "************************************************* end of weight optimization **************************************************" << endline;
      out_integration.close();
      /*
      size_proceeding[2] = 0;
      size_proc_generic = accumulate(size_proceeding.begin(), size_proceeding.end(), 0);
      ///    psi->MC_opt_end = 2;
      */
    }
  }
  /*
  if (psi->IS_tau.active_optimization == 1){
    if (psi->IS_tau.end_optimization < 2 && (*psi->i_step_mode % psi->IS_tau.n_event_per_step == 0 || *psi->i_step_mode != 0)){
      if (*psi->i_step_mode == psi->IS_tau.n_event_per_step * psi->IS_tau.n_optimization_step){
	size_proceeding[3] = 0;
	size_proc_generic = accumulate(size_proceeding.begin(), size_proceeding.end(), 0);
      }
    }
  }

  if (psi->IS_x1x2.active_optimization == 1){
    if (psi->IS_x1x2.end_optimization < 2 && (*psi->i_step_mode % psi->IS_x1x2.n_event_per_step == 0 || *psi->i_step_mode != 0)){
      if (*psi->i_step_mode == psi->IS_x1x2.n_event_per_step * psi->IS_x1x2.n_optimization_step){
	size_proceeding[4] = 0;
	size_proc_generic = accumulate(size_proceeding.begin(), size_proceeding.end(), 0);
      }
    }
  }
  */

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_step_result(){
  Logger logger("observable_set::output_step_result");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ofstream out_result;
  out_result.open(filename_result.c_str(), ofstream::out | ofstream::trunc);
  out_result << psi->i_gen << endline << setprecision(15) << unit_factor_calculation * Xsection << endline << setprecision(15) << unit_factor_calculation * Xsection_delta << endline;

  if (switch_CV){
    for (int i_c = 0; i_c < output_n_qTcut; i_c++){
      for (int i_s = 0; i_s < n_scales_CV; i_s++){
	out_result << setprecision(15) << unit_factor_calculation * Xsection_CV[i_c][i_s] << endline << setprecision(15) << unit_factor_calculation * Xsection_delta_CV[i_c][i_s] << endline;
      }
    }
  }

  long long temp_time = 3600 * h + 60 * min + sec;
  out_result << temp_time << endline;
  out_result << setprecision(15) << temp_time / double(psi->i_acc) << endline;
  out_result << setprecision(15) << pow(Xsection_delta, 2) * temp_time << endline;
  out_result << setprecision(15) << pow(Xsection_delta, 2) * double(psi->i_acc) << endline;
  /*
  // Adapt summary routine first !!!
  xmunich->end_time_point = chrono::system_clock::now();
  xmunich->end_time = chrono::system_clock::to_time_t(xmunich->end_time_point);
  out_result << setprecision(15) << noshowpoint << difftime(xmunich->end_time, xmunich->y2k_time) << endline;
  out_result << ctime(&xmunich->end_time);
  */
  out_result.close();

  if (switch_TSV){output_step_result_TSV();}

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

void observable_set::output_step_moment(){
  Logger logger("observable_set::output_step_moment");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (int i_m = 0; i_m < n_moments; i_m++){
    ofstream out_moment;
    logger << LOG_DEBUG << "Moment output into " << filename_moment[i_m] << endline_flush;

    out_moment.open(filename_moment[i_m].c_str(), ofstream::out | ofstream::trunc);
    out_moment << psi->i_gen << endline << setprecision(15) << unit_factor_calculation * Xsection_moment[i_m] << endline << setprecision(15) << unit_factor_calculation * Xsection_moment_delta[i_m] << endline;
    if (switch_CV){
      for (int i_c = 0; i_c < output_n_qTcut; i_c++){
	for (int i_s = 0; i_s < n_scales_CV; i_s++){
	  out_moment << setprecision(15) << unit_factor_calculation * Xsection_moment_CV[i_m][i_c][i_s] << endline << setprecision(15) << unit_factor_calculation * Xsection_moment_delta_CV[i_m][i_c][i_s] << endline;
	}
      }
    }
    long long temp_time = 3600 * h + 60 * min + sec;
    out_moment << temp_time << endline;
    out_moment << setprecision(15) << temp_time / double(psi->i_acc) << endline;
    out_moment << setprecision(15) << pow(Xsection_moment_delta[i_m], 2) * temp_time << endline;
    out_moment << setprecision(15) << pow(Xsection_moment_delta[i_m], 2) * double(psi->i_acc) << endline;
    out_moment.close();
  }

  if (switch_TSV){output_step_moment_TSV();}

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_step_time(){
  Logger logger("observable_set::output_step_time");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "filename_time = " << filename_time << endline_flush;
  ofstream out_time;
  out_time.open(filename_time.c_str(), ofstream::out | ofstream::trunc);
  xmunich->walltime_now();
  std::chrono::duration<double> elapsed_seconds = xmunich->now_time_point - xmunich->start_time_point;
  out_time << setprecision(15) << difftime(xmunich->now_time, xmunich->y2k_time) << endline;
  out_time << double2hexastr(elapsed_seconds.count()) << endline;
  out_time << "# start:   " << ctime(&xmunich->start_time);
  out_time << "# final:   " << ctime(&xmunich->now_time);
  out_time << "# elapsed walltime (sec):   " << setprecision(15) << setw(23) << elapsed_seconds.count() << endline;
  out_time << "# elapsed CPU-time (sec):   " << setprecision(15) << setw(23) << 3600 * h + 60 * min + sec << endline;
  out_time.close();

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_step_execution(){
  Logger logger("observable_set::output_step_execution");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ofstream out_execution;
  if (*psi->i_step_mode == psi->n_step || psi->enforce_step){
    out_execution.open(filename_execution.c_str(), ofstream::out | ofstream::trunc);
    out_execution << "started" << endline;
    out_execution.close();
  }

  if (int_end == 1){
    out_execution.open(filename_execution.c_str(), ofstream::out | ofstream::trunc);
    out_execution << "final result" << endline;
    out_execution.close();
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


  // Should be obsolete and removed !!!
void observable_set::output_step_errorplot(){
  Logger logger("observable_set::output_step_errorplot");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ofstream out_gnuplot;
  out_gnuplot.open(filename_gnuplot.c_str(), ofstream::out | ofstream::app);
  out_gnuplot << setw(10) << psi->i_acc << setw(23) << Xsection << setw(23) << Xsection_delta << setw(23) << abs(Xsection_delta / sigma_normalization) << endline;
  //  if (psi->i_acc == temp_n_step || psi->i_acc % (10 * temp_n_step) == 0 || psi->i_acc == psi->n_events_max || int_end == 1){
  double delta = abs(Xsection_delta / sigma_normalization);
  output_gnuplotfile(filename_makegnuplot, psi->i_acc, delta, psi->n_events_max);
  //  }
  out_gnuplot.close();

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_distribution_complete(){
  Logger logger("observable_set::output_distribution_complete");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  if (switch_distribution){
    if (switch_CV){
      output_distribution_CV();
      output_dddistribution_CV();
    }
    if (switch_TSV){output_distribution_TSV();}
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


// still required ???
void observable_set::output_distribution(){
  Logger logger("observable_set::output_distribution");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ofstream out_distribution;
  for (int i_d = 0; i_d < dat.size(); i_d++){
  logger << LOG_DEBUG_VERBOSE << "filename_distribution[" << i_d << "] = " << filename_distribution[i_d] << endline_flush;
    out_distribution.open(filename_distribution[i_d].c_str(), ofstream::out | ofstream::trunc);
    out_distribution << psi->i_gen << endline;
    for (int i_b = 0; i_b < dat[i_d].n_bins; i_b++){
      out_distribution << bin_counts[i_d][i_b] << endline;
      out_distribution << setprecision(15) << unit_factor_calculation * bin_weight[i_d][i_b] << endline;
      out_distribution << setprecision(15) << unit_factor2_calculation * bin_weight2[i_d][i_b] << endline;
    }
    out_distribution.close();
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


// still required ???
void observable_set::output_dddistribution(){
  Logger logger("observable_set::output_dddistribution");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ofstream out_dddistribution;

  for (int i_d = 0; i_d < dddat.size(); i_d++){
  logger << LOG_DEBUG_VERBOSE << "XXX filename_dddistribution[" << i_d << "] = " << filename_dddistribution[i_d] << endline_flush;
    out_dddistribution.open(filename_dddistribution[i_d].c_str(), ofstream::out | ofstream::trunc);
    out_dddistribution << psi->i_gen << endline;
    for (int i_b = 0; i_b < dddat[i_d].n_bins; i_b++){
      out_dddistribution << bin_counts[dat.size() + i_d][i_b] << endline;
      out_dddistribution << setprecision(15) << unit_factor_calculation * bin_weight[dat.size() + i_d][i_b] << endline;
      out_dddistribution << setprecision(15) << unit_factor2_calculation * bin_weight2[dat.size() + i_d][i_b] << endline;
    }
    out_dddistribution.close();
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_distribution_CV(){
  Logger logger("observable_set::output_distribution_CV");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ofstream out_distribution;
  for (int i_s = 0; i_s < n_scales_CV; i_s++){
    for (int i_d = 0; i_d < dat.size(); i_d++){
      logger << LOG_DEBUG_VERBOSE << "filename_distribution_CV[" << i_s << "][" << i_d << "] = " << filename_distribution_CV[i_s][i_d] << endline_flush;
      out_distribution.open(filename_distribution_CV[i_s][i_d].c_str(), ofstream::out | ofstream::trunc);
      //      out_distribution << "# " << dat[i_d].distname() << endline;
      out_distribution << psi->i_gen << endline;
      for (int i_b = 0; i_b < dat[i_d].n_bins; i_b++){
	out_distribution << bin_counts[i_d][i_b] << endline;
	out_distribution << setprecision(15) << unit_factor_calculation * bin_weight_CV[i_s][i_d][i_b] << endline;
	out_distribution << setprecision(15) << unit_factor2_calculation * bin_weight2_CV[i_s][i_d][i_b] << endline;
      }
      out_distribution.close();
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_dddistribution_CV(){
  Logger logger("observable_set::output_dddistribution_CV");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ofstream out_dddistribution;
  for (int i_s = 0; i_s < n_scales_CV; i_s++){
    logger << LOG_DEBUG_VERBOSE << "i_s = " << i_s << endline_flush;
    for (int i_d = 0; i_d < dddat.size(); i_d++){
      logger << LOG_DEBUG_VERBOSE << "filename_dddistribution_CV[" << i_s << "][" << i_d << "] = " << filename_dddistribution_CV[i_s][i_d] << endline_flush;
      out_dddistribution.open(filename_dddistribution_CV[i_s][i_d].c_str(), ofstream::out | ofstream::trunc);
      out_dddistribution << psi->i_gen << endline;
      for (int i_b = 0; i_b < dddat[i_d].n_bins; i_b++){
	out_dddistribution << bin_counts[dat.size() + i_d][i_b] << endline;
	out_dddistribution << setprecision(15) << unit_factor_calculation * bin_weight_CV[i_s][dat.size() + i_d][i_b] << endline;
	out_dddistribution << setprecision(15) << unit_factor2_calculation * bin_weight2_CV[i_s][dat.size() + i_d][i_b] << endline;
      }
      out_dddistribution.close();
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_distribution_TSV(){
  Logger logger("observable_set::output_distribution_TSV");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ofstream out_distribution;
  for (int i_s = 0; i_s < n_extended_set_TSV; i_s++){
    if (switch_distribution_TSV[i_s] > 0){
      out_distribution.open((filename_distribution_TSV[i_s]).c_str(), ofstream::out | ofstream::trunc);
      // Replaced 'psi->i_gen' by 'n_ps * psi->i_gen' !!!
      //      out_distribution << setw(12) << left << n_ps * psi->i_gen << endline;
      out_distribution << setw(12) << left << psi->i_gen << endline;
      out_distribution << setw(12) << left << n_ps << endline;
       int next_no_qTcut = 0;
      for (int i_q = 0; i_q < output_n_qTcut; i_q++){
	// output only for values selected by 'selection_qTcut_distribution':
	if (i_q != no_qTcut_distribution[next_no_qTcut]){continue;}
	logger << LOG_DEBUG << "i_q = " << i_q << "   no_qTcut_distribution[next_no_qTcut = " << next_no_qTcut << "] = " << no_qTcut_distribution[next_no_qTcut] << endline_flush;

	if (active_qTcut == 0){out_distribution << "#   no qTcut" << endline;}
	else {out_distribution << "#   qTcut[" << setw(3) << i_q << "] = " << setw(15) << setprecision(8) << value_qTcut[i_q] << endline;}
	for (int i_d = 0; i_d < dat.size(); i_d++){
	  out_distribution << "#   dat.name[" << setw(3) << i_d << "] = " << dat[i_d].xdistribution_name << endline;
	  for (int i_r = 0; i_r < n_scale_ren_TSV[i_s]; i_r++){
	    for (int i_f = 0; i_f < n_scale_fact_TSV[i_s]; i_f++){
	      out_distribution << "#   ";
	      out_distribution << "relative_scale_ren[" << setw(3) << i_r << "] = " << setw(15) << setprecision(8) << relative_scale_ren_TSV[i_s][i_r];
	      out_distribution << "   relative_scale_fact[" << setw(3) << i_f << "] = " << setw(15) << setprecision(8) << relative_scale_fact_TSV[i_s][i_f];
	      out_distribution << endline;
	      for (int i_b = 0; i_b < dat[i_d].n_bins; i_b++){
		out_distribution << setw(12) << bin_count_TSV[i_q][i_d][i_b] << "  ";
		out_distribution << setw(24) << setprecision(15) << unit_factor_calculation * bin_weight_TSV[i_s][i_q][i_r][i_f][i_d][i_b] << "  ";
		out_distribution << setw(24) << setprecision(15) << unit_factor2_calculation * bin_weight2_TSV[i_s][i_q][i_r][i_f][i_d][i_b];
		out_distribution << endline;
	      }
	    }
	  }
	}

	for (int i_d = 0; i_d < dddat.size(); i_d++){
	  int i_ddd = dat.size() + i_d;
	  out_distribution << "#   dddat.name[" << setw(3) << i_d << " - " << setw(3) << i_ddd << "] = " << dddat[i_d].name << endline;
	  for (int i_r = 0; i_r < n_scale_ren_TSV[i_s]; i_r++){
	    for (int i_f = 0; i_f < n_scale_fact_TSV[i_s]; i_f++){
	      out_distribution << "#   ";
	      out_distribution << "relative_scale_ren[" << setw(3) << i_r << "] = " << setw(15) << setprecision(8) << relative_scale_ren_TSV[i_s][i_r];
	      out_distribution << "   relative_scale_fact[" << setw(3) << i_f << "] = " << setw(15) << setprecision(8) << relative_scale_fact_TSV[i_s][i_f];
	      out_distribution << endline;
	      for (int i_b = 0; i_b < dddat[i_d].n_bins; i_b++){
		out_distribution << setw(12) << bin_count_TSV[i_q][i_ddd][i_b] << "  ";
		out_distribution << setw(24) << setprecision(15) << unit_factor_calculation * bin_weight_TSV[i_s][i_q][i_r][i_f][i_ddd][i_b] << "  ";
		out_distribution << setw(24) << setprecision(15) << unit_factor2_calculation * bin_weight2_TSV[i_s][i_q][i_r][i_f][i_ddd][i_b];
		out_distribution << endline;
	      }
	    }
	  }
	}
	if (next_no_qTcut < no_qTcut_distribution.size() - 1){next_no_qTcut++;}
      }
      out_distribution.close();
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::output_testpoint_input(){
  Logger logger("observable_set::output_testpoint_input");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  vector<int> list_external_mass;
  for (int i_p = 1; i_p < csi->type_parton[0].size(); i_p++){
    if (csi->mass_parton[0][i_p] != 0.){
      int flag = -1;
      for (int i_l = 0; i_l < list_external_mass.size(); i_l++){
	if (abs(csi->type_parton[0][i_p]) == list_external_mass[i_l]){flag = i_l; break;}
      }
      if (flag == -1){list_external_mass.push_back(abs(csi->type_parton[0][i_p]));}
    }
  }

  int nd = 0;
  for (int i_a = 0; i_a < esi->cut_ps.size(); i_a++){if (esi->cut_ps[i_a] > -1){nd++;}}
  nd = esi->cut_ps.size(); // would print every phase-space point
  if (nd == esi->cut_ps.size()){
    stringstream temp;
    temp << "Print testpoint" << endline;
    temp << endline;
    
    //    int tab_txt = 1;
    if (list_external_mass.size() > 0){
      temp << "if (";
      for (int i_l = 0; i_l < list_external_mass.size(); i_l++){
	temp << "msi->M[" << list_external_mass[i_l] << "] == " << setprecision(15) << psi->M[list_external_mass[i_l]];
	if (i_l < list_external_mass.size() - 1){temp << " && ";}
      }
      temp << "){" << endline;
    }
    else {temp << "{" << endline;}
    // Write x_psf output for all contributions:
    ///    if (csi->class_contribution_collinear){
      for (int j = 0; j < 3; j++){
	temp << "  psi->x_pdf[" << j << "] = " << right << setw(23) << setprecision(15) << psi->x_pdf[j] << ";" << endline;
      }
      temp << endline;
    if (csi->class_contribution_collinear){
      for (int j = 1; j < 3; j++){
	temp << "  psi->z_coll[" << j << "] = " << right << setw(23) << setprecision(15) << psi->z_coll[j] << ";" << endline;
      }
      temp << endline;
    }

    for (int i_p = 0; i_p < esi->p_parton[0].size(); i_p++){
      temp << "  p_parton[0][" << i_p << "] = fourvector(" << right << setw(23) << setprecision(15) << esi->p_parton[0][i_p].x0() << "," << right << setw(23) << setprecision(15) << esi->p_parton[0][i_p].x1() << "," << right << setw(23) << setprecision(15) << esi->p_parton[0][i_p].x2() << "," << right << setw(23) << setprecision(15) << esi->p_parton[0][i_p].x3() << ");" << endline;
    }
    temp << "}" << endline;
    temp << endline;
    logger << LOG_DEBUG << temp.str() << endline_flush;
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


