#include "header.hpp"

void observable_set::set_testpoint_psp(){
  static Logger logger("observable_set::set_testpoint_psp");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  // Check what is really required !!!
  psi->xbp_all = psi->start_xbp_all;
  psi->xbs_all = psi->start_xbs_all;
  psi->xbsqrts_all = psi->start_xbsqrts_all;
  psi->boost = (psi->x_pdf[1] - psi->x_pdf[2]) / (psi->x_pdf[1] + psi->x_pdf[2]);
  psi->xbs_all[0][0] = esi->p_parton[0][0].m2();
  psi->xbsqrts_all[0][0] = esi->p_parton[0][0].m();

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::set_testpoint_psp_dipoles(){
  static Logger logger("observable_set::set_testpoint_psp_dipoles");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (int i_p = 0; i_p < esi->p_parton[0].size(); i_p++){psi->xbp_all[0][intpow(2, i_p - 1)] = esi->p_parton[0][i_p];}
  psi->determine_dipole_phasespace_RA();
  for (int i_a = 1; i_a < n_ps; i_a++){
    for (int i_p = 0; i_p < esi->p_parton[i_a].size(); i_p++){
      esi->p_parton[i_a][i_p] = psi->xbp_all[i_a][intpow(2, i_p - 1)];
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::set_testpoint_x_pdf(){
  static Logger logger("observable_set::set_testpoint_x_pdf");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  if (psi->x_pdf[1] == 0. || psi->x_pdf[2] == 0.){
    psi->x_pdf[0] = pow(esi->p_parton[0][0].x0() / E_CMS, 2);
    psi->x_pdf[1] = pow(psi->x_pdf[0], 0.6);
    psi->x_pdf[2] = psi->x_pdf[0] / psi->x_pdf[1];
  }
  psi->boost = (psi->x_pdf[1] - psi->x_pdf[2]) / (psi->x_pdf[1] + psi->x_pdf[2]);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::set_testpoint_g_z_coll(){
  static Logger logger("observable_set::set_testpoint_g_z_coll");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  psi->g_z_coll[0] = 1.;
  if (psi->g_z_coll[1] == 0. && psi->g_z_coll[2] == 0.){
    //	for (int i_z = 1; i_z < 3; i_z++){psi->g_z_coll[i_z] = 1. / (1. - psi->x_pdf[i_z]);}
    for (int i_z = 1; i_z < 3; i_z++){psi->g_z_coll[i_z] = 1.;}
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::set_testpoint_scale_complete(){
  static Logger logger("observable_set::calculate_dynamic_scale_complete");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  if (esi->cut_ps[0] == -1){
    //    out_cut << "Phase-space 0 is cut. -> Default scale is used." << endline;
    //	esi->cut_ps[0] = 0;
    set_testpoint_scale();
    set_testpoint_scale_TSV(0);
  }
  else {
    // Probably not implemented like this ...
    // 1 - Testpoint is calculated at basic fixed scale (prefactor * scale_ren).
    // 2 - Testpoint is calculated at output scale.
    calculate_dynamic_scale(0);
    calculate_dynamic_scale_TSV(0);
  }
  determine_scale();
  //  if (csi->class_contribution_CS_collinear){determine_scale_CA();}
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::set_testpoint_scale_complete_RA(){
  static Logger logger("observable_set::calculate_dynamic_scale_complete_RA");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (int i_a = 0; i_a < n_ps; i_a++){
    if (esi->cut_ps[i_a] == -1){
      //      out_cut << "Phase-space " << i_a << " is cut. -> Default scale is used." << endline;
      //      esi->cut_ps[i_a] = 0;
      set_testpoint_scale_RA(i_a);
      set_testpoint_scale_TSV(i_a);
    }
    else {
      calculate_dynamic_scale_RA(i_a);
      calculate_dynamic_scale_TSV(i_a);
    }
    determine_scale_RA(i_a);
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::set_testpoint_scale(){
  static Logger logger("observable_set::set_testpoint_scale");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (int sd = 1; sd < value_mu_ren_rel.size(); sd++){
    if (value_mu_ren_rel[sd].size() != 0){
      double temp_mu_central = scale_ren / prefactor_reference;
      //      double temp_mu_central = scale_ren;
      for (int ss = 0; ss < value_mu_ren_rel[sd].size(); ss++){
        value_mu_ren[sd][ss] = temp_mu_central * value_mu_ren_rel[sd][ss];
        value_alpha_S[sd][ss] = LHAPDF::alphasPDF(value_mu_ren[sd][ss]);
        value_factor_alpha_S[sd][ss] = pow(value_alpha_S[sd][ss] / alpha_S, csi->contribution_order_alpha_s);
      }
    }
  }
  if (id_scales == 1){
    value_mu_fact = value_mu_ren;
  }
  else {
    for (int sd = 1; sd < value_mu_fact_rel.size(); sd++){
      if (value_mu_fact_rel[sd].size() != 0){
	double temp_mu_central = scale_fact / prefactor_reference;
	//	double temp_mu_central = scale_fact;
        for (int ss = 0; ss < value_mu_fact_rel[sd].size(); ss++){
          value_mu_fact[sd][ss] = temp_mu_central * value_mu_fact_rel[sd][ss];
        }
      }
    }
  }
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::set_testpoint_scale_RA(int i_a){
  static Logger logger("observable_set::set_testpoint_scale_RA");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  for (int sd = 1; sd < value_mu_ren_rel.size(); sd++){
    if (value_mu_ren_rel[sd].size() != 0){
      double temp_mu_central = scale_ren / prefactor_reference;
      //      double temp_mu_central = scale_ren;
      for (int ss = 0; ss < value_mu_ren_rel[sd].size(); ss++){
        RA_value_mu_ren[i_a][sd][ss] = temp_mu_central * value_mu_ren_rel[sd][ss];
        RA_value_alpha_S[i_a][sd][ss] = LHAPDF::alphasPDF(RA_value_mu_ren[i_a][sd][ss]);
        RA_value_factor_alpha_S[i_a][sd][ss] = pow(RA_value_alpha_S[i_a][sd][ss] / alpha_S, csi->contribution_order_alpha_s);
      }
    }
  }
  if (id_scales == 1){
    RA_value_mu_fact[i_a] = RA_value_mu_ren[i_a];
  }
  else {
    for (int sd = 1; sd < value_mu_fact_rel.size(); sd++){
      if (value_mu_fact_rel[sd].size() != 0){
        double temp_mu_central = scale_fact / prefactor_reference;
	//        double temp_mu_central = scale_fact;
        for (int ss = 0; ss < value_mu_fact_rel[sd].size(); ss++){
          RA_value_mu_fact[i_a][sd][ss] = temp_mu_central * value_mu_fact_rel[sd][ss];
        }
      }
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void observable_set::set_testpoint_scale_TSV(int i_a){
  static Logger logger("observable_set::set_testpoint_scale_TSV");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;
  for (int sd = 1; sd < max_dyn_ren + 1; sd++){
    double temp_mu_central = scale_ren / prefactor_reference;
    //    double temp_mu_central = scale_ren;
    for (int ss = 0; ss < n_scale_dyn_ren[sd]; ss++){
      value_scale_ren[i_a][sd][ss] = temp_mu_central * value_relative_scale_ren[sd][ss];
      if (needed_scale2_ren){value_scale2_ren[i_a][sd][ss] = pow(value_scale_ren[i_a][sd][ss], 2);}
      value_alpha_S_TSV[i_a][sd][ss] = LHAPDF::alphasPDF(value_scale_ren[i_a][sd][ss]);
      value_relative_factor_alpha_S[i_a][sd][ss] = pow(value_alpha_S_TSV[i_a][sd][ss] / alpha_S, csi->contribution_order_alpha_s);
    }
  }
  for (int sd = 1; sd < max_dyn_fact + 1; sd++){
    double temp_mu_central = scale_fact / prefactor_reference;
    //    double temp_mu_central = scale_fact;
    value_central_scale_fact[sd] = temp_mu_central;
    if (needed_scale2_fact){value_central_logscale2_fact[sd] = 2 * log(temp_mu_central);}
    for (int ss = 0; ss < n_scale_dyn_fact[sd]; ss++){
      value_scale_fact[i_a][sd][ss] = value_central_scale_fact[sd] * value_relative_scale_fact[sd][ss];
      if (needed_scale2_fact){value_scale2_fact[i_a][sd][ss] = pow(value_scale_fact[i_a][sd][ss], 2);}
    }
  }
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


string observable_set::output_momenta(){
  Logger logger("observable_set::output_momenta");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  stringstream out_comparison;
  out_comparison << endline << "Particle momenta:" << endline << endline;
  for (int i_a = 0; i_a < n_ps; i_a++){
    if (csi->class_contribution_CS_real){
      out_comparison << "Phase space " << setw(2) << i_a << ":   " << setw(12) << csi->dipole[i_a].name() << endline << endline;
    }
    for (int ib = 1; ib < esi->p_parton[0].size(); ib++){
      out_comparison << "p[" << setw(3) << csi->type_parton[0][ib] << "] = " << esi->p_parton[0][ib] << "   " << sqrt(abs(esi->p_parton[0][ib].m2())) << endline;
    }
    out_comparison << endline;
    fourvector check;
    for (int ib = 1; ib < esi->p_parton[0].size(); ib++){
      if (ib < 3){check = check + esi->p_parton[0][ib];}
      else {check = check - esi->p_parton[0][ib];}
    }
    if (sqrt(abs(check.m2() / esi->p_parton[0][0].m2())) > 1.e-14){
      out_comparison << "momentum conservation violation = " << check << "   " << sqrt(abs(check.m2() / esi->p_parton[0][0].m2())) << endline;
      out_comparison << endline;
    }
  }

  if (csi->class_contribution_collinear){
    out_comparison << endline << "Momentum fractions (for pdfs): " << endline << endline;
    for (int i_x = 1; i_x < 3; i_x++){
      out_comparison << "x_pdf[" << i_x << "] = " << setw(23) << showpoint << setprecision(15) << psi->x_pdf[i_x] << endline;
    }
    out_comparison << endline;
    for (int i_x = 1; i_x < 3; i_x++){
      out_comparison << "x_pdf[" << i_x << "] / z_coll[" << i_x << "] = " << setw(23) << showpoint << setprecision(15) << psi->x_pdf[i_x] / psi->z_coll[i_x] << endline;
    }

    out_comparison << endline << "Collinear radiation fractions: " << endline << endline;
    for (int i_z = 1; i_z < 3; i_z++){
      out_comparison << "z_coll[" << i_z << "] = " << setw(23) << showpoint << setprecision(15) << psi->z_coll[i_z] << endline;
    }
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished - return (string):" << endline << out_comparison.str() << endline_flush;
  return out_comparison.str();
}


string observable_set::output_scales(){
  Logger logger("observable_set::output_scales");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  stringstream out_comparison;

  if (switch_CV){  
  out_comparison << endline << "Renormalization and factorization scales (CV variation):" << endline << endline;
  //    out_comparison << "CV variation" << endline << endline;
    /*
    for (int i_vr = 0; i_vr < value_mu_ren.size(); i_vr++){
      for (int i_mr = 0; i_mr < value_mu_ren[i_vr].size(); i_mr++){
	out_comparison << "µR[" << i_vr << "][" << i_mr << "] = " << value_mu_ren[i_vr][i_mr] << endline;
      }
    }
    out_comparison << endline;
    for (int i_vf = 0; i_vf < value_mu_fact.size(); i_vf++){
      for (int i_mf = 0; i_mf < value_mu_fact[i_vf].size(); i_mf++){
	out_comparison << "µF[" << i_vf << "][" << i_mf << "] = " << value_mu_fact[i_vf][i_mf] << endline;
      }
    }
    out_comparison << endline;
    */
    //    out_comparison << "dynamic_scale = " << dynamic_scale << "   map_value_scale_ren = " << map_value_scale_ren << "   map_value_scale_fact = " << map_value_scale_fact << endline;
    out_comparison << "µF_ref    = µF[" << dynamic_scale << "][" << map_value_scale_fact << "] = " << left << setw(23) << setprecision(15) << value_mu_fact[dynamic_scale][map_value_scale_fact] << "   "
		   << "µR_ref    = µR[" << dynamic_scale << "][" << map_value_scale_ren << "] = " << left << setw(23) << setprecision(15) << value_mu_ren[dynamic_scale][map_value_scale_fact] << endline;
    out_comparison << endline;
    for (int i_s = 0; i_s < n_scales_CV; i_s++){
      //    out_comparison << "dynamic_scale_CV = " << dynamic_scale_CV << "   map_value_scale_ren_CV[" << i_s << "] = " << map_value_scale_ren_CV[i_s] << "   map_value_scale_fact_CV[" << i_s << "] = " << map_value_scale_fact_CV[i_s] << endline;
      out_comparison << "µF_CV[" << setw(2) << right << i_s << "] = µF[" << dynamic_scale_CV << "][" << map_value_scale_fact_CV[i_s] << "] = " << left << setw(23) << setprecision(15) << value_mu_fact[dynamic_scale_CV][map_value_scale_fact_CV[i_s]] << "   "
		     << "µR_CV[" << setw(2) << right << i_s << "] = µR[" << dynamic_scale_CV << "][" << map_value_scale_ren_CV[i_s] << "] = " << left << setw(23) << setprecision(15) << value_mu_ren[dynamic_scale_CV][map_value_scale_ren_CV[i_s]] << endline;
    }
    out_comparison << endline;
    
  }
  if (switch_TSV){
    out_comparison << "Renormalization and factorization scales (TSV variation):" << endline << endline;
    for (int i_a = 0; i_a < n_ps; i_a++){
      if (csi->class_contribution_CS_real){
	if (i_a > 0){out_comparison << endline;}
	out_comparison << left << "Phase space " << setw(2) << i_a << ":   " << setw(12) << csi->dipole[i_a].name();
	if (esi->cut_ps[i_a] == -1){
	  out_comparison << "is cut. -> Fixed default scales µR and µF are used.";
	}
	out_comparison << endline << endline;
      }
      else {
	if (esi->cut_ps[0] == -1){
	  out_comparison << "Phase-space 0 is cut. -> Fixed default scales µR and µF are used." << endline;
	  out_comparison << endline;
	}
      }
          
      for (int i_s = 0; i_s < n_set_TSV; i_s++){
	for (int i_r = 0; i_r < n_scale_ren_TSV[i_s]; i_r++){
	  if (switch_output_testpoint_reference_only && (i_s != no_reference_TSV || i_r != no_scale_ren_reference_TSV)){continue;}
	  out_comparison << left << setw(25+1) << name_scale_ren_TSV[i_s][i_r] << " = " << setw(23) << setprecision(15) << *(pointer_scale_ren)[i_a][i_s][i_r] << "   " << "alpha_S(µR) = " << setw(23) << setprecision(15) << LHAPDF::alphasPDF(*(pointer_scale_ren)[i_a][i_s][i_r]) << endline;
	}
      }
      if (!switch_output_testpoint_reference_only){out_comparison << endline;}
      for (int i_s = 0; i_s < n_set_TSV; i_s++){
	for (int i_f = 0; i_f < n_scale_fact_TSV[i_s]; i_f++){
	  if (switch_output_testpoint_reference_only && (i_s != no_reference_TSV || i_f != no_scale_fact_reference_TSV)){continue;}
	  out_comparison << left << setw(25+1) << name_scale_fact_TSV[i_s][i_f] << " = " << left << setw(23) << setprecision(15) << *(pointer_scale_fact)[i_a][i_s][i_f] << endline;
	}
      }
    }
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished - return (string):" << endline << out_comparison.str() << endline_flush;
  return out_comparison.str();
}


string observable_set::output_momentum_conservation_check(int x_a){
  Logger logger("observable_set::output_momentum_conservation_check");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  stringstream out_comparison;
  fourvector check;
  for (int i_p = 1; i_p < esi->p_parton[x_a].size(); i_p++){
    if (i_p < 3){check = check + esi->p_parton[x_a][i_p];}
    else {check = check - esi->p_parton[x_a][i_p];}
  }

  double check_criterion = 0.;
  if (abs(pow(check.x0(), 2) / esi->p_parton[x_a][0].m2()) > check_criterion){check_criterion = abs(pow(check.x0(), 2) / esi->p_parton[x_a][0].m2());}
  if (abs(pow(check.x1(), 2) / esi->p_parton[x_a][0].m2()) > check_criterion){check_criterion = abs(pow(check.x1(), 2) / esi->p_parton[x_a][0].m2());}
  if (abs(pow(check.x2(), 2) / esi->p_parton[x_a][0].m2()) > check_criterion){check_criterion = abs(pow(check.x2(), 2) / esi->p_parton[x_a][0].m2());}
  if (abs(pow(check.x3(), 2) / esi->p_parton[x_a][0].m2()) > check_criterion){check_criterion = abs(pow(check.x3(), 2) / esi->p_parton[x_a][0].m2());}
  if (abs(check.m2() / esi->p_parton[x_a][0].m2()) > check_criterion){check_criterion = abs(check.m2() / esi->p_parton[x_a][0].m2());}
  
  if (check_criterion > 1.e-14){
    out_comparison << "momentum conservation check = (" <<
      setprecision(15) << setw(19) << check.x0() << "; " <<
      setprecision(15) << setw(19) << check.x1() << ", " <<
      setprecision(15) << setw(19) << check.x2() << ", " <<
      setprecision(15) << setw(19) << check.x3() << ")" <<
      "   " << sqrt(abs(check.m2())) << endline;
    out_comparison << "sqrt(abs((p(in) - p(out))^2 / p(in)^2)) = " << sqrt(abs(check.m2() / esi->p_parton[x_a][0].m2())) << endline;
    out_comparison << endline;
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished - return" << endline_flush;
  return out_comparison.str();
}


string observable_set::output_momenta_phasespace(int x_a){
  Logger logger("observable_set::output_momenta");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  stringstream out_comparison;
  for (int i_p = 1; i_p < esi->p_parton[x_a].size(); i_p++){
    out_comparison << "   p[" << setw(3) << csi->type_parton[x_a][i_p] << "] = (" <<
      setprecision(15) << setw(19) << esi->p_parton[x_a][i_p].x0() << "; " <<
      setprecision(15) << setw(19) << esi->p_parton[x_a][i_p].x1() << ", " <<
      setprecision(15) << setw(19) << esi->p_parton[x_a][i_p].x2() << ", " <<
      setprecision(15) << setw(19) << esi->p_parton[x_a][i_p].x3() << ")" <<
      "   " << sqrt(abs(esi->p_parton[x_a][i_p].m2())) << endline;
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished - return" << endline_flush;
  return out_comparison.str();
}


string observable_set::output_testpoint_kinematics(){
  Logger logger("observable_set::output_testpoint_kinematics");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  stringstream out_comparison;
  out_comparison << "Particle momenta in CMS frame: " << endline << endline;
  for (int i_a = 0; i_a < n_ps; i_a++){
    if (csi->class_contribution_CS_real){
      out_comparison << "Phase space " << setw(2) << i_a << ":   " << setw(12) << csi->dipole[i_a].name() << endline << endline;
    }
    out_comparison << output_momenta_phasespace(i_a) << endline;
    out_comparison << output_momentum_conservation_check(i_a);
  }
  
  vector<vector<fourvector> > temp_p_CMS = esi->p_parton;
  for (int i_p = 1; i_p < esi->p_parton[0].size(); i_p++){esi->p_parton[0][i_p] = esi->p_parton[0][i_p].zboost(-psi->boost);}
  out_comparison << "Particle momenta in LAB frame: " << endline << endline;
  out_comparison << output_momenta_phasespace(0) << endline;
  out_comparison << "Momentum fractions (for pdfs): " << endline << endline;
  for (int i_x = 1; i_x < 3; i_x++){
    out_comparison << "x_pdf[" << i_x << "] = " << setw(23) << showpoint << setprecision(15) << psi->x_pdf[i_x] << endline;
  }
  out_comparison << endline;
  //  out_comparison << setprecision(15) << setw(19) << "psi->x_pdf[1] = " << psi->x_pdf[1] << endline;
  //  out_comparison << setprecision(15) << setw(19) << "psi->x_pdf[2] = " << psi->x_pdf[2] << endline;
  //  out_comparison << setprecision(15) << setw(19) << "psi->boost    = " << psi->boost << endline;
  if (csi->class_contribution_collinear){
    for (int i_x = 1; i_x < 3; i_x++){
      out_comparison << "x_pdf[" << i_x << "] / z_coll[" << i_x << "] = " << setw(23) << showpoint << setprecision(15) << psi->x_pdf[i_x] / psi->z_coll[i_x] << endline;
    }

    out_comparison << endline << "Collinear radiation fractions: " << endline << endline;
    for (int i_z = 1; i_z < 3; i_z++){
      //      out_comparison << "z_coll[" << i_z << "] = " << setw(23) << showpoint << setprecision(15) << psi->z_coll[i_z] << endline;
      out_comparison << "z_coll[" << i_z << "] = " << setw(23) << showpoint << setprecision(15) << psi->z_coll[i_z] << "   " << "g_z_coll[" << i_z << "] = " << setw(23) << showpoint << setprecision(15) << psi->g_z_coll[i_z] << endline;
    }
    out_comparison << endline;
  }
  
  esi->p_parton = temp_p_CMS;

  logger << LOG_DEBUG_VERBOSE << "finished - return" << endline_flush;
  return out_comparison.str();
}


string observable_set::output_testpoint_integrand(){
  Logger logger("observable_set::output_testpoint_integrand");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  stringstream out_comparison;
  out_comparison << "Contribution to cross section (without phase space factors):" << endline;
  out_comparison << endline;
  for (int i_s = 0; i_s < n_set_TSV; i_s++){
    for (int i_r = 0; i_r < n_scale_ren_TSV[i_s]; i_r++){
      for (int i_f = 0; i_f < n_scale_fact_TSV[i_s]; i_f++){
	if (switch_output_testpoint_reference_only && (i_s != no_reference_TSV || i_r != no_scale_ren_reference_TSV || i_f != no_scale_fact_reference_TSV)){continue;}
	for (int i_h = 0; i_h < max_n_integrand_TSV[i_s]; i_h++){
	  if (csi->class_contribution_CS_collinear){
	    vector<double> temp_c(n_pc, 0.);
	    for (int i_c = 0; i_c < n_pc; i_c++){
	      for (int i_z = 0; i_z < n_pz; i_z++){
		if (i_z == 0 && (*CA_collinear)[i_c][0].in_collinear()[i_z] == 1){
		  temp_c[i_c] += *(pointer_pdf_factor)[i_c][i_z][i_s][i_f][i_h]
		    * (  *(pointer_ME2term)[i_c][1][i_s][i_r][i_f] / psi->g_z_coll[(*CA_collinear)[i_c][0].no_emitter()]
			 + *(pointer_ME2term)[i_c][2][i_s][i_r][i_f]);
		}
		else if ((*CA_collinear)[i_c][0].in_collinear()[i_z] == 1){ // always i_z == dipole_phasespace[i_c][4] !!!
		  temp_c[i_c] += *(pointer_pdf_factor)[i_c][i_z][i_s][i_f][i_h]
		    * *(pointer_ME2term)[i_c][0][i_s][i_r][i_f] / psi->g_z_coll[i_z];
		}
	      }
	    }
	    if (i_h == 0){out_comparison << "integrand (scaleset " << i_s << ", µR " << i_r << ", µF " << i_f << ") = " << setprecision(15) << setw(19) << accumulate(temp_c.begin(), temp_c.end(), 0.) * *(pointer_relative_factor_alpha_S)[0][i_s][i_r] << endline;
	    }
	  }
	  else {
	    if (i_h == 0){out_comparison << "integrand (scaleset " << i_s << ", µR " << i_r << ", µF " << i_f << ") = " << setprecision(15) << setw(19) << *(pointer_ME2term)[0][0][i_s][i_r][i_f]
		* *(pointer_pdf_factor)[0][0][i_s][i_f][i_h]
		* *(pointer_relative_factor_alpha_S)[0][i_s][i_r] << endline;
	    }
	  }
	}
      }
    }
  }

  logger << LOG_DEBUG_VERBOSE << "finished - return" << endline_flush;
  return out_comparison.str();
}

