#include "header.hpp"

summary_subprocess::summary_subprocess(){
  Logger logger("summary_subprocess::summary_subprocess ()");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  name = "";
  error2_time = 0.;
  time_per_event = 0.;
  used_runtime = 0.;
  used_n_event = 0;
  extrapolated_runtime = 0.;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


summary_subprocess::summary_subprocess(summary_generic * _ygeneric){
  Logger logger("summary_subprocess::summary_subprocess (ygeneric)");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ygeneric = _ygeneric;
  osi = ygeneric->osi;
  name = "";

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


summary_subprocess::summary_subprocess(string _name, summary_contribution * _contribution){
  Logger logger("summary_subprocess::summary_subprocess (name, contribution)");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ycontribution = _contribution;
  ygeneric = ycontribution->ygeneric;
  osi = ygeneric->osi;

  name = _name;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


summary_subprocess::summary_subprocess(string _name, int _n_seed, int _default_size_CV, summary_contribution * _contribution){
  Logger logger("summary_subprocess::summary_subprocess (name, n_seed, default_size_CV, contribution)");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  initialization(_name, _n_seed, _default_size_CV, _contribution);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void summary_subprocess::initialization(string _name, int _n_seed, int _default_size_CV, summary_contribution * _contribution){
  Logger logger("summary_subprocess::initialization");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ycontribution = _contribution;
  ygeneric = ycontribution->ygeneric;
  osi = ygeneric->osi;

  name = _name;
  n_seed = _n_seed;
  default_size_CV = _default_size_CV;

  no_results = 0;
  N_all = 0;

  result = 0.;
  deviation = 0.;
  chi2 = 0.;

  logger << LOG_DEBUG << "initialization:  " << setw(20) << name << "  in  " << setw(15) << ycontribution->type_contribution << endline_flush;

  remove_run_qTcut_CV.resize(osi->n_qTcut, vector<int> (n_seed, 0));

  // Check if needed beforehand !!!

  N.resize(n_seed, 0);
  result_run.resize(n_seed, 0.);
  deviation_run.resize(n_seed, 0.);
  // might be needed for CV !!!
  used_n_event_run.resize(n_seed, 0);

  N_event_run_CV.resize(n_seed, 0);
  N_event_CV = 0;
  Nd_event_CV = 0.;
  N_valid_event_CV.resize(osi->n_qTcut, 0);

  n_parallel_runs_reference_CV.resize(osi->n_qTcut, 0);
  n_valid_parallel_runs_reference_CV.resize(osi->n_qTcut, 0);

  result_CV.resize(osi->n_qTcut, vector<double> (osi->n_scales_CV, 0.));
  deviation_CV.resize(osi->n_qTcut, vector<double> (osi->n_scales_CV, 0.));
  chi2_CV.resize(osi->n_qTcut, vector<double> (osi->n_scales_CV, 0.));
  n_parallel_runs_CV.resize(osi->n_qTcut, vector<int> (osi->n_scales_CV, 0));
  n_valid_parallel_runs_CV.resize(osi->n_qTcut, vector<int> (osi->n_scales_CV, 0));

  result_run_CV.resize(ycontribution->output_n_qTcut, vector<vector<double> > (osi->n_scales_CV, vector<double> (n_seed, 0.)));
  deviation_run_CV.resize(ycontribution->output_n_qTcut, vector<vector<double> > (osi->n_scales_CV, vector<double> (n_seed, 0.)));
  
  if (osi->switch_TSV){
    result_TSV.resize(osi->n_moments, vector<vector<vector<vector<double> > > > (osi->n_qTcut, vector<vector<vector<double> > > (osi->n_extended_set_TSV)));
    for (int i_m = 0; i_m < osi->n_moments; i_m++){
      for (int i_q = 0; i_q < osi->n_qTcut; i_q++){
	for (int i_s = 0; i_s < osi->n_extended_set_TSV; i_s++){
	  result_TSV[i_m][i_q][i_s].resize(osi->n_scale_ren_TSV[i_s], vector<double> (osi->n_scale_fact_TSV[i_s], 0.));
	}
      }
    }
    deviation_TSV = result_TSV;
    chi2_TSV = result_TSV;

    result_group_TSV.resize(ycontribution->extended_directory.size());
    result_group_run_TSV.resize(ycontribution->extended_directory.size());
    deviation_group_run_TSV.resize(ycontribution->extended_directory.size());
    N_event_group_run_TSV.resize(ycontribution->extended_directory.size());
    for (int i_o = 0; i_o < ycontribution->extended_directory.size(); i_o++){
      result_group_TSV[i_o].resize(osi->n_moments);
      for (int i_m = 0; i_m < osi->n_moments; i_m++){
	result_group_TSV[i_o][i_m].resize(ycontribution->output_n_qTcut, vector<vector<vector<double> > > (osi->n_extended_set_TSV));
	for (int i_s = 0; i_s < osi->n_extended_set_TSV; i_s++){
	  if (!ygeneric->switch_output_scaleset[i_s]){continue;}
	  for (int i_q = 0; i_q < ycontribution->output_n_qTcut; i_q++){
	    result_group_TSV[i_o][i_m][i_q][i_s].resize(osi->n_scale_ren_TSV[i_s], vector<double> (osi->n_scale_fact_TSV[i_s], 0.));
	  }
	}
      }
      result_group_run_TSV[i_o].resize(ycontribution->extended_directory[i_o].size(), result_group_TSV[i_o]);
      N_event_group_run_TSV[i_o].resize(ycontribution->extended_directory[i_o].size(), 0);
    }
    deviation_group_TSV = result_group_TSV;
    chi2_group_TSV = result_group_TSV;
    deviation_group_run_TSV = result_group_run_TSV;

    N_event_group_TSV.resize(ycontribution->extended_directory.size(), 0);
    Nd_event_group_TSV.resize(ycontribution->extended_directory.size(), 0.);
    N_valid_event_group_TSV.resize(ycontribution->extended_directory.size(), vector<long long> (ycontribution->output_n_qTcut, 0));
  
    n_parallel_runs_group_reference_TSV.resize(ycontribution->extended_directory.size(), vector<int> (ycontribution->output_n_qTcut, 0)); 
    n_valid_parallel_runs_group_reference_TSV = n_parallel_runs_group_reference_TSV;

    remove_group_run_qTcut_TSV.resize(ycontribution->output_n_qTcut, vector<vector<int> > (ycontribution->extended_directory.size()));
    for (int i_q = 0; i_q < ycontribution->output_n_qTcut; i_q++){
      for (int i_o = 0; i_o < ycontribution->extended_directory.size(); i_o++){
	logger << LOG_DEBUG_VERBOSE << "i_o = " << i_o << "   i_q = " << i_q << endline_flush;
	remove_group_run_qTcut_TSV[i_q][i_o].resize(ycontribution->extended_directory[i_o].size(), 0);
      }
    }

    n_parallel_runs_group_TSV.resize(ycontribution->extended_directory.size(), vector<vector<vector<vector<vector<int> > > > > (osi->n_moments, vector<vector<vector<vector<int> > > > (ycontribution->output_n_qTcut, vector<vector<vector<int> > > (osi->n_extended_set_TSV)))); 
    for (int i_o = 0; i_o < ycontribution->extended_directory.size(); i_o++){
      for (int i_m = 0; i_m < osi->n_moments; i_m++){
	for (int i_q = 0; i_q < ycontribution->output_n_qTcut; i_q++){
	  for (int i_s = 0; i_s < osi->n_extended_set_TSV; i_s++){
	    n_parallel_runs_group_TSV[i_o][i_m][i_q][i_s].resize(osi->n_scale_ren_TSV[i_s], vector<int> (osi->n_scale_fact_TSV[i_s], 0));
	  }
	}
      }
    }
    n_valid_parallel_runs_group_TSV = n_parallel_runs_group_TSV;
  }
  
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


// What are these two functions needed for ???
void summary_subprocess::combine_result_original(){
  Logger logger("summary_subprocess::combine_result_original");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  N_all = 0;
  no_results = 0;
  result = 0.;
  deviation = 0.;
  for (int i_z = 0; i_z < n_seed; i_z++){
    if (!(result_run[i_z] == 0. && deviation_run[i_z] == 0.)){
      N_all += N[i_z];
      result += N[i_z] * result_run[i_z];
      deviation += pow((N[i_z] - 1) * deviation_run[i_z], 2) + N[i_z] * pow(result_run[i_z], 2);
      no_results++;
    }
  }
  deviation = sqrt((N_all * deviation - pow(result, 2)) / N_all) / (N_all - 1);
  result = result / N_all;
  for (int i_z = 0; i_z < n_seed; i_z++){
    if (!(result_run[i_z] == 0. && deviation_run[i_z] == 0.)){
      chi2 += pow((result_run[i_z] - result) / deviation_run[i_z], 2.);
    }
  }
  chi2 = chi2 / no_results;
  for (int i_z = 0; i_z < n_seed; i_z++){
    if (N[i_z] != 0 && deviation_run[i_z] != 0. && result_run[i_z] != 0.){
      logger << LOG_DEBUG << "run " << right << setw(2) << i_z << ":   " << showpoint << setw(13) << right << setprecision(6) << result_run[i_z] << " +- " << setw(13) << setprecision(6) << deviation_run[i_z] << "   " << setw(11) << right << setprecision(4) << (result_run[i_z] - result) / deviation_run[i_z] << " sigma   " << setw(15) << N[i_z] << endline_flush;
    }
  }
  logger << LOG_DEBUG << endl << setw(10) << left << "old av:   " << setw(13) << right << setprecision(6) << result << " +- " << setw(13) << setprecision(6) << deviation << endl << endline_flush;

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


// Check what this is used for !!!
void summary_subprocess::combine_result_alternative(){
  Logger logger("summary_subprocess::combine_result_alternative");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ///////////////////////////////////////////////////////////////
  //  combination of different-seed runs for all subprocesses  //
  ///////////////////////////////////////////////////////////////
  double temp_singleweight;

  N_all = 0;
  result = 0.;
  deviation = 0.;
  no_results = 0;
  for (int i_z = 0; i_z < n_seed; i_z++){
    if (N[i_z] != 0 && deviation_run[i_z] != 0. && result_run[i_z] != 0.){
      temp_singleweight = 1. / pow(deviation_run[i_z], 2);
      result += temp_singleweight * result_run[i_z];
      deviation += temp_singleweight;
      N_all += N[i_z];
      no_results++;
    }
  }
  if (N_all != 0){
    result = result / deviation;
    deviation = 1. / sqrt(deviation);
    if (result != result){
      logger << LOG_DEBUG_VERBOSE << N_all << endline_flush;
      for (int i_z = 0; i_z < n_seed; i_z++){logger << LOG_DEBUG_VERBOSE << N[i_z] << endline_flush;}
    }
  }
  else{
    result = 0.;
    deviation = 0.;
  }
  for (int i_z = 0; i_z < n_seed; i_z++){
    if (N[i_z] != 0 && deviation_run[i_z] != 0. && result_run[i_z] != 0.){
      logger << LOG_DEBUG << "run " << right << setw(2) << i_z << ":   " << showpoint << setw(13) << right << setprecision(6) << result_run[i_z] << " +- " << setw(13) << setprecision(6) << deviation_run[i_z] << "   " << setw(11) << right << setprecision(4) << (result_run[i_z] - result) / deviation_run[i_z] << " sigma   " << setw(15) << N[i_z] << " EVENTS" << endline_flush;
    }
  }
  logger << LOG_DEBUG << endl << setw(10) << left << "average:  " << setw(13) << right << setprecision(6) << result << " +- " << setw(13) << setprecision(6) << deviation << endl << endline_flush;
  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}



double summary_subprocess::get_time_stamp(int i_s, int i_m, int i_z){
  Logger logger("summary_subprocess::get_time_stamp");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  ifstream in_time_file(filepath_time[i_s][i_m][i_z].c_str());
  string temp_time_stamp;
  in_time_file.seekg(0);//, is.beg);
  getline(in_time_file, temp_time_stamp);
  in_time_file.close();
  double time_stamp = atof(temp_time_stamp.c_str());

  logger << LOG_DEBUG_VERBOSE << "finished - time_stamp = " << time_stamp << endline_flush;
  return time_stamp;
}


void summary_subprocess::set_file_distribution(int i_s, int i_m, int i_z){
  Logger logger("summary_subprocess::set_file_distribution");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  in_result_vector[i_s][i_m][i_z] = new ifstream(filepath_distribution[i_s][i_m][i_z].c_str());
  char c;
  while (in_result_vector[i_s][i_m][i_z]->get(c)){
    if (c == '\n'){pos_line[i_s][i_m][i_z].push_back(in_result_vector[i_s][i_m][i_z]->tellg());}
  }
  in_result_vector[i_s][i_m][i_z]->clear();

  int temp_counter = 2; // 1 without 'group counter' (old version)
  for (int x_q = 0; x_q < ycontribution->selection_n_qTcut; x_q++){
    temp_counter++;
    for (int i_d = 0; i_d < osi->extended_distribution.size(); i_d++){
      logger << LOG_DEBUG << "set pos_qTcut_distribution:   i_s = " << i_s << "   x_q = " << x_q << "   i_d = " << i_d << endline_flush;
      no_line_distribution[i_s][i_d][x_q][0] = temp_counter;
      temp_counter += (osi->extended_distribution[i_d].n_bins + 1) * osi->n_scale_ren_TSV[i_s] * osi->n_scale_fact_TSV[i_s];

      pos_qTcut_distribution[i_s][i_m][i_z][x_q][i_d] = pos_line[i_s][i_m][i_z][no_line_distribution[i_s][i_d][x_q][0]];

      temp_counter++;
    }
  }

  // pos_line[i_s][i_m][i_z] not needed any longer

  // Probably only for checks (no function ???):
  for (int x_q = 0; x_q < ycontribution->selection_n_qTcut; x_q++){
    for (int i_d = 0; i_d < osi->extended_distribution.size(); i_d++){
      string temp_s;
      in_result_vector[i_s][i_m][i_z]->seekg(pos_qTcut_distribution[i_s][i_m][i_z][x_q][i_d]);//, is.beg);
      getline(*in_result_vector[i_s][i_m][i_z], temp_s);
    }
  }
  in_result_vector[i_s][i_m][i_z]->clear();
  // ... until here.


  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void summary_subprocess::cleanup_file_distribution(){
  Logger logger("summary_subprocess::cleanup_file_distribution");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  vector<vector<vector<int> > > ().swap(index_contributing_run);
  vector<vector<vector<ifstream * > > > ().swap(in_result_vector);
  vector<vector<vector<vector<streampos> > > > ().swap(pos_line);
  vector<vector<vector<string> > > ().swap(filepath_time);
  vector<vector<vector<string> > > ().swap(filepath_distribution);
  vector<vector<vector<double> > > ().swap(time_stamp);
  vector<vector<vector<vector<int> > > > ().swap(no_line_distribution);
  vector<vector<vector<vector<vector<streampos> > > > > ().swap(pos_qTcut_distribution);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void summary_subprocess::check_reset_file_distribution(int i_s, int i_m, int i_z){
  Logger logger("summary_subprocess::check_reset_file_distribution");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  logger << LOG_DEBUG << "Open file  " << filepath_distribution[i_s][i_m][i_z] << endline_flush;

  double check_time_stamp = get_time_stamp(i_s, i_m, i_z);
  if (check_time_stamp == time_stamp[i_s][i_m][i_z]){
    logger << LOG_DEBUG << "Open " << name << " at [" << i_s << "][" << i_m << "][" << i_z << "]  " << filepath_distribution[i_s][i_m][i_z] << "   " << "time_stamp = " << setprecision(15) << setw(23) << time_stamp[i_s][i_m][i_z] << "   successful -> return;" << endline_flush;
    return;
  }

  logger << LOG_WARN << "File has changed : filepath_distribution[" << i_s << "][" << i_m << "][" << i_z << "] = " << filepath_distribution[i_s][i_m][i_z] << "   " << "time_stamp = " << time_stamp[i_s][i_m][i_z] << " =/= " << check_time_stamp << " = check_time_stamp - > needs to be reset!" << endline_flush;

  time_stamp[i_s][i_m][i_z] = check_time_stamp;
  in_result_vector[i_s][i_m][i_z]->close();

  pos_line[i_s][i_m][i_z].clear();

  delete in_result_vector[i_s][i_m][i_z];
  in_result_vector[i_s][i_m][i_z] = new ifstream(filepath_distribution[i_s][i_m][i_z].c_str());
  char c;
  while (in_result_vector[i_s][i_m][i_z]->get(c)){
    if (c == '\n'){pos_line[i_s][i_m][i_z].push_back(in_result_vector[i_s][i_m][i_z]->tellg());}
  }
  in_result_vector[i_s][i_m][i_z]->clear();

  logger << LOG_DEBUG << "pos_line[" << i_s << "][" << i_m << "][" << i_z << "].size() = " << pos_line[i_s][i_m][i_z].size() << endline_flush;

  int temp_counter = 2; // 1 without 'group counter' (old version)
  for (int x_q = 0; x_q < ycontribution->selection_n_qTcut; x_q++){
    temp_counter++;
    for (int i_d = 0; i_d < osi->extended_distribution.size(); i_d++){
      logger << LOG_DEBUG << "set pos_qTcut_distribution:   i_s = " << i_s << "   x_q = " << x_q << "   i_d = " << i_d << endline_flush;
      no_line_distribution[i_s][i_d][x_q][0] = temp_counter;
      temp_counter += (osi->extended_distribution[i_d].n_bins + 1) * osi->n_scale_ren_TSV[i_s] * osi->n_scale_fact_TSV[i_s];

      pos_qTcut_distribution[i_s][i_m][i_z][x_q][i_d] = pos_line[i_s][i_m][i_z][no_line_distribution[i_s][i_d][x_q][0]];

      temp_counter++;
    }
  }

  // pos_line[i_s][i_m][i_z] not needed any longer

  // Probably only for checks (no function ???):
  for (int x_q = 0; x_q < ycontribution->selection_n_qTcut; x_q++){
    for (int i_d = 0; i_d < osi->extended_distribution.size(); i_d++){
      string temp_s;
      logger << LOG_DEBUG << "set in_result_vector:   i_s = " << i_s << "   x_q = " << x_q << "   i_d = " << i_d << endline_flush;
      in_result_vector[i_s][i_m][i_z]->seekg(pos_qTcut_distribution[i_s][i_m][i_z][x_q][i_d]);//, is.beg);
      getline(*in_result_vector[i_s][i_m][i_z], temp_s);
    }
  }
  in_result_vector[i_s][i_m][i_z]->clear();
  // ... until here.

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}


void summary_subprocess::cleanup(){
  Logger logger("summary_subprocess::cleanup");
  logger << LOG_DEBUG_VERBOSE << "called" << endline_flush;

  vector<vector<vector<vector<vector<vector<double> > > > > > ().swap(distribution_result_TSV);
  vector<vector<vector<vector<vector<vector<double> > > > > > ().swap(distribution_deviation_TSV);
  vector<vector<vector<vector<vector<vector<double> > > > > > ().swap(distribution_chi2_TSV);

  vector<vector<vector<vector<vector<vector<vector<double> > > > > > > ().swap(distribution_group_result_TSV);
  vector<vector<vector<vector<vector<vector<vector<double> > > > > > > ().swap(distribution_group_deviation_TSV);
  vector<vector<vector<vector<vector<vector<vector<double> > > > > > > ().swap(distribution_group_chi2_TSV);

  vector<vector<vector<vector<int> > > > ().swap(distribution_group_counter_removal_run_qTcut_TSV);
  vector<vector<vector<vector<int> > > > ().swap(distribution_group_counter_nonzero_run_qTcut_TSV);

  vector<vector<vector<vector<long long> > > > ().swap(distribution_group_N_binwise_TSV);

  logger << LOG_DEBUG_VERBOSE << "finished" << endline_flush;
}

